#========================================================================================
# package      DE_EPAGES::DHLIntraship::API::Object::RemoteRESTDHL
# state        public
#----------------------------------------------------------------------------------------
# description  object for REST requests for DHL Intraship
#========================================================================================
package DE_EPAGES::DHLIntraship::API::Object::RemoteRESTDHL;

use strict;
use HTTP::Request;
use JSON qw(from_json to_json);
use LWP::UserAgent;
use Encode qw ( encode );
use DE_EPAGES::Core::API::Log            qw( GetCommunicationLog );
use DE_EPAGES::Core::API::TestParameter  qw( TestString TestObject );
use DE_EPAGES::Core::API::Url            qw ( BuildParameterString );
use DE_EPAGES::Object::API::Factory      qw( LoadObjectByPath );
use DE_EPAGES::DHLIntraship::API::Constants qw ( $DHLVersendenSandboxURL $DHLVersendenSandboxProductionURL );
use DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::ShipmentOrder;

#ERRORDEFINITION
sub ErrorDefinition { return {}
}

sub new {
    my $class = shift;
    my %options = @_;

    my $Env = $options{'Environment'};

    $Env = LoadObjectByPath('ShippingFulfillmentTypes/DHLIntraship') unless defined $Env;
    $options{'Environment'} = $Env;

    return bless \%options, $class;
}

#========================================================================================
# function     sendRESTRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Shop->sendRESTRequest($Method, $APIUrl, $Route, $hParams);
# syntax       $Shop->sendRESTRequest('POST', 'https://api-sandbox.dhl.com/parcel/de/shipping/v2/',
#                                      /orders , $hParams);
#----------------------------------------------------------------------------------------
# description  Sends a request to the DHL API server with the given parameter.
#----------------------------------------------------------------------------------------
# input        $Method | request method  | string
# input        $APIUrl | API url  | string
# input        $Route | route to object | string
# input        $hParams | with following keys:
#               <ul>
#                   <li>Header - http header parameters (optional) - ref.hash
#                   <li>Query - query parameters for call (optional) - ref.hash
#                   <li>Content - payload for call (optional) - ref.hash
#                   <li>Authorization - authorization data - ref.array
#                   <li>IsTestMode - mode for getting APIKey - string
#               </ul> | ref.hash
# input        $aAuthorizationData | User authorization data   | ref.array
# input        $IsTestMode         | TesMode to be used 1 (Test) , 0 (Live)  | boolean

# return       $Response | JSON decoded response or undef | ref.hash
#========================================================================================
sub sendRestRequest {
    my $self = shift;
    my ($Method, $APIUrl, $Route, $hParams) = @_;

    $Method //= 'GET';
    TestString( 'Method', $Method );
    TestString( 'APIUrl', $APIUrl );
    TestString( 'Route', $Route );


    # encode content
    my $EncodedContent;
    $EncodedContent = encode('UTF-8',to_json($hParams->{'Content'})) if defined $hParams->{'Content'};

    #build correct request URL
    my $RequestURL;
    if (defined $hParams->{'URL'}) {
        $RequestURL = $self->_buildRequestURL($APIUrl, $Route, $hParams);
    } else {
        # build the correct request URL by adding route and query params
        $RequestURL = $self->_buildRequestURL($APIUrl, $Route, $hParams->{'Query'});
    }

    my $Request = HTTP::Request->new($Method => $RequestURL);
    my $hHTTPHeader = {
        'Content-Type'       => 'application/json',
        'Content-Length'     => length($hParams->{'Content'}),
        'Accept'             => 'application/json',
        'Accept-Charset'     => 'utf-8',
        'Accept-Language'    => $hParams->{'Header-Accept-Language'},
        'dhl-api-key'        => $self->appCredentials($hParams->{'IsTestMode'}),
    };

    $Request->authorization_basic(@{$hParams->{'Authorization'}});
    $Request->header(%$hHTTPHeader);

    # adds the content and send request
    $Request->content($EncodedContent) if defined $EncodedContent;

    my $LogCategory = $self->_getLogCategory();
    $self->logCommunication($LogCategory, $Request);
    my $Response = LWP::UserAgent->new->request($Request);
    $self->logCommunication( $LogCategory, $Response);

    # return decoded content if we have a GET request only
    my $JSONResponse = $Response->decoded_content;
    return $JSONResponse ? from_json($JSONResponse) : undef;

}

#========================================================================================
# function     environment
# state        public
#----------------------------------------------------------------------------------------
# syntax       $store = $self->environment;
#----------------------------------------------------------------------------------------
# description  Returns DHLIntraship environment.
#----------------------------------------------------------------------------------------
# return       $environment | DHLIntraship Environment | Object
#========================================================================================
sub environment {
    return shift->{'Environment'};
}

#========================================================================================
# function     appCredentials
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hCredentials = $self->appCredentials($IsTestMode);
#----------------------------------------------------------------------------------------
# description  Returns app api keys.
#----------------------------------------------------------------------------------------
# input        $IsTestMode | TestMode to be used 1 (Test) or 0 (Live) | boolean
# return       $result | with following keys:
#               <ul>
#                   <li>DHLApiKey - DHLVersenden application key - string
#               </ul> | ref.hash
#========================================================================================
sub appCredentials {
    my $self = shift;
    my ($IsTestMode) = @_;

    if ($IsTestMode){
        return $self->environment->get('DHLSandboxApiKey');
    } else {
        return $self->environment->get('DHLProductionApiKey');
    }
}

#========================================================================================
# function     _buildRequestURL
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Shop->_buildRequestURL($APIUrl, $Route, $hQueryParams);
# syntax       $Shop->_buildRequestURL($APIUrl, '/product_summary/search', {'validate'=> 'true'});
#----------------------------------------------------------------------------------------
# description  inserts the $Route into the given $APIUrl and adds adds $hQueryParams
#               afterwards to the URL
#----------------------------------------------------------------------------------------
# input        $APIUrl | API url | string
# input        $Route | route to add to the url | string
# input        $hQueryParams | locale id for call (optional) | string
# return       $APIUrl |  api URL including the route and the query params | string
#========================================================================================
sub _buildRequestURL{
    my $self = shift;
    my ($APIUrl, $Route, $hQueryParams) = @_;

    my $ParamString = defined $hQueryParams ? BuildParameterString($hQueryParams) . '&' : '';
    $APIUrl .= "$Route?$ParamString";

    return $APIUrl;
}

#========================================================================================
# function     logCommunication
# state        public
#----------------------------------------------------------------------------------------
# syntax       $RemoteRESTDHL->logCommunication( $Category, $HTTPObject );
# example      $RemoteRESTDHL->logCommunication( 'DHLIntraship::API::Object::$RemoteRESTDHL', $Response );
#----------------------------------------------------------------------------------------
# description  Logs HTTP communications.
#----------------------------------------------------------------------------------------
# input        $Category | log category, undef for no logging | string
# input        $Message | Message to be logged | HTTP::Request or HTTP::Response object
#========================================================================================
sub logCommunication {
    shift;
    my ($Category, $Message) = @_;

    my $Logger = GetCommunicationLog($Category);

    if (defined $Category && $Logger->is_debug) {
        if( $Message->isa('HTTP::Request') ) {
            $Logger->debug( "Request headers:\n" . $Message->headers_as_string );
            $Logger->debug( $Message->method .' '. $Message->uri ."\n". $Message->content);
        }
        else {
            $Logger->debug( "Response headers:\n" . $Message->headers_as_string );
            $Logger->debug( $Message->status_line . "\n" . $Message->content);
        }
    }
    return;
}

#========================================================================================
# function     _getLogCategory
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_getLogCategory();
#----------------------------------------------------------------------------------------
# description  Returns the log category if logging is enabled.
#----------------------------------------------------------------------------------------
# return       $LogCategory |  log category for enabled logging | string
#========================================================================================
sub _getLogCategory {
    my $self = shift;

    return undef unless $self->environment->get('Logging');
    return  "DHLIntraship::RemoteRESTDHL"
}

#========================================================================================
# function     _areValidCredentials
# state        private
#----------------------------------------------------------------------------------------
# syntax       $areValid = $RemoteRESTDHL->areValidCredentials( $Form );
#----------------------------------------------------------------------------------------
# description  Validate provided credentials against REST endpoint.
#               This will not create a shipping label.
#----------------------------------------------------------------------------------------
# input        $DHLBillingNumber | 14 digit DHL billing number | string
# input        $ShopAddress | Shops address object | object
# input        $aAuthorizationData | User authorization data | ref.array
# input        $IsTestMode | Is in test mode | boolean
# return       $areValid | credentials are successfully validated | boolean
#========================================================================================
sub validateCredentials {
    my $self = shift;
    my ($DHLBillingNumber, $ShippingFulfillmentMethod, $aAuthorizationData, $IsTestMode) = @_;

    my $Shop = $ShippingFulfillmentMethod->get('Shop');
    my $ShopAddress = $Shop->get('Address');
    my $ShopLocaleID = $Shop->get('LocaleID');

    my $hParams;
    my $Formatter = DE_EPAGES::Core::API::Object::DateTimeFormatter->new();
    my $FormattedDate = $Formatter->format_datetime(DateTime->today(), 'date', '%Y-%m-%d');
        # create dummy ShippingLabelData with the shop address for credential validation test
        $hParams->{'Content'} = {
            'profile' => 'STANDARD_GRUPPENPROFIL',
            'shipments' => [
                {
                    'product' => 'V01PAK',
                    'billingNumber' => $DHLBillingNumber,
                    'refNo' => 'Order No. 1234',
                    'shipDate' => $FormattedDate,
                    'shipper' => {
                        'name1' => $ShopAddress->get('FullName'),
                        'addressStreet' => $ShopAddress->get('Street'),
                        'postalCode' => $ShopAddress->get('Zipcode'),
                        'city' => $ShopAddress->get('City'),
                        'country' => $ShopAddress->get('Country')->{'Code3'},
                        'email' => $ShopAddress->get('EMail'),
                        'phone' => $ShopAddress->get('Phone')
                    },
                    'consignee' => {
                        'name1' => $ShopAddress->get('FullName'),
                        'addressStreet' => $ShopAddress->get('Street'),
                        'postalCode' => $ShopAddress->get('Zipcode'),
                        'city' => $ShopAddress->get('City'),
                        'country' => $ShopAddress->get('Country')->{'Code3'},
                    },
                    'details' => {
                        'dim' => {
                            'uom' => 'mm',
                            'height' => 100,
                            'length' => 200,
                            'width' => 150
                        },
                        'weight' => {
                            'uom' => 'g',
                            'value' => 500
                        }
                    }
                }
            ]
        };

    my $APIUrl = $self->_getDHLVersendenApiUrl($IsTestMode);

    my $Route = "orders";
    $hParams->{'Query'} =  {'validate' => 'true'};
    $hParams->{'IsTestMode'} = $IsTestMode;
    $hParams->{'Authorization'} = $aAuthorizationData;
    $hParams->{'Header-Accept-Language'} = 'en-GB';
    if ($ShopLocaleID eq 'de_DE') {
        $hParams->{'Header-Accept-Language'} = 'de-DE';
    }

    my $RemoteRESTDHL = DE_EPAGES::DHLIntraship::API::Object::RemoteRESTDHL->new();
    my $hParsedResponse = $RemoteRESTDHL->sendRestRequest(
        'POST',
        $APIUrl,
        $Route,
        $hParams,
    );

    return (defined $hParsedResponse->{'status'}->{'statusCode'} && $hParsedResponse->{'status'}->{'statusCode'} == 200) ? 1 : 0;
}

#========================================================================================
# function     createShipmentOrder
# state        public
#----------------------------------------------------------------------------------------
# syntax       createShipmentOrder( $ShippingLabel, 1 );
#----------------------------------------------------------------------------------------
# description  execute the createShipmentOrder REST request
#               throws HttpError if result is empty.
#----------------------------------------------------------------------------------------
# input        $ShippingLabel | Shipping label Object | Object
# input        $IsValidate | status if label data will only be validated | boolean
# return       $aResponse | Simplified response array | refarray
#========================================================================================
sub createShipmentOrder {
    my $self = shift;
    my ($ShippingLabel, $IsValidate) = @_;

    TestObject('ShippingLabelDHLIntraship', $ShippingLabel, 'DE_EPAGES::DHLIntraship::API::Object::ShippingLabelDHLIntraship');
    my $Shop = $ShippingLabel->getSite();
    my $ShopLocaleID = $Shop->get('LocaleID');

    my $hParams->{'Content'} = DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::ShipmentOrder->assemble($ShippingLabel);
    my $ShippingFullfilmentMethod = $ShippingLabel->get('ShippingFulfillmentMethod');

    $hParams->{'Authorization'} = $self->_getAuthorizationData($ShippingFullfilmentMethod);
    $hParams->{'Header-Accept-Language'} = 'en-GB';
    if ($ShopLocaleID eq 'de_DE') {
        $hParams->{'Header-Accept-Language'} = 'de-DE';
    }

    my $IsTestMode = $ShippingLabel->get('IsTestLabel');
    $hParams->{'IsTestMode'} = $IsTestMode;

    if ($IsValidate) {
        # We want only check if a Shippinglabel can be created
        $hParams->{'Query'} =  {'validate' => 'true'};
    } else {
        # We want the URL for the Shippinglabel
        $hParams->{'Query'} =  {'includeDocs'=> 'URL'};
    }


    my $DHLVersendenAPIUrl = $self->_getDHLVersendenApiUrl($IsTestMode);
    my $Route = "orders";

    my $RemoteRESTDHL = DE_EPAGES::DHLIntraship::API::Object::RemoteRESTDHL->new();
    my $hParsedResponse = $RemoteRESTDHL->sendRestRequest(
        'POST',
        $DHLVersendenAPIUrl,
        $Route,
        $hParams,
    );

    return $hParsedResponse;
}

sub _getAuthorizationData {
    my $self = shift;
    my ($ShippingFulfillmentMethod) = @_;

    my $DHLApiUsername = $ShippingFulfillmentMethod->get('WebserviceUsername');
    my $DHLApiPassword = $ShippingFulfillmentMethod->get('WebservicePassword');

    return [$DHLApiUsername,$DHLApiPassword];
}

sub _getDHLVersendenApiUrl {
    my $self = shift;
    my ($IsTestMode) = @_;

    my $DHLVersendenAPIUrl;
    if ($IsTestMode) {
        $DHLVersendenAPIUrl = $DHLVersendenSandboxURL;
    } else {
        $DHLVersendenAPIUrl = $DHLVersendenSandboxProductionURL;
    }

    return $DHLVersendenAPIUrl;
}
1;
