#========================================================================================
# package      DE_EPAGES::DHLIntraship::API::Object::ShippingLabelDHLIntraship
# state        public
# base         DE_EPAGES::ExternalShipping::API::Object::ShippingLabel
# description  object interface for ShippingLabelDHLIntraship.
#========================================================================================
package DE_EPAGES::DHLIntraship::API::Object::ShippingLabelDHLIntraship;
use base DE_EPAGES::ExternalShipping::API::Object::ShippingLabel;

use strict;
use DE_EPAGES::Core::API::Error qw ( Error ExistsError );
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Core::API::PerlTools qw ( GetCurrentDateTime );
use DE_EPAGES::DHLIntraship::API::ShippingLabelAttributes::CustomerAddress;
use DE_EPAGES::DHLIntraship::API::ShippingLabelAttributes::Products;
use DE_EPAGES::DHLIntraship::API::ShippingLabelAttributes::ShopAddress;
use DE_EPAGES::DHLIntraship::API::Tools;
use DE_EPAGES::Dictionary::API::Template qw ( ProcessFileTemplate );
use DE_EPAGES::Object::API::Table::Language qw ( GetInfoLanguage );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );

sub ErrorDefinition {
    return {
        'TotalWeightMissing' => 'TotalWeight for shippinglabel (#ShippingLabelID) is missing.',
        'ProductCodeMissing' => 'ProductCode for shippinglabel (#ShippingLabelID) is missing.',
        'AddressMissing'     => 'Address for shippinglabel (#ShippingLabelID) is missing.',
    };
}

#========================================================================================
# function     className
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClassName = $ShippingLabelDHLIntraship->className;
#----------------------------------------------------------------------------------------
# description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $ClassName | class name | String
#========================================================================================
sub className  { 'ShippingLabelDHLIntraship' }

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ShippingLabelDHLIntraship = DE_EPAGES::DHLIntraship::API::Object::ShippingLabelDHLIntraship->insert($Package, $hInfo);
# example      my $Class = LoadClassByAlias('ShippingLabelDHLIntraship');
#               my $ShippingLabelDHLIntraship = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertShippingLabelDHLIntraship' is triggered.
#----------------------------------------------------------------------------------------
# input        $Package | class package | string
# input        $hInfo | attributes of new object | ref.hash
# return       $ShippingLabelDHLIntraship | new object | object
# hook         OBJ_InsertShippingLabelDHLIntraship | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingLabelDHLIntrashipID - shipping label id - int
#               </ul>
#========================================================================================
sub insert {
    my ($Package, $hInfo) = @_;

    my $self = $Package->SUPER::insert($hInfo);

    # Default value for EstimatedDeliveryDate attribute
    $self->set({'EstimatedDeliveryDate' => GetCurrentDateTime()});
    $self->set({'ProductCode' => $self->_getPackageType()});

    TriggerHook('OBJ_InsertShippingLabelDHLIntraship',
                {'Object' => $self, 'ShippingLabelDHLIntrashipID' => $self->id });

    return $self;
}

#========================================================================================
# function     isDefined
# state        public
#----------------------------------------------------------------------------------------
# syntax       isDefined( $Value );
#----------------------------------------------------------------------------------------
# description  wrapping the isValid Function from PerlTools because 
#               in the external Shipping Frame work is a second Method with the same name
#----------------------------------------------------------------------------------------
# input        $Value | a String | string
# return       $Valid | parameter is simple valid | boolean
#========================================================================================
sub isDefined {
    my $self = shift;
    my ($Value) = @_;
    return DE_EPAGES::Core::API::PerlTools::isValid( $Value );
}
#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ShippingLabelDHLIntraship->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteShippingLabelDHLIntraship' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteShippingLabelDHLIntraship | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingLabelDHLIntrashipID - shipping label id - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeleteShippingLabelDHLIntraship',
                {'Object' => $self, 'ShippingLabelDHLIntrashipID' => $self->id });
    $self->SUPER::delete;
    return;
}

#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ShippingLabelDHLIntraship->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateShippingLabelDHLIntraship'
#               and 'OBJ_AfterUpdateShippingLabelDHLIntraship'.
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateShippingLabelDHLIntraship | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingLabelDHLIntrashipID - shipping label id - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateShippingLabelDHLIntraship | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingLabelDHLIntrashipID - shipping label id - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#----------------------------------------------------------------------------------------
# input       $hValues    | attribute name => value hash | ref.hash
# input       $LanguageID | ID of Language | String

#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    TriggerHook('OBJ_BeforeUpdateShippingLabelDHLIntraship',
        {'Object'=>$self,
            'Values'=> $hValues,
            'ShippingLabelDHLIntrashipID' => $self->id,
            'LanguageID' => $LanguageID});
    $self->SUPER::set($hValues, $LanguageID);
    TriggerHook('OBJ_AfterUpdateShippingLabelDHLIntraship',
                {'Object'=>$self,
                    'Values'=> $hValues,
                    'ShippingLabelDHLIntrashipID' => $self->id,
                    'LanguageID' => $LanguageID});
    return;
}

#========================================================================================
# function     validateLabel
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShippingLabel->validateLabel();
#----------------------------------------------------------------------------------------
# description  Validates shipping label. Throws an error if
#               mandatory values are missing.
#========================================================================================
sub validateLabel {
    my $self = shift;

   my @Mandatories = qw( TotalWeight EstimatedDeliveryDate );

    my $hInfo = $self->get(\@Mandatories);
    foreach my $Mandatory (@Mandatories) {
        Error($Mandatory.'Missing',{
            'ShippingLabel' => $self,
            'ShippingLabelID' => $self->id,
        }) unless($hInfo->{$Mandatory});
    }

    my $EstimatedDeliveryDate = $hInfo->{EstimatedDeliveryDate};
    if ($self->_isDateInThePast($EstimatedDeliveryDate)) {
        GetLog->debug(sprintf('EstimatedDeliveryDate is in the past: %s', $EstimatedDeliveryDate));
        Error('ShippingLabelError',{
            'Message' => 'EstimatedDeliveryDate is in the past.',
        });
    }

    my $ShopAddressValidation = DE_EPAGES::DHLIntraship::API::ShippingLabelAttributes::ShopAddress->new( $self );
    my $aShopAddressWarnings = $ShopAddressValidation->getNotProcessableWarningStates();
    if ( scalar(@{$aShopAddressWarnings}) > 0 ) {
        my $ShopAddress = $self->get('ShopAddress');
        Error('ShopAddressError', {
            'ShopAddress' => $ShopAddress,
            'ShopAddressID' => $ShopAddress->id(),
            'Message' => join('|', @{$aShopAddressWarnings}[0]->{'attribute'}.' '.@{$aShopAddressWarnings}[0]->{'reason'}),
        });
    }

    my $CustomerAddressValidation = DE_EPAGES::DHLIntraship::API::ShippingLabelAttributes::CustomerAddress->new( $self );
    my $aCustomerAddressWarnings = $CustomerAddressValidation->getNotProcessableWarningStates();
    if ( scalar(@{$aCustomerAddressWarnings}) > 0 ) {
        Error('CustomerAddressError',{
            'ShippingLabel' => $self,
            'ShippingLabelID' => $self->id,
            'Message' => join('|', @{$aCustomerAddressWarnings}[0]->{'attribute'}.' '.@{$aCustomerAddressWarnings}[0]->{'reason'}),
        });
    }

    my $ProductsValidation = DE_EPAGES::DHLIntraship::API::ShippingLabelAttributes::Products->new( $self );
    unless ( $ProductsValidation->isComplete() ) {
        Error('ProductError',{
           'Product' => $self,
           'ProductsID' => $self->id(),
           'Message' => join('|', @{$ProductsValidation->{'WarningStates'}}),
        });
    }
   return ;
}

#========================================================================================
# function     isValid
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShippingLabel->isValid();
#----------------------------------------------------------------------------------------
# description  checks that label is processable through DHL web services
#               $ShippingLabel->validateLabel();
# return       $IsValid - true if shipping label validates - boolean
#========================================================================================
sub isValid {
    my $self = shift;

   eval {
       $self->validateLabel();
   };
   return !ExistsError();
}

#========================================================================================
# function     generateTrackingLink
# state        public
#----------------------------------------------------------------------------------------
# syntax       $TrackingLink = $ShippingLabel->generateTrackingLink( $LangCode );
#----------------------------------------------------------------------------------------
# description  Returns tracking link generated from template
#               MBO/ShippingLabel/TrackingLink.txt
#----------------------------------------------------------------------------------------
# return       $TrackingLink | tracking link | string
#========================================================================================
sub generateTrackingLink {
    my $self     = shift;

    my $hLangInfo = GetInfoLanguage($self->get('LineItemContainer')->get('LanguageID'));

    my $TrackingLink = ProcessFileTemplate(
        $ENV{'EPAGES_CARTRIDGES'}
          . '/DE_EPAGES/DHLIntraship/Templates/MBO/ShippingLabel/TrackingLink.txt',
        $hLangInfo->{'Code2'}, {
            'ShippingLabel' => $self->tleHash,
            'Language'      => $hLangInfo,
        }
    );

    return $TrackingLink;
}

#---------------------------------------------------------------------------
#  Private methods
#---------------------------------------------------------------------------
sub _getPackageType {
    my $self = shift;

    my $ShippingFulfillmentMethod = $self->get('ShippingFulfillmentMethod');
    my $Order = $self->get('Order');
    my $Address = ($Order->get('ShippingAddress')) ? $Order->get('ShippingAddress') : $Order->get('BillingAddress');

    my $CustomerNumber;
    my $Tools = DE_EPAGES::DHLIntraship::API::Tools->new();
    if (defined($Address) && defined($ShippingFulfillmentMethod) ) {
        $CustomerNumber = $Tools->GetCustomerNumber($ShippingFulfillmentMethod, $Address);
    }

    return $Tools->GetPackageType($CustomerNumber);
}

sub _isDateInThePast {
    my $self = shift;
    my ($EstimatedDeliveryDate) = @_;

    my $Tools = DE_EPAGES::DHLIntraship::API::Tools->new();

    return $Tools->IsDateInThePast($EstimatedDeliveryDate);
}


1;
