#========================================================================================
# package      DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::Receiver
# state        public
# description  Receiver element for the DHL Versenden shipping label request builder
#========================================================================================
package DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::Receiver;

use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::Address;
use DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::Communication;

use strict;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Receiver = DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::Receiver->new();
#----------------------------------------------------------------------------------------
# description  Object contructor.
#----------------------------------------------------------------------------------------
# return       $Receiver | DHL Versenden receiver element object | object
#========================================================================================
sub new {
    my $class = shift;

    my $self  = bless {}, $class;

    return $self;
}

#========================================================================================
# function     assemble
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hReceiver = $self->assemble();
#----------------------------------------------------------------------------------------
# description  Create the hash with receiver data for the Shipping label request
#----------------------------------------------------------------------------------------
# input        $ShippingLabel | DHL Versenden shipping label object | object
# return       $hReceiver | DHL Versenden receiver element | ref.hash
#========================================================================================
sub assemble {
    my $self = shift;
    my ($ShippingLabel) = @_;

    my $CustomerAddress = $ShippingLabel->get('Address');

    # Due to the DSGVO, the merchant would need the consent of the customer to send email and phone number to DHL.
    # We can not be sure if the merchant has the consent so we do not email and phone if country of the customer is DE.
    my $CustomerCountryCode = $CustomerAddress->get('Country')->{'Code2'};
    my $hCommunication ={};
    if ($CustomerCountryCode ne 'DE') {
        # Phone and Email address
        $hCommunication = $self->_assembleCommunication($CustomerAddress);
    }

    my $hReceiver =  {%{$hCommunication}};
    my $hReceiverNames = $self->_getNamesPrioritizingCompanyOverFullName($CustomerAddress);

    $hReceiver->{'name1'} = $hReceiverNames->{'name1'};
    my $hAddress = $self->_assembleAddressReceiver($CustomerAddress, $hReceiverNames->{'name2'});

    return {
        'consignee' => {
            %{$hReceiver},
            %{$hAddress},
        }
    };
}

#========================================================================================
# Local methods
#========================================================================================
sub _assembleCommunication {
    my $self = shift;
    my ($Address) = @_;

    return DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::Communication->assemble($Address);
}

sub _assembleAddressReceiver {
    my $self = shift;
    my ($Address, $ReceiverName2) = @_;

    return DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::Address->assemble($Address, $ReceiverName2);
}

sub _getNamesPrioritizingCompanyOverFullName {
    my $self = shift;
    my ($Address) = @_;

    my $hNames = {};
    my $Company = $Address->get('Company');
    if (isValid($Company)) {
        $hNames->{'name1'} = $Company;
        $hNames->{'name2'} = $Address->get('FullName');
    } else {
        $hNames->{'name1'} = $Address->get('FullName');
        $hNames->{'name2'} = undef;
    }

    return $hNames;
}

1;