#========================================================================================
# package      DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::ShipmentDetails
# state        public
# description  ShipmentDetails element for the DHL Versenden shipping label request builder
#========================================================================================
package DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::ShipmentDetails;

use Date::Format qw ( time2str );

use DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::ShipmentItem;

use DE_EPAGES::DHLIntraship::API::Constants qw (
    $EKP_LENGHT
    $PRODUCTCODE_LENGHT
    %MAPPING_PRODUCTCODE_PRODUCTNAME
);

use strict;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShipmentDetails = DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::ShipmentDetails->new();
#----------------------------------------------------------------------------------------
# description  Object contructor.
#----------------------------------------------------------------------------------------
# return       $ShipmentDetails | DHL Versenden shipment details element object | object
#========================================================================================
sub new {
    my $class = shift;

    my $self  = bless {}, $class;

    return $self;
}

#========================================================================================
# function     assemble
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hShipmentDetails = $self->assemble();
#----------------------------------------------------------------------------------------
# description  Create the hash with shipment details data for the Shipping label request
#----------------------------------------------------------------------------------------
# input        $ShippingLabel | DHL Versenden shipping label object | object
# return       $hShipmentDetails | DHL Versenden shipment details element | ref.hash
#========================================================================================
sub assemble {
    my $self = shift;
    my ($ShippingLabel) = @_;

    my $ShippingFulfillmentMethod = $ShippingLabel->get('ShippingFulfillmentMethod');
    my $CustomerAddress = $ShippingLabel->get('Address');


    my $BillingNumber = $self->_getAccountNumber($ShippingFulfillmentMethod, $CustomerAddress);

    my $hShipmentItem = $self->_assembleShipmentItem($ShippingLabel);

    # The refNo. for DHL must be between 8 and 35 characters.
    my $RefNumber;
    my $ShippingLabelCharLength = length($ShippingLabel->alias());
    if ($ShippingLabelCharLength < 8) {
        my $NeededZeros = 8 - $ShippingLabelCharLength;
        my $AddedZeros = "0" x $NeededZeros;
        $RefNumber = $ShippingLabel->alias().'-'.$AddedZeros;
    } else {
        $RefNumber = $ShippingLabel->alias();
    }

    return {
                'product' => $self->_getProductCode($ShippingFulfillmentMethod, $CustomerAddress),
                'billingNumber' => $BillingNumber,
                'refNo' => $RefNumber,
                'shipDate' => $self->_getShipmentDate($ShippingLabel),
                # returnShipmentAccountNumber (optional. field to include again the DHL account number)
                # returnShipmentReference (optional.A reference number for better association purposes. It appears on return)
                %{$hShipmentItem},
                # no optional BankData (we don't use it)
    };
}

#========================================================================================
# Local methods
#========================================================================================
sub _getAccountNumber {
    my $self = shift;
    my ($ShippingFulfillmentMethod, $CustomerAddress) = @_;

    return DE_EPAGES::DHLIntraship::API::Tools->GetCustomerNumber($ShippingFulfillmentMethod, $CustomerAddress);
}

#========================================================================================
sub _getProductCode {
    my $self = shift;
    my ($ShippingFulfillmentMethod, $CustomerAddress) = @_;

    my $AccountNumber = $self->_getAccountNumber($ShippingFulfillmentMethod, $CustomerAddress);
    my $ProductCodeID = substr($AccountNumber, $EKP_LENGHT, $PRODUCTCODE_LENGHT);

    return $MAPPING_PRODUCTCODE_PRODUCTNAME{$ProductCodeID};
}

#========================================================================================
sub _getShipmentDate {
    my $self = shift;
    my ($ShippingLabel) = @_;

    my $EstimatedDeliveryDate = $ShippingLabel->get('EstimatedDeliveryDate');

    if ( defined $EstimatedDeliveryDate ) {
        return time2str('%Y-%m-%d', $EstimatedDeliveryDate->epoch);
    }
    return '';
}

#========================================================================================
sub _assembleShipmentItem {
    my $self = shift;
    my ($ShippingLabel) = @_;

    return DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::ShipmentItem->assemble($ShippingLabel);
}


1;