#========================================================================================
# package      DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::ShipmentItem
# state        public
# description  ShipmentItem element for the DHL Versenden shipping label request builder
#========================================================================================
package DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::ShipmentItem;

use DE_EPAGES::Object::API::Factory qw ( LoadObjectByPath );

use strict;

use constant MAX_WEIGHT => 31.5;
use constant MIN_WEIGHT => 0.1;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShipmentItem = DE_EPAGES::DHLIntraship::API::REST::ShippingLabelRequest::ShipmentItem->new();
#----------------------------------------------------------------------------------------
# description  Object contructor.
#----------------------------------------------------------------------------------------
# return       $ShipmentItem | DHL Versenden shipment item element object | object
#========================================================================================
sub new {
    my $class = shift;

    my $self  = bless {}, $class;

    return $self;
}

#========================================================================================
# function     assemble
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hShipmentItem = $self->assemble();
#----------------------------------------------------------------------------------------
# description  Create the hash with shipment item data for the Shipping label request
#----------------------------------------------------------------------------------------
# input        $ShippingLabel | DHL Versenden shipping label object | object
# return       $hShipmentItem | DHL Versenden shipment item element and warnings | ref.hash
#========================================================================================
sub assemble {
    my $self = shift;
    my ($ShippingLabel) = @_;

    my @Warnings = ();
    my $hWeight = $self->_getWeight($ShippingLabel);
    push(@Warnings, $hWeight->{'Warning'}) if (defined $hWeight->{'Warning'});

    my $UnitOfMeasurement = $ShippingLabel->get('SizeUnit');
    my $OriginalLengthFactor = $UnitOfMeasurement->get('Factor');
    my $CentimetreFactor = $self->_getCentimetreFactor();
    my $Length = $ShippingLabel->get('ParcelLength');
    my $Width = $ShippingLabel->get('ParcelWidth');
    my $Height = $ShippingLabel->get('ParcelDepth');

    my $hShipmentItemData = { 'weight' => { 'uom'   => 'kg',
                                            'value' => $hWeight->{'Value'}

                                          },
                                'dim'   => {    'uom'       => 'cm',
                                                'height'    => $self->_refactorAndConvertToInteger($Height, $OriginalLengthFactor, $CentimetreFactor) // '',
                                                'length'    => $self->_refactorAndConvertToInteger($Length, $OriginalLengthFactor, $CentimetreFactor) // '',
                                                'width'     => $self->_refactorAndConvertToInteger($Width, $OriginalLengthFactor, $CentimetreFactor) // ''
                                           }
                            };

    return {
        'details' => {%{$hShipmentItemData}},
    };
}

#========================================================================================
# Local methods
#========================================================================================
sub _getWeight {
    my $self = shift;
    my ($ShippingLabel) = @_;

    my $totalWeight = $ShippingLabel->get('TotalWeight');

    $totalWeight = sprintf('%.3f', $totalWeight);


    if ( defined $totalWeight ){
        if ( $totalWeight > MAX_WEIGHT ){
            return { 'Value' => $totalWeight, 'Warning' => { 'weightInKG' => 'TotalWeightTooHigh' } };
        }
        elsif ( $totalWeight < MIN_WEIGHT ){
            return { 'Value' => $totalWeight, 'Warning' => { 'weightInKG' => 'TotalWeightTooLow' } };
        }
        return { 'Value' => $totalWeight, 'Warning' => undef };
    }

    return { 'Value' => $totalWeight, 'Warning' => undef };
}

#========================================================================================
sub _refactorAndConvertToInteger {
    my $self = shift;
    my ( $valueToConvert, $OriginalFactor,  $GoalFactor) = @_;

    my $ConversionFactor = $OriginalFactor / $GoalFactor;

    return int($valueToConvert * $ConversionFactor);
}

#========================================================================================
# Wrappers
#========================================================================================
sub _getCentimetreFactor {
    my $self = shift;

    return LoadObjectByPath('/Units/metre/centimetre')->get('Factor');
}

1;