#========================================================================================
# package      DE_EPAGES::DHLIntraship::t::RemoteCreateShippingOrder
# base         Test::Class
# state        private
#----------------------------------------------------------------------------------------
# description  Tests for createShippingOrder SOAP call (integration tests)
#========================================================================================
package DE_EPAGES::DHLIntraship::t::RemoteCreateShippingOrder;

use DE_EPAGES::Core::API::PerlTools qw ( GetCurrentDateTime );
use DE_EPAGES::Core::API::Script qw ( RunScript );
use DE_EPAGES::Database::API::Connection qw ( RunOnStore GetTestSite );
use DE_EPAGES::Object::API::Factory qw(LoadObjectByPath ExistsObjectByPath LoadRootObject);
use DE_EPAGES::Object::API::XML::ObjectImportHandler qw ( ImportObjects );

use DE_EPAGES::DHLIntraship::API::SOAP::ShippingLabelRequest::Builder;
use DE_EPAGES::DHLIntraship::API::SOAP::ShippingLabelResponse::Parser;
use DE_EPAGES::DHLIntraship::API::SOAP::WebServiceClient;

use Test::More;

use strict;

use constant ROOTOBJECT => 'Test';
use constant ROOTPATH => '/' . ROOTOBJECT;
use constant SHOPPATH => ROOTPATH . '/TestShopDHLVersenden';
use constant RESPONSE_SHIPMENT_NUMBER_MAX_LENGTH => 39;

sub Init {

    return if ExistsObjectByPath(ROOTPATH) && $ENV{'SKIP_CLEANUP'};

    Cleanup();

    my $System = LoadRootObject();
    $System->insertChild( 'Object', { Alias => ROOTOBJECT } );
    note 'importing shop ' . SHOPPATH;
    ImportObjects('import.xml', ROOTPATH);

    return;
}

sub Cleanup {

    return if $ENV{'SKIP_CLEANUP'};

    my $System = LoadRootObject();
    if( $System->existsChild(ROOTOBJECT) ) {
        note 'deleting shop ' . SHOPPATH;
        $System->child(ROOTOBJECT)->delete();
    }

    return;
}

sub test {
    my $self = shift;

    $self->shipping_label_request_positive_test();
    $self->shipping_label_request_negative_test();
}

#========================================================================================
# Test methods
#========================================================================================
sub test_xml_import {
    my $subTestName = (caller(0))[3];

    my $testImportOK = 0;

    my $Shop = LoadObjectByPath(SHOPPATH);
    $testImportOK = isa_ok($Shop, 'DE_EPAGES::Shop::API::Object::Shop', "$subTestName - test shop found");

    my $ShippingFulfillmentMethod = $Shop->path('ShippingFulfillmentMethods/DHLIntraship');
    $testImportOK = isa_ok($ShippingFulfillmentMethod,
            'DE_EPAGES::DHLIntraship::API::Object::ShippingFulfillmentMethodDHLIntraship',
            "$subTestName - test ShippingFulfillmentMethodDHLIntraship found")
                        && $testImportOK;

    my $Customer = $Shop->path('Customers/TestCustomer');
    $testImportOK = isa_ok($Customer,
            'DE_EPAGES::Customer::API::Object::Customer',
             "$subTestName - test customer found")
                        && $testImportOK;

    my $Order = $Shop->path('Customers/TestCustomer/Orders/1234-5678');
    $testImportOK = isa_ok($Order,
            'DE_EPAGES::Order::API::Object::CustomerOrder',
             "$subTestName - test order found")
                        && $testImportOK;

    my $ShippingLabel = $Shop->path('Customers/TestCustomer/Orders/1234-5678/ShippingLabels/1234-5678-1');
    $testImportOK = isa_ok($ShippingLabel,
            'DE_EPAGES::DHLIntraship::API::Object::ShippingLabelDHLIntraship',
             "$subTestName - test ShippingLabelDHLIntraship found")
                        && $testImportOK;

    return $testImportOK;
}

sub shipping_label_request_positive_test {
    my $self = shift;
    my $subTestName = (caller(0))[3];

    my $Shop = LoadObjectByPath(SHOPPATH);
    my $ShippingLabel = $Shop->path('Customers/TestCustomer/Orders/1234-5678/ShippingLabels/1234-5678-1');
    $ShippingLabel->set( { 'EstimatedDeliveryDate' => GetCurrentDateTime() } );

    my $RemoteRESTDHL = DE_EPAGES::DHLIntraship::API::Object::RemoteRESTDHL->new();
    my $Response = $RemoteRESTDHL->createShipmentOrder($ShippingLabel);

    is($Response->{status}->{'statusCode'}, 200, "$subTestName - Shipping label response status code is ok (200)");
    cmp_ok( length($Response->{'items'}[0]->{'shipmentNo'}), '>', 0,
            "$subTestName - Shipping label response shipment number is > 0");

    cmp_ok( length($Response->{'items'}[0]->{'shipmentNo'}), '<=', RESPONSE_SHIPMENT_NUMBER_MAX_LENGTH,
             "$subTestName - Shipping label response shipment number is <= " . RESPONSE_SHIPMENT_NUMBER_MAX_LENGTH);

}

sub shipping_label_request_negative_test {
    my $self = shift;
    my $subTestName = (caller(0))[3];

    my $Shop = LoadObjectByPath(SHOPPATH);
    my $ShippingLabel = $Shop->path('Customers/TestCustomer/Orders/1234-5678/ShippingLabels/1234-5678-1');

    $ShippingLabel->set( { 'TotalWeight' => '' } );

    my $RemoteRESTDHL = DE_EPAGES::DHLIntraship::API::Object::RemoteRESTDHL->new();
    my $Response = $RemoteRESTDHL->createShipmentOrder($ShippingLabel);

    is($Response->{items}[0]->{'validationMessages'}[0]->{'validationMessage'}, 'The weight is too small.', "$subTestName - Response is a hard validation error (parcel weight is to small)");
}


RunScript(
    Sub => sub {
        RunOnStore( 'Store' => GetTestSite(), Sub => sub {
            Init();
            unless ( test_xml_import() ) {
                done_testing();
                exit;
            }

            SKIP: {
                skip 'external services tests' if $ENV{'SKIP_EXTERNAL_TESTS'};
                my $self = __PACKAGE__;
                $self->test();
            }
            done_testing();
            Cleanup();
        });
    }
);

1;
