#========================================================================================
# package      DE_EPAGES::Ebay::API::Object::EbayEnvironment
# state        public
# base         DE_EPAGES::Shop::API::Object::ShopObject
# description  object interface for EbayEnvironment.
#========================================================================================
package DE_EPAGES::Ebay::API::Object::EbayEnvironment;
use base DE_EPAGES::Shop::API::Object::ShopObject;

use strict;
use DE_EPAGES::Object::API::Factory qw (LoadClassByAlias);
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::Ebay::API::Table::EbayCall qw ( DeleteEbayCall GetPKeyEbayCallsByEbayEnvironmentID );

#========================================================================================
# function     className
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClassName = $EbayEnvironment->className;
#----------------------------------------------------------------------------------------
# description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $ClassName | class name | String
#========================================================================================
sub className  { 'EbayEnvironment' }

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Object = $Package->new('ObjectID' => $ObjectID, ... );
#----------------------------------------------------------------------------------------
# description  overloads DE_EPAGES::Core::API::BaseObject->new
#               to decrypt attributes before store in cache
#----------------------------------------------------------------------------------------
# input        %Attributes | attribute value hash | hash
# return       $Object | existing object | object
#========================================================================================
sub new {
    my $self = shift;
    my (%options) = @_;

    $self->decryptBaseAttributes(\%options);

    return $self->SUPER::new(%options);
}

#========================================================================================
# function     dbPackage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DbPackage = $EbayEnvironment->dbPackage;
#----------------------------------------------------------------------------------------
# description  Returns the database package, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $DbPackage | database package name | String
#========================================================================================
sub dbPackage { __PACKAGE__ }

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $EbayEnvironment = DE_EPAGES::Ebay::API::Object::EbayEnvironment->insert($hInfo);
# example      my $Class = LoadClassByAlias('EbayEnvironment');
#               my $EbayEnvironment = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertEbayEnvironment' is triggered.
#----------------------------------------------------------------------------------------
# input        $hInfo | attributes of new object | ref.hash
# return       $EbayEnvironment | new object | object
# hook         OBJ_InsertEbayEnvironment | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>EbayEnvironmentID - object id of environment - int
#               </ul>
#========================================================================================
sub insert {
    my $class = shift;
    my ($hInfo) = @_;

    my $self = $class->SUPER::insert($hInfo);

    TriggerHook('OBJ_InsertEbayEnvironment', {'Object' => $self, 'EbayEnvironmentID' => $self->id });
    delete $hInfo->{$_} foreach qw ( EbaySiteId EbaySiteCode IsSandBox IsVisible CurrencyID LanguageID IconPath ApiUrl WebUrl SignInUrl Name LocaleID HasGetItFast HasEbayPlus AttrSetIDCondition ); ## no critic qw(ProhibitLongLines)
    return $self;
}

#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $EbayEnvironment->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteEbayEnvironment' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteEbayEnvironment | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>EbayEnvironmentID - object id of environment - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeleteEbayEnvironment', {'Object' => $self, 'EbayEnvironmentID' => $self->id });
    DeleteEbayCall($_) foreach @{GetPKeyEbayCallsByEbayEnvironmentID($self->id)};
    LoadClassByAlias('EbayPaymentMethod')->deleteObjects('ByEbayEnvironmentID', [$self->id]);
    LoadClassByAlias('EbayShippingLocation')->deleteObjects('ByEbayEnvironmentID', [$self->id]);
    LoadClassByAlias('EbayShippingMethod')->deleteObjects('ByEbayEnvironmentID', [$self->id]);
    LoadClassByAlias('EbayShopEnvironment')->deleteObjects('ByEbayEnvironmentID', [$self->id]);
    $self->SUPER::delete;
    return;
}

#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $EbayEnvironment->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateEbayEnvironment' and 'OBJ_AfterUpdateEbayEnvironment'.
#----------------------------------------------------------------------------------------
# input        $hValues | attribute name => value hash | ref.hash
# input        $LanguageID | (optional; required for localized attributes) language id | int
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateEbayEnvironment | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>EbayEnvironmentID - object id of environment - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateEbayEnvironment | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>EbayEnvironmentID - object id of environment - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    TriggerHook('OBJ_BeforeUpdateEbayEnvironment', {'Object'=>$self, 'Values'=> $hValues, 'EbayEnvironmentID' => $self->id, 'LanguageID' => $LanguageID});
    $self->SUPER::set($hValues, $LanguageID);
    TriggerHook('OBJ_AfterUpdateEbayEnvironment',  {'Object'=>$self, 'Values'=> $hValues, 'EbayEnvironmentID' => $self->id, 'LanguageID' => $LanguageID});
    return;
}

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::Ebay::API::Object::EbayEnvironment->default( $Parent, $hInfo );
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;
    return {
        %{$class->SUPER::default($Parent, $hInfo)},
        'IsSandBox' => 0,
        'IsVisible' => 0,
        'HasGetItFast' => 0,
        'HasEbayPlus' => 0,

    };
}

1;
