#========================================================================================
# package      DE_EPAGES::Ebay::API::Object::EbayShopEnvironment
# state        public
# base         DE_EPAGES::Shop::API::Object::ShopObject
# description  object interface for EbayShopEnvironment.
#========================================================================================
package DE_EPAGES::Ebay::API::Object::EbayShopEnvironment;
use base DE_EPAGES::Shop::API::Object::ShopObject;

use strict;
use DE_EPAGES::Object::API::Factory qw (LoadClassByAlias);
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::Ebay::API::Object::RemoteRESTEbay;
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Ebay::API::Table::EbayCustomerMap qw ( DeleteEbayCustomerMap GetPKeyEbayCustomerMapsByEbayShopEnvironmentID );

#========================================================================================
# function     className
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClassName = $EbayShopEnvironment->className;
#----------------------------------------------------------------------------------------
# description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $ClassName | class name | String
#========================================================================================
sub className  { 'EbayShopEnvironment' }

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Object = $Package->new('ObjectID' => $ObjectID, ... );
#----------------------------------------------------------------------------------------
# description  overloads DE_EPAGES::Core::API::BaseObject->new
#               to decrypt attributes before store in cache
#----------------------------------------------------------------------------------------
# input        %Attributes | attribute value hash | hash
# return       $Object | existing object | object
#========================================================================================
sub new {
    my $self = shift;
    my (%options) = @_;

    $self->decryptBaseAttributes(\%options);

    return $self->SUPER::new(%options);
}

#========================================================================================
# function     dbPackage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DbPackage = $EbayShopEnvironment->dbPackage;
#----------------------------------------------------------------------------------------
# description  Returns the database package, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $DbPackage | database package name | String
#========================================================================================
sub dbPackage { __PACKAGE__ }

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $EbayShopEnvironment = DE_EPAGES::Ebay::API::Object::EbayShopEnvironment->insert($hInfo);
# example      my $Class = LoadClassByAlias('EbayShopEnvironment');
#               my $EbayShopEnvironment = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertEbayShopEnvironment' is triggered.
#----------------------------------------------------------------------------------------
# input        $hInfo | attributes of new object | ref.hash
# return       $EbayShopEnvironment | new object | object
# hook         OBJ_InsertEbayShopEnvironment | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>EbayShopEnvironmentID - object id of EbayShopEnvironment - int
#               </ul>
#========================================================================================
sub insert {
    my $class = shift;
    my ($hInfo) = @_;

    my $self = $class->SUPER::insert($hInfo);

    TriggerHook('OBJ_InsertEbayShopEnvironment', {'Object' => $self, 'EbayShopEnvironmentID' => $self->id });
    delete $hInfo->{$_} foreach qw ( Shop EbayEnvironment EbayTokenExpiryDate EbayUserName EbayCountry EbayRegion EbayLocation Name LastUpdate  EbayPostalCode  PaypalAccount ); ## no critic qw(ProhibitLongLines)
    return $self;
}

#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $EbayShopEnvironment->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteEbayShopEnvironment' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteEbayShopEnvironment | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>EbayShopEnvironmentID - object id of EbayShopEnvironment - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeleteEbayShopEnvironment', {'Object' => $self, 'EbayShopEnvironmentID' => $self->id });
    LoadClassByAlias('EbayAuction')->deleteObjects('ByEbayShopEnvironmentID', [$self->id]);
    DeleteEbayCustomerMap($_) foreach @{GetPKeyEbayCustomerMapsByEbayShopEnvironmentID($self->id)};
    $self->SUPER::delete;
    TriggerHook('OBJ_AfterDeleteEbayShopEnvironment', {'Object' => $self });

    return;
}

#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $EbayShopEnvironment->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateEbayShopEnvironment' and 'OBJ_AfterUpdateEbayShopEnvironment'.
#----------------------------------------------------------------------------------------
# input        $hValues | attribute name => value hash | ref.hash
# input        $LanguageID | (optional; required for localized attributes) language id | int
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateEbayShopEnvironment | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>EbayShopEnvironmentID - object id of EbayShopEnvironment - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateEbayShopEnvironment | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>EbayShopEnvironmentID - object id of EbayShopEnvironment - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    TriggerHook('OBJ_BeforeUpdateEbayShopEnvironment', {'Object'=>$self, 'Values'=> $hValues, 'EbayShopEnvironmentID' => $self->id, 'LanguageID' => $LanguageID});
    $self->SUPER::set($hValues, $LanguageID);
    TriggerHook('OBJ_AfterUpdateEbayShopEnvironment',  {'Object'=>$self, 'Values'=> $hValues, 'EbayShopEnvironmentID' => $self->id, 'LanguageID' => $LanguageID});
    return;
}

#========================================================================================
# function     remoteEbay
# state        private
#----------------------------------------------------------------------------------------
# syntax       $RemoteEbay = $EbayShopEnvironment->remoteEbay;
#----------------------------------------------------------------------------------------
# description  Returns the <package DE_EPAGES::Ebay::API::Object::RemoteEbay> object for
#               the ebay shop environment which is assigned this this EbayAuction.
#----------------------------------------------------------------------------------------
# return       $RemoteEbay | perl package for communication between epages and ebay | object
#========================================================================================
sub remoteEbay {
    my $self = shift;

    return DE_EPAGES::Ebay::API::Object::RemoteRESTEbay->new(
        'ShopEnvironment' => $self,
    );
}

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::Ebay::API::Object::EbayShopEnvironment->default( $Parent, $hInfo );
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;
    return {
        %{$class->SUPER::default($Parent, $hInfo)},
        'Shop' => $Parent->getSite(),
    };
}

#========================================================================================
# function     getSellerProfiles
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $hBusinessPolicySellerProfiles = $EbayShopEnvironment->getSellerProfiles;
#----------------------------------------------------------------------------------------
# description  Returns the ebay seller profile of an ebay shop environment
#----------------------------------------------------------------------------------------
# return       $hresult | hash with attribute values of the seller profile | ref.hash
#========================================================================================
sub getSellerProfiles {
    my $self = shift;
    my ($hSellerProfiles) = @_;

    return $hSellerProfiles
        unless $self->isSellerProfileEmpty($hSellerProfiles);


    my $RemoteEbay = DE_EPAGES::Ebay::API::Object::RemoteRESTEbay->new(
        'ShopEnvironment' => $self,
    );

    my $hresult = $RemoteEbay->getUserPreferences([
        'ShowSellerPaymentPreferences',
        'ShowSellerExcludeShipToLocationPreference',
        'ShowSellerProfilePreferences']);

    return $hresult;
}

#========================================================================================
# function     isSellerProfileEmpty
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $isSellerProfileEmpty = $EbayShopEnvironment->isSellerProfileEmpty;
#----------------------------------------------------------------------------------------
# description  Returns true if the sellerProfile hash is empty
#----------------------------------------------------------------------------------------
# return       $isSellerProfileEmpty | isSellerProfileEmpty | boolean
#========================================================================================
sub isSellerProfileEmpty {
    my $self = shift;
    my ($hSellerProfiles) = @_;

    return 0
        if (defined $hSellerProfiles && %{$hSellerProfiles});

    return 1;
}

#========================================================================================
# function     filterSellerProfilesByDefault
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hDefaultProfile = $EbayShopEnvironment->filterSellerProfilesByDefault(
#                                       $hBusinessPolicySellerProfiles, 'PAYMENT');
#----------------------------------------------------------------------------------------
# description  Returns the default profile for a given seller profile type (business policy type)
#----------------------------------------------------------------------------------------
# input        $hResult | the whole result of the seller profile call | ref.hash
# input        $sellerProfileType | type of the profile to be filtered | string
#----------------------------------------------------------------------------------------
# return       $hFilteredSellerProfile | hash with attribute values of the default seller profile | ref.hash
#========================================================================================
sub filterSellerProfilesByDefault {
    my $self = shift;
    my ($hresult, $sellerProfileType) = @_;

    my $ahFilteredSellerProfiles = $self->filterSellerProfiles($hresult, $sellerProfileType);

    foreach my $hFilteredSellerProfile (@{$ahFilteredSellerProfiles}) {
        return $hFilteredSellerProfile if $hFilteredSellerProfile->{'CategoryGroup'}->{'IsDefault'} eq 'true';
    }

    return undef;
}


#========================================================================================
# function     filterSellerProfiles
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ahFilteredSellerProfiles = $self->filterSellerProfiles($hresult, $sellerProfileType);
#----------------------------------------------------------------------------------------
# description  Returns the ebay seller profiles for a given type ('PAYMENT', 'SHIPPING' or 'RETURN_POLICY')
#----------------------------------------------------------------------------------------
# input        $hResult | the whole result of the seller profile call | ref.hash
# input        $sellerProfileType | type of the profile to be filtered | string
#----------------------------------------------------------------------------------------
# return       $hresult | hash with attribute values of the filtered seller profile | ref.hash
#========================================================================================
sub filterSellerProfiles {
    my $self = shift;
    my ($hresult, $sellerProfileType) = @_;

    my @filteredSellerProfile = ();

    my $hSellerProfilePreferences = $hresult->{'SellerProfilePreferences'};
    return unless isValid($hSellerProfilePreferences);

    my $hSupportedSellerProfiles = $hSellerProfilePreferences->{'SupportedSellerProfiles'};
    return unless isValid($hSupportedSellerProfiles);

    my $ahSupportedSellerProfiles = $hSupportedSellerProfiles->{'SupportedSellerProfile'};
    return unless isValid($ahSupportedSellerProfiles);

    foreach my $hSupportedSellerProfile (@{$ahSupportedSellerProfiles}) {
       push(@filteredSellerProfile,$hSupportedSellerProfile) if $hSupportedSellerProfile->{'ProfileType'} eq uc($sellerProfileType);
    }

    return \@filteredSellerProfile;
}


#========================================================================================
# function     getBusinessPolicies
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aBusinessPolicies = $ShopEnv->getBusinessPolicies();
#----------------------------------------------------------------------------------------
# description  Returns an array with all business policies for a shop environment
#----------------------------------------------------------------------------------------
# return       $hArray | array with all business policies | ref.array
#========================================================================================
sub getBusinessPolicies {
    my $self = shift;

    my $RemoteEbay = DE_EPAGES::Ebay::API::Object::RemoteRESTEbay->new(
        'ShopEnvironment' => $self,
    );

    return $RemoteEbay->getUserPreferences(['ShowSellerProfilePreferences']);
}

#========================================================================================
# function     setEbayOutOfStockPreference
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShopEnv->setEbayOutOfStockPreference();
#----------------------------------------------------------------------------------------
# description  Set the value for the OtOfStockPreference from the sellers ebay account settings
#----------------------------------------------------------------------------------------
# return       $hArray | array with all business policies | ref.array
#========================================================================================
sub setEbayOutOfStockPreference {
    my $self = shift;

    my $RemoteEbay = DE_EPAGES::Ebay::API::Object::RemoteRESTEbay->new(
        'ShopEnvironment' => $self,
    );

    my $EbayOutOfStockPreference;
    my $response = $RemoteEbay->getUserPreferences(['ShowOutOfStockControlPreference']);

    if ( $response->{'OutOfStockControlPreference'} eq 'true') {
        $EbayOutOfStockPreference = 1;
    } elsif ( $response->{'OutOfStockControlPreference'} eq 'false') {
        $EbayOutOfStockPreference = 0;
    }

    $self->set({'EbayOutofStockPreference' => $EbayOutOfStockPreference} ) if defined $EbayOutOfStockPreference;

    return;
}

#========================================================================================
# function     setEbayPlusStatus
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShopEnv->setEbayPlusStatus();
#----------------------------------------------------------------------------------------
# description  Set the value for the EbayPlus account settings from the sellers ebay account
#----------------------------------------------------------------------------------------
# return       $hArray | array with all business policies | ref.array
#========================================================================================
sub setEbayPlusStatus {
    my $self = shift;

    my $RemoteEbay = DE_EPAGES::Ebay::API::Object::RemoteRESTEbay->new(
        'ShopEnvironment' => $self,
    );

    my $EbayPlusStatus;
    my $EbayCountry = $self->get('EbayCountry');

    my $response = $RemoteEbay->getUserPreferences(['ShoweBayPLUSPreference']);
    if ( defined $response->{'eBayPLUSPreference'} && $response->{'eBayPLUSPreference'}->{'Country'} eq $EbayCountry) {
        $EbayPlusStatus = 1;
    } else {
        $EbayPlusStatus = 0;
    }

    $self->set({'EbayPlus' => $EbayPlusStatus} ) if defined $EbayPlusStatus;

    return;
}


#========================================================================================
# function     isBusinessPolicyActivated
# state        public
#----------------------------------------------------------------------------------------
# syntax       $isBusinessPoliciesActivated = $ShopEnv->isBusinessPolicyActivated($aSellerProfiles);
#----------------------------------------------------------------------------------------
# description  Returns 1 if the business policy feature is activated for the used ebay account
#----------------------------------------------------------------------------------------
# input        $hResult | the whole result of the seller profile call | ref.hash
#----------------------------------------------------------------------------------------
# return       $result | returns 1 if business policy is activate | boolean
#========================================================================================
sub isBusinessPolicyActivated {
    my $self = shift;
    my ($hresult) = @_;

    my $hSellerProfilePreferences = $hresult->{'SellerProfilePreferences'};
    return unless defined $hSellerProfilePreferences;

    my $isBusinessPoliciesActivated = $hSellerProfilePreferences->{'SellerProfileOptedIn'};
    return unless defined $isBusinessPoliciesActivated;

    return $isBusinessPoliciesActivated eq 'true' ? 1 : 0;
}

1;
