#========================================================================================
# package      DE_EPAGES::MetaCatalog::API::Object::MetaCatalogPortalConfig
# base         DE_EPAGES::ProductPortal::API::Object::PortalConfig
# state        public
#----------------------------------------------------------------------------------------
# description  object interface for PortalConfig.
#========================================================================================
package DE_EPAGES::MetaCatalog::API::Object::MetaCatalogPortalConfig;
use base DE_EPAGES::ProductPortal::API::Object::PortalConfig;

use strict;
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias LoadObject );
use DE_EPAGES::Core::API::PerlTools qw ( unique );
use DE_EPAGES::ProductPortal::API::Table::PortalProductMap qw ( GetPKeyPortalProductMapsByPortalConfigID );

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::MetaCatalog::API::Object::MetaCatalogPortalConfig->default( $Parent, $hInfo );
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;

    my $ProductClass = LoadClassByAlias('Product');
    return {
        %{$class->SUPER::default($Parent, $hInfo)},
        'ImageAttribute' => $ProductClass->attribute('ImageLarge'),
        'ExportInterval' => 1, # days automatic export launches since last export
        'ExportAllProducts' => 1,
        'ExportAllVariations' => 1,
    };
}

#========================================================================================
# function     items
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aProductIDs = $PortalConfig->items;
#----------------------------------------------------------------------------------------
# description  Returns all product identifier of portal config
#                 Taken from the base DE_EPAGES::ProductPortal::API::Object::PortalConfig
#                 and overwritten because we will always export variations individually.
#                 The option for "ExportAllVariaions" checks for the attribute "DefaultSubProduct",
#                 which is NOT set for NOW, and the logic should probably use "IsMaster" instead anyway!
#----------------------------------------------------------------------------------------
# return       $aProductIDs | product identifier | ref.array.int
#========================================================================================
sub items {
    my $self = shift;

    if ($self->get('ExportAllProducts')) {
        # export all products
        return $self->allProducts();
    }

    my @Products = ();
    my @AssignedProductIDs = map { $_->{'ProductID'} } @{GetPKeyPortalProductMapsByPortalConfigID($self->id)};

    if ($self->get('ExportAllVariations')) {
        # export SubProducts only
        foreach my $ProductID (@AssignedProductIDs) {
            my $Product = LoadObject($ProductID);
            if ($Product->get('IsMaster')) {
                my $aSubProducts = $Product->get('SubProducts');
                push(@Products, $_) foreach grep { $_->get('IsVisible') } @$aSubProducts;
            }
            else {
                push(@Products, $Product) if ($Product->get('IsVisible'));
            }
        }
    }
    else {
        # if SuperProduct then export DefaultSubProduct only
        foreach my $ProductID (@AssignedProductIDs) {
            my $Product = LoadObject($ProductID);
            my $DefaultSubProduct = $Product->get('DefaultSubProduct');
            if (defined($DefaultSubProduct) && $DefaultSubProduct->get('IsVisible')) {
                push(@Products, $DefaultSubProduct);
            }
            else {
                push(@Products, $Product) if ($Product->get('IsVisible'));
            }
        }
    }

    if (!$self->get('PortalSite')->get('ExportsDownloadProducts')) {
        @Products = grep { !$_->get('IsDownloadProduct') } @Products;
    }

    # The base products are replaced by their subproducts, so we could have repeated items if the merchant
    # mixes base and subproducts.
    # To avoid repeated items we apply the unique function ...
    @Products = unique( map { $_->id } @Products ) ;

    return \@Products ;
}

1;
