#========================================================================================
# package      DE_EPAGES::MultiStoreCommon::API::Cartridge
# base         DE_EPAGES::Design::API::DesignInstaller
# state        public
#----------------------------------------------------------------------------------------
# description  This is the main cartridge class for install/patch/uninstall.
#========================================================================================
package DE_EPAGES::MultiStoreCommon::API::Cartridge;
use base DE_EPAGES::Design::API::DesignInstaller;

use strict;
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Object::API::Factory qw (LoadRootObject);
use DE_EPAGES::Trigger::DAL::Hook qw (DAL_InsertHook DAL_InsertHookFunction);

#========================================================================================
# function     new
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer = DE_EPAGES::MultiStoreCommon::API::Cartridge->new(%options);
#----------------------------------------------------------------------------------------
# description  Creates a new installer object of DE_EPAGES::MultiStoreCommon.
#----------------------------------------------------------------------------------------
# input        %options | options for un/install
#               <ul>
#                 <li>IsRecursive - recursive un/installation of required or dependent
#                     cartridges
#               </ul>
#               | hash
# return       $Installer | cartridge installer | object
#========================================================================================
## no critic qw( ProhibitLongLines )
sub new {
    my ($class, %options) = @_;

    my $self = __PACKAGE__->SUPER::new(
        %options,
        'CartridgeDirectory' => 'DE_EPAGES/MultiStoreCommon',
        'Version'            => '7.93.0',
        'Patches'            => ['6.15.1','6.15.2','6.15.3','6.15.4','6.16.0','6.16.1','6.16.2','6.16.3','6.17.0','6.17.1','6.17.2','6.17.3','6.17.4','6.17.5','6.17.6','6.17.7','6.17.8','6.17.9','6.17.9.1','6.17.9.2','6.17.9.3','6.17.10','6.17.11','6.17.12','6.17.12.1','6.17.12.2','6.17.13','6.17.13.1','6.17.13.2','6.17.14','6.17.14.1','6.17.15','6.17.15.1','6.17.16','6.17.17','6.17.17.1','6.17.18','6.17.19','6.17.20','6.17.20.1','6.17.21','6.17.21.1','6.17.22','6.17.23','6.17.24','6.17.24.1','6.17.25','6.17.26','6.17.26.1','6.17.27','6.17.27.1','6.17.28','6.17.28.1','6.17.29','6.17.29.1','6.17.30','6.17.31','6.17.32','6.17.33','6.17.33.1','6.17.34','6.17.34.1','6.17.35','6.17.36','6.17.36.1','6.17.37','6.17.38','6.17.38.1','6.17.39','6.17.39.1','6.17.40','6.17.40.1','6.17.41','6.17.41.1','6.17.42','6.17.43','6.17.43.1','6.17.44','6.17.45','6.17.45.1','6.17.46','6.17.47','6.17.47.1','6.17.48','6.17.48.1','6.17.49','6.17.49.1','6.17.50','6.17.51','6.17.51.1','6.17.52','6.17.53','7.0.0','7.1.0','7.2.0','7.3.0','7.4.0','7.5.0','7.6.0','7.7.0','7.8.0','7.9.0','7.10.0','7.11.0','7.12.0','7.13.0','7.14.0','7.15.0','7.16.0','7.17.0','7.18.0','7.19.0','7.20.0','7.21.0','7.22.0','7.23.0','7.24.0','7.25.0','7.26.0','7.27.0','7.28.0','7.29.0','7.30.0','7.31.0','7.32.0','7.33.0','7.34.0','7.35.0','7.36.0','7.37.0','7.38.0','7.39.0','7.40.0','7.41.0','7.42.0','7.43.0','7.44.0','7.45.0','7.46.0','7.47.0','7.48.0','7.49.0','7.50.0','7.51.0','7.52.0','7.53.0','7.54.0','7.55.0','7.56.0','7.57.0','7.58.0','7.59.0','7.60.0','7.61.0','7.62.0','7.63.0','7.64.0','7.65.0','7.66.0','7.67.0','7.68.0','7.69.0','7.70.0','7.71.0','7.72.0','7.73.0','7.74.0','7.75.0','7.76.0','7.77.0','7.78.0','7.79.0','7.80.0','7.81.0','7.82.0','7.83.0','7.84.0','7.85.0','7.86.0','7.87.0','7.88.0','7.89.0','7.90.0','7.91.0','7.92.0'],
    );
    return bless $self, $class;
}
## use critic

#========================================================================================
# function     install
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer->install;
#----------------------------------------------------------------------------------------
# description  Installs the cartridge.
#========================================================================================
sub install {
    my $self = shift;

    $self->startInstall;
    $self->SUPER::install;
    $self->finishInstall;
    return;
}

#========================================================================================
# function     uninstall
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer->uninstall;
#----------------------------------------------------------------------------------------
# description  Uninstalls the cartridge.
#========================================================================================
sub uninstall {
    my $self = shift;

    $self->startUninstall;
    $self->SUPER::uninstall;
    $self->finishUninstall;
    return;
}

sub patch_6_15_4 {
    my $self = shift;

    my $dbi = $self->getCurrentDBHandle();
    if($dbi->driverName() ne 'mysql'){
        die 'MultiStore is supported only for MySQL.';
    }

    $self->importPatchDir('Patch_6_15_4', 'Patch');
    $self->logInfo('-> Reset shop attribute MS_LastUpdate.');

    foreach my $Shop (@{ LoadRootObject()->child('Shops')->children }){
        $Shop->set({ 'MS_LastUpdate' => { $Shop->alias => 0 } });
    }

    return '6.16.0';
}

#========================================================================================
# function     patchDBStructure_6_16_0
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer->patchDBStructure_6_16_0;
#----------------------------------------------------------------------------------------
# description  Patches cartridge db structure in active store (database connection)
#               from version '6.16.1' to version '6.16.1'.
#----------------------------------------------------------------------------------------
# return       $Version | version of cartridge after patch function, '6.16.1' | string
#========================================================================================
sub patchDBStructure_6_16_0 {
    my $self = shift;

    # Fixes the patch.
    # Webservice cartridge imports del_AttributeWebservice.xml
    # -> triggers hook OBJ_AfterAttributeUpdate
    # -> used in MultiStoreMaster/MultiStore cartridge
    # -> call GetDefaultValues in MultiStoreCommon cartridge
    # -> MultiStoreCommon triggers hook "API_MS_GetDefaultValues".
    GetLog()->info('Import hook "API_MS_GetDefaultValues".');
    my $HookID = DAL_InsertHook({'Name' => 'API_MS_GetDefaultValues'});
    if (defined $HookID) {
        GetLog()->info('Import hookfunction "OnGetDefaultValues" for "API_MS_GetDefaultValues".');
        my $HookFunctionID = DAL_InsertHookFunction({
            'HookID' => $HookID,
            'OrderNo' => 10,
            'FunctionName' => 'DE_EPAGES::MultiStoreCommon::Hooks::DefaultValues::OnGetDefaultValues'
        });
        die 'HookFunction "OnGetDefaultValues" for hook "API_MS_GetDefaultValues" is not imported.' unless defined $HookFunctionID;
    }
    else {
        die 'Hook "API_MS_GetDefaultValues" is not imported.';
    }

    return '6.16.1';
}

#========================================================================================
# function     patchDBStructure_6_17_1
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer->patchDBStructure_6_17_1;
#----------------------------------------------------------------------------------------
# description  Patches cartridge db structure in active store (database connection)
#               from version '6.17.1' to version '6.17.2'.
#----------------------------------------------------------------------------------------
# return       $Version | version of cartridge after patch function, '6.17.2' | string
#========================================================================================
sub patchDBStructure_6_17_1 {
    my $self = shift;

    my $dbi = $self->getCurrentDBHandle();

    if($dbi->driverName() eq 'mysql'){

        #changes for file /Database/mysql/Tables/objectupdate.sql
        unless ($dbi->existsTableColumn('objectupdate', 'value')) {
            $self->logInfo('add column objectupdate.value');
            $dbi->addTableColumnAfterColumn('objectupdate', 'updatetime', 'value longtext null');
        }

    }

    return '6.17.2';
}

#========================================================================================
# function     patchDBStructure_6_17_4
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer->patchDBStructure_6_17_4;
#----------------------------------------------------------------------------------------
# description  Patches cartridge db structure in active store (database connection)
#               from version '6.17.4' to version '6.17.5'.
#----------------------------------------------------------------------------------------
# return       $Version | version of cartridge after patch function, '6.17.5' | string
#========================================================================================
sub patchDBStructure_6_17_4 {
    my $self = shift;
    #dont use API function, to avoid inconistencies between perl modules and database content

    my $dbi = $self->getCurrentDBHandle();

    if($dbi->driverName() eq 'mysql'){

        #changes for file /Database/mysql/Tables/objectupdate.sql
        $self->logInfo('recreate index objectupdate.i_objectpath');
        $dbi->recreateIndex('objectupdate', 'i_objectpath', undef,
            ['objectpath(255)'],
            { 'objectpath(255)' => 'ASC' }
        );

    }

    return '6.17.5';
}

1;
