#========================================================================================
# package      DE_EPAGES::Object::API::Object::SiteLocale
# state        public
# base         DE_EPAGES::Object::API::Object::Object
#----------------------------------------------------------------------------------------
# description  object interface for SiteLocale.
#               Provides information about regional settings.
#               Includes functions for locale-specific number and currency formats.
#========================================================================================
package DE_EPAGES::Object::API::Object::SiteLocale;
use base DE_EPAGES::Object::API::Object::Object;

use strict;
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::Core::API::Error qw( Error ExistsError GetError );
use DE_EPAGES::Core::API::File qw ( ExistsFile RealPath );
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Core::API::Object::NumberFormat;
use DE_EPAGES::Core::API::Object::DateTimeFormatter;
use DE_EPAGES::Core::API::PerlTools qw (GetLocalTimeZone SubtractDuration trim);
use DE_EPAGES::Core::API::TestParameter qw( TestString );
use DE_EPAGES::Dictionary::API::XML::DictionaryImportHandler qw ( ImportDictionary );
use DE_EPAGES::Object::API::Factory qw ( LoadRootObject );
use DE_EPAGES::Object::API::Table::Locale   qw ( GetInfoLocale );
use DE_EPAGES::Object::API::Table::Currency qw ( GetInfoCurrency );

#========================================================================================
# function     className
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClassName = $SiteLocale->className;
#----------------------------------------------------------------------------------------
# description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $ClassName | class name | String
#========================================================================================
sub className  { 'SiteLocale' }

#========================================================================================
# function     dbPackage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DbPackage = $SiteLocale->dbPackage;
#----------------------------------------------------------------------------------------
# description  Returns the database package, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $DbPackage | database package name | String
#========================================================================================
sub dbPackage { __PACKAGE__ }

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SiteLocale = DE_EPAGES::Object::API::Object::SiteLocale->insert($hInfo);
# example      my $Class = LoadClassByAlias('SiteLocale');
#               my $SiteLocale = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertSiteLocale' is triggered.
#----------------------------------------------------------------------------------------
# input        $hInfo | attributes of new object | ref.hash
# return       $SiteLocale | new object | object
# hook         OBJ_InsertSiteLocale | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>SiteLocaleID  - site locale id       - int
#               </ul>
#========================================================================================
sub insert {
    my $class = shift;
    my ($hInfo) = @_;

    my $self = $class->SUPER::insert($hInfo);

    TriggerHook('OBJ_InsertSiteLocale', {'Object' => $self, 'SiteLocaleID' => $self->id });
    delete $hInfo->{$_} foreach qw ( LocaleID ); ## no critic qw(ProhibitLongLines)
    return $self;
}

#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SiteLocale->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteSiteLocale' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteSiteLocale | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>SiteLocaleID  - site locale id       - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeleteSiteLocale', {'Object' => $self, 'SiteLocaleID' => $self->id });
    $self->SUPER::delete;
    return;
}

#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SiteLocale->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateSiteLocale' and 'OBJ_AfterUpdateSiteLocale'.
#----------------------------------------------------------------------------------------
# input        $hValues | attribute name => value hash | ref.hash
# input        $LanguageID | (optional; required for localized attributes) language id | int
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateSiteLocale | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>SiteLocaleID  - site locale id       - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateSiteLocale | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>SiteLocaleID  - site locale id       - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    TriggerHook('OBJ_BeforeUpdateSiteLocale', {'Object'=>$self, 'Values'=> $hValues, 'SiteLocaleID' => $self->id, 'LanguageID' => $LanguageID});
    $self->SUPER::set($hValues, $LanguageID);
    TriggerHook('OBJ_AfterUpdateSiteLocale',  {'Object'=>$self, 'Values'=> $hValues, 'SiteLocaleID' => $self->id, 'LanguageID' => $LanguageID});
    return;
}

#========================================================================================
# function     localeID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Type = $SiteLocale->localeID;
#----------------------------------------------------------------------------------------
# description  Returns the default locale.
#               This is an optimized version of $SiteLocale->get('LocaleID').
#----------------------------------------------------------------------------------------
# return       $Type | attribute type | string
#========================================================================================
sub localeID {  ## no critic qw(RequireArgUnpacking RequireCommentForExportedFunctions)
    return $_[0]->{'Attributes'}{'LocaleID'};
}

#========================================================================================
# function     formatAttributes
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hFormatAttributes = $Object->formatAttributes($CurrencyID);
#----------------------------------------------------------------------------------------
# description  Returns the format options for number formatter.
#----------------------------------------------------------------------------------------
# input        $CurrencyID | currency identifier | char(3)
# return       $hFormatAttributes | options of formatter, keys:
#               <ul>
#                   <li>SymbolCurrency
#                   <li>PosFormatCurrency
#                   <li>NegFormatCurrency
#                   <li>MinorUnitCurrency
#                   <li>DecimalSep
#                   <li>ThousandSep
#                   <li>NegFormatNumbers
#                   <li>NegFormatCurrency
#                   <li>PosFormatCurrency
#               </ul> | ref.hash
#========================================================================================
sub formatAttributes {
    my $self = shift;
    my ($CurrencyID) = @_;

    my $Locale = GetInfoLocale($self->localeID);
    $CurrencyID = $Locale->{'CurrencyID'} unless defined $CurrencyID;
    my $Currency = GetInfoCurrency($CurrencyID);

    my %Values;
    $Values{$_.'Currency'} = $Currency->{$_} foreach qw (Symbol PosFormat NegFormat MinorUnit);
    $Values{$_} = $Locale->{$_} foreach grep { defined $Locale->{$_} } qw (DecimalSep ThousandSep NegFormatNumbers NegFormatCurrency PosFormatCurrency);
    return \%Values;
}

#========================================================================================
# function     formatInteger
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $Formatted = $Object->formatInteger($Number);
#----------------------------------------------------------------------------------------
# description  Returns the formatted number (insert thousand sep).
#----------------------------------------------------------------------------------------
# input        $Number | number | int
# return       $Formatted | formatted value | string
#========================================================================================
sub formatInteger {
    my $self = shift;
    my ($Number) = @_;
    my $FormatAttributes = $self->formatAttributes();
    my $Formatter = DE_EPAGES::Core::API::Object::NumberFormat->new(
        decimal_point  => $FormatAttributes->{'DecimalSep'},
        thousands_sep  => $FormatAttributes->{'ThousandSep'},
        neg_format     => $FormatAttributes->{'NegFormatNumbers'},
        decimal_digits => 0,
    );
    return $Formatter->format_number($Number);
}

#========================================================================================
# function     formatFloat
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Formatted = $Object->formatFloat($Number);
# syntax       my $Formatted = $Object->formatFloat($Number, $Precision);
# example      print $Object->formatFloat( 12345.6789, 3 );
#----------------------------------------------------------------------------------------
# description  Returns the formatted number (insert thousand sep and decimal point).
#----------------------------------------------------------------------------------------
# input        $Number | number | float
# input        $Precision | (optional) maximum number of digits right to the decimal
#                       point | float
# return       $Formatted | formatted value | string
#========================================================================================
sub formatFloat {
    my $self = shift;
    my ($Number, $Precision) = @_;
    my $FormatAttributes = $self->formatAttributes();
    my $Formatter = DE_EPAGES::Core::API::Object::NumberFormat->new(
        decimal_point  => $FormatAttributes->{'DecimalSep'},
        thousands_sep  => $FormatAttributes->{'ThousandSep'},
        neg_format     => $FormatAttributes->{'NegFormatNumbers'},
    );
    return $Formatter->format_number($Number, $Precision);
}

#========================================================================================
# function     formatDateTime
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Formatted = $Object->formatDateTime($DateTime, $Type);
# syntax       $Formatted = $Object->formatDateTime($DateTime, $Type, $Pattern);
# example      print $Object->formatDateTime( DateTime->now(), 'date' );
#----------------------------------------------------------------------------------------
# description  Returns the formatted date and time string in the Store time zone.
#----------------------------------------------------------------------------------------
# input        $DateTime | DateTime object | object
# input        $Type | date/time/datetime/date_tz/time_tz/datetime_utc | string
# input        $Pattern | (optional) short/medium/long/full/mnemonic
#                               or custom strftime format, for example '%Y-%m-%d'
#                               (default: short)| string
# return       $Formatted | formatted value | string
#========================================================================================
sub formatDateTime {
    my $self = shift;
    my ($DateTime, $Type, $Pattern) = @_;
    TestString('Type', $Type);
    if (not ref $DateTime) {
        GetLog('BUGTRACKER::EPG-7351')->warn("Can't format datetime string $DateTime.");
        return $DateTime;
    }
    my $dt = $DateTime->clone;
    my $TimeZone = LoadRootObject()->get('TimeZone');
    die 'no store time zone' if not defined $TimeZone;

    my $LocaleID = $self->localeID;

    # convert from server time to the timezone of the database
    eval {
        $dt->set_time_zone( GetLocalTimeZone() );

        # don't change date values, for example Birthdate and Time
        $dt->set_time_zone( $TimeZone) if ($Type eq 'datetime') or ($Type eq 'date_tz') or ($Type eq 'time_tz');
        $dt->set_time_zone( 'UTC' ) if $Type eq 'datetime_utc';
        $dt->set( locale => $LocaleID );
    };
    if( ExistsError() ) {
        my $Error = GetError();
        if( $Error->code eq 'NativeDie' && $Error->shortMessage =~ /^Invalid local time for date/ ) {
            # EPG-12193: workaround for orders taken during non-existing time intervals
            warn 'Setting timezone to UTC for invalid datetime: '.$dt->iso8601;
            $dt->set_time_zone('UTC');
        }
        else {
            Error( $Error );
        }
    }

    $Type =~ s/_(?:tz|utc)//g;

    my $Formatter = DE_EPAGES::Core::API::Object::DateTimeFormatter->new( locale => $LocaleID );
    if ($Type ne 'time' && $Pattern && $Pattern eq 'mnemonic') {

        my $ymd = $dt->ymd;
        my $CurrentDate = LoadRootObject()->currentDate; # current date in system timezone

        my $formatString;
        if( $ymd eq $CurrentDate->ymd ) {
            $formatString = $self->_localizedDate('today');
        }
        elsif( $ymd eq SubtractDuration($CurrentDate, 'days', 1)->ymd ) {
            $formatString = $self->_localizedDate('yesterday');
        }
        else {
            $formatString = $Formatter->formatString('date', 'short');
            #remove year
            $formatString =~ s/%\{ce_year\}//g;
            $formatString =~ s/^[\/\.:]//g;
            $formatString =~ s/[\/\.:]$//g;
        }

        if ($Type eq 'datetime') {
            $formatString .= ', ' . $Formatter->formatString('time', 'short');
        }
        return $Formatter->format_datetime($dt, $Type, $formatString);
    } else {
        return $Formatter->format_datetime($dt, $Type, $Pattern);
    }
}

sub _localizedDate {
    my $self = shift;
    my ($Key) = @_;
    my $LanguageCode = $self->get('Language')->{'Code2'};
    my $FileName = RealPath($ENV{'EPAGES_CARTRIDGES'})."/DE_EPAGES/Dictionary/Templates/DatetimeFormatter.$LanguageCode.xml";
    my $Dictionary = {};
    ImportDictionary($FileName, $Dictionary) if ExistsFile($FileName);;
    return $Dictionary->{$LanguageCode}{$Key} || '{'.$Key.'}';
}

#========================================================================================
# function     formatMoney
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Formatted = $Object->formatMoney($Number, $CurrencyID);
# syntax       $Formatted = $Object->formatMoney($Number, $CurrencyID, $Precision);
#----------------------------------------------------------------------------------------
# description  Returns the formatted number (insert thousand sep and decimal point).
#----------------------------------------------------------------------------------------
# input        $Number | number | float
# input        $CurrencyID | currency identifier | char(3)
# input        $Precision | (optional) number of digits after the decimal point | integer
# return       $Formatted | formatted value | string
#========================================================================================
sub formatMoney {
    my $self = shift;
    my ($Number, $CurrencyID, $Precision) = @_;
    my $FormatAttributes = $self->formatAttributes($CurrencyID);

    my $Formatter = DE_EPAGES::Core::API::Object::NumberFormat->new(
        int_currencyid     => $CurrencyID,
        curr_neg_format    => $FormatAttributes->{'NegFormatCurrency'},
        curr_pos_format    => $FormatAttributes->{'PosFormatCurrency'},
        decimal_fill       => 1,
        decimal_point      => $FormatAttributes->{'DecimalSep'},
        decimal_digits     => $FormatAttributes->{'MinorUnitCurrency'},
        int_curr_symbol    => $FormatAttributes->{'SymbolCurrency'},
        mon_decimal_point  => $FormatAttributes->{'DecimalSep'},
        mon_thousands_sep  => $FormatAttributes->{'ThousandSep'},
        thousands_sep      => $FormatAttributes->{'ThousandSep'},
    );
    return $Formatter->format_price($Number, $Precision);
}

#========================================================================================
# function     formatRoundMoney
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Formatted = $Object->formatRoundMoney($Number, $CurrencyID);
# syntax       $Formatted = $Object->formatRoundMoney($Number, $CurrencyID, $Precision);
#----------------------------------------------------------------------------------------
# description  Returns the rounded and formatted number (insert thousand sep and decimal point).
#----------------------------------------------------------------------------------------
# input        $Number | number | float
# input        $CurrencyID | currency identifier | char(3)
# input        $Precision | (optional) number of digits after the decimal point | integer
# return       $Formatted | rounded and formatted value | string
#========================================================================================
sub formatRoundMoney {
    my $self = shift;
    my ($Number, $CurrencyID, $Precision) = @_;
    my $FormatAttributes = $self->formatAttributes($CurrencyID);

    my $Formatter = DE_EPAGES::Core::API::Object::NumberFormat->new(
        int_currencyid     => $CurrencyID,
        curr_neg_format    => $FormatAttributes->{'NegFormatCurrency'},
        curr_pos_format    => $FormatAttributes->{'PosFormatCurrency'},
        decimal_fill       => 1,
        decimal_point      => $FormatAttributes->{'DecimalSep'},
        decimal_digits     => $FormatAttributes->{'MinorUnitCurrency'},
        int_curr_symbol    => $FormatAttributes->{'SymbolCurrency'},
        mon_decimal_point  => $FormatAttributes->{'DecimalSep'},
        mon_thousands_sep  => $FormatAttributes->{'ThousandSep'},
        thousands_sep      => $FormatAttributes->{'ThousandSep'},
    );

    my $RoundNumber =  $Formatter->round($Number, $FormatAttributes->{'MinorUnitCurrency'});

    return $Formatter->format_price($RoundNumber, $Precision);
}

#========================================================================================
# function     formatMoneyPlain
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Formatted = $Object->formatMoneyPlain($Number, );
# syntax       $Formatted = $Object->formatMoneyPlain($Number, $Precision);
#----------------------------------------------------------------------------------------
# description  Returns the formatted number (insert thousand sep and decimal point)
#               without currency symbol
#----------------------------------------------------------------------------------------
# input        $Number | number | float
# input        $Precision | (optional) number of digits after the decimal point | integer
# return       $Formatted | formatted value | string
#========================================================================================
sub formatMoneyPlain {
    my $self = shift;
    my ($Number, $Precision) = @_;

    my $FormatAttributes = $self->formatAttributes();
    my $Formatter = DE_EPAGES::Core::API::Object::NumberFormat->new(
        decimal_fill       => 1,
        decimal_point      => $FormatAttributes->{'DecimalSep'},
        decimal_digits     => $FormatAttributes->{'MinorUnitCurrency'},
        mon_decimal_point  => $FormatAttributes->{'DecimalSep'},
        mon_thousands_sep  => $FormatAttributes->{'ThousandSep'},
        thousands_sep      => $FormatAttributes->{'ThousandSep'},
        neg_format         => $FormatAttributes->{'NegFormatNumbers'},
    );

    my ($Price) = trim($Formatter->format_price($Number, $Precision));
    return $Price;
}

#========================================================================================
# function     formatRoundMoneyPlain
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Formatted = $Object->formatRoundMoneyPlain($Number, );
# syntax       $Formatted = $Object->formatRoundMoneyPlain($Number, $CurrencyID, $Precision);
#----------------------------------------------------------------------------------------
# description  Returns the rounded and formatted number (insert thousand sep and decimal point)
#               without currency symbol
#----------------------------------------------------------------------------------------
# input        $Number | number | float
# input        $CurrencyID | currency identifier | char(3)
# input        $Precision | (optional) number of digits after the decimal point | integer
# return       $Formatted | rounded and formatted value | string
#========================================================================================
sub formatRoundMoneyPlain {
    my $self = shift;
    my ($Number, $CurrencyID, $Precision) = @_;

    my $FormatAttributes = $self->formatAttributes();
    my $Formatter = DE_EPAGES::Core::API::Object::NumberFormat->new(
        int_currencyid     => $CurrencyID,
        curr_neg_format    => $FormatAttributes->{'NegFormatCurrency'},
        curr_pos_format    => $FormatAttributes->{'PosFormatCurrency'},
        decimal_fill       => 1,
        decimal_point      => $FormatAttributes->{'DecimalSep'},
        decimal_digits     => $FormatAttributes->{'MinorUnitCurrency'},
        int_curr_symbol    => $FormatAttributes->{'SymbolCurrency'},
        mon_decimal_point  => $FormatAttributes->{'DecimalSep'},
        mon_thousands_sep  => $FormatAttributes->{'ThousandSep'},
        thousands_sep      => $FormatAttributes->{'ThousandSep'},
    );

    my $RoundNumber =  $Formatter->round($Number, $FormatAttributes->{'MinorUnitCurrency'});
    return $self->formatMoneyPlain($RoundNumber, $Precision );
}

#========================================================================================
# function     formatMoneyMicrodata
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Formatted = $Object->formatMoneyMicrodata($Number, $CurrencyID);
# syntax       $Formatted = $Object->formatMoneyMicrodata($Number, $CurrencyID, $Precision);
#----------------------------------------------------------------------------------------
# description  Returns the formatted number within a microdata span tag.
#----------------------------------------------------------------------------------------
# input        $Number | number | float
# input        $CurrencyID | currency identifier | char(3)
# input        $Precision | (optional) number of digits after the decimal point | integer
# return       $Formatted | formatted value | string
#========================================================================================
sub formatMoneyMicrodata {
    my $self = shift;
    my ($Number, $CurrencyID, $Precision) = @_;
    my $FormatAttributes = $self->formatAttributes($CurrencyID);

    my $Formatter = DE_EPAGES::Core::API::Object::NumberFormat->new(
        int_currencyid     => $CurrencyID,
        curr_neg_format    => $FormatAttributes->{'NegFormatCurrency'},
        curr_pos_format    => $FormatAttributes->{'PosFormatCurrency'},
        decimal_fill       => 1,
        decimal_point      => $FormatAttributes->{'DecimalSep'},
        decimal_digits     => $FormatAttributes->{'MinorUnitCurrency'},
        int_curr_symbol    => $FormatAttributes->{'SymbolCurrency'},
        mon_decimal_point  => $FormatAttributes->{'DecimalSep'},
        mon_thousands_sep  => $FormatAttributes->{'ThousandSep'},
        thousands_sep      => $FormatAttributes->{'ThousandSep'},
    );
    return $Formatter->format_price($Number, $Precision, 'microdata');
}

#========================================================================================
# function     roundMoney
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Rounded = $Object->roundMoney($Number, $CurrencyID);
#----------------------------------------------------------------------------------------
# description  Round number for money (<function formatAttributes> includes MinorUnitCurrency).
#----------------------------------------------------------------------------------------
# input        $Number | number in perl | float
# input        $CurrencyID | currency identifier | char(3)
# input        $RoundMode | (optional) round mode, see Math::BigFloat->round_mode.
#                   Default: standard commercial practice (Kaufmnnisches Runden http://de.wikipedia.org/wiki/Rundung).
#                   Possible values are: even|odd|+inf|-inf|zero|trunc.|ceil | string
# return       $Rounded | rounded number | string
#========================================================================================
sub roundMoney {
    my $self = shift;
    my ($Number, $CurrencyID, $RoundMode ) = @_;
    my $FormatAttributes = $self->formatAttributes($CurrencyID);

    my $Formatter = DE_EPAGES::Core::API::Object::NumberFormat->new();
    return $Formatter->round($Number, $FormatAttributes->{'MinorUnitCurrency'}, $RoundMode);
}

#========================================================================================
# function     format
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Formatted = $Object->format($Value, $Format, $CurrencyID);
#----------------------------------------------------------------------------------------
# description  Formats numbers in float/integer/money and DateTime objects to date/datetime/timedate_tz/time_tz/datetime_utc
#               strings.
#----------------------------------------------------------------------------------------
# input        $Value | number in perl or DateTime object| float/object
# input        $Format | (integer/float/money/date)/datetime/time/date_tz/time_tz/datetime_utc) | string
# input        $CurrencyID | currency identifier (for money) | char(3)
# return       $Formatted | formatted value | string
#========================================================================================
sub format {
    my $self = shift;
    my ($Value, $Format, @Parameter) = @_;
    if ($Format eq 'integer') {
        return $self->formatInteger($Value, @Parameter);
    } elsif ($Format eq 'float') {
        return $self->formatFloat($Value, @Parameter);
    } elsif ($Format eq 'money') {
        return $self->formatMoney($Value, @Parameter);
    } elsif ($Format eq 'roundmoney') {
        return $self->formatRoundMoney($Value, @Parameter);
    } elsif ($Format eq 'roundmoney_plain') {
        return $self->formatRoundMoneyPlain($Value, @Parameter);
    } elsif ($Format eq 'money_plain') {
        return $self->formatMoneyPlain($Value, @Parameter);
    } elsif ($Format eq 'money_microdata') {
        return $self->formatMoneyMicrodata($Value, @Parameter);
    } elsif ( $Format eq 'datetime' or $Format eq 'date' or $Format eq 'time' or $Format eq 'date_tz' or $Format eq 'time_tz' or $Format eq 'datetime_utc') {
        return $self->formatDateTime($Value, $Format, @Parameter);
    }
    die 'unknown format';
}

#========================================================================================
# function     parse
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Value = $Object->parse($Formatted, $CurrencyID);
#----------------------------------------------------------------------------------------
# description  Parse a formatted numbers of float/integer/money.
#----------------------------------------------------------------------------------------
# input        $Formatted | formatted value | string
# input        $CurrencyID | currency identifier (for money) | char(3)
# return       $Value | number in perl or DateTime object| float
#========================================================================================
sub parse {
    my $self = shift;
    my ($String, $CurrencyID) = @_;

    my $FormatAttributes = $self->formatAttributes($CurrencyID);

    my $Formatter = DE_EPAGES::Core::API::Object::NumberFormat->new(
        int_currencyid     => $CurrencyID,
        curr_neg_format    => $FormatAttributes->{'NegFormatCurrency'},
        curr_pos_format    => $FormatAttributes->{'PosFormatCurrency'},
        decimal_fill       => 1,
        decimal_point      => $FormatAttributes->{'DecimalSep'},
        decimal_digits     => $FormatAttributes->{'MinorUnitCurrency'},
        int_curr_symbol    => $FormatAttributes->{'SymbolCurrency'},
        mon_decimal_point  => $FormatAttributes->{'DecimalSep'},
        mon_thousands_sep  => $FormatAttributes->{'ThousandSep'},
        neg_format         => $FormatAttributes->{'NegFormatNumbers'},
        thousands_sep      => $FormatAttributes->{'ThousandSep'},
    );
    return $Formatter->unformat_number($String);
}

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::Object::API::Object::Object->default( $Parent, $hInfo );
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;
    return {
        %{$class->SUPER::default($Parent, $hInfo)},
        'LocaleID' => $hInfo->{'Alias'},
        'Site'=> $Parent->getSite(),
    };
}

#========================================================================================
# function     parseDateTime
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DateTime = $Object->parseDateTime($Formatted, $Type, $Pattern);
#----------------------------------------------------------------------------------------
# description  Parse a formatted date and time strings. The current Store time zone is
#               assumed.
#----------------------------------------------------------------------------------------
# input        $Formatted | formatted value | string
# input        $Type | date/time/datetime | string
# input        $Pattern | (optional) short/medium/long/full
#                               or custom strftime format, for example '%Y-%m-%d'
#                               (default: short)| string
# return       $DateTime | DateTime object | object
#========================================================================================
sub parseDateTime {
    my $self = shift;
    my ($String, $Type, $Pattern) = @_;

    my $Formatter = DE_EPAGES::Core::API::Object::DateTimeFormatter->new(
        locale => $self->localeID,
    );

    my $CleanType = $Type;
    $CleanType =~ s/_(?:tz|utc)//g;
    my $dt = $Formatter->parse_datetime($String, $CleanType, $Pattern);

    my $TimeZone = LoadRootObject()->get('TimeZone');
    die 'no store time zone' if not defined $TimeZone;

    $dt->set_time_zone( 'floating');

    # don't change date values, for example Birthdate or Time
    $dt->set_time_zone( $TimeZone) if ($Type eq 'datetime') or ($Type eq 'date_tz') or ($Type eq 'time_tz');
    $dt->set_time_zone( GetLocalTimeZone());

    return $dt;
}

1;
