#==============================================================================
# package      DE_EPAGES::Object::API::TLE::SiteFormatter
# state        public
# description  Implements locale-dependent TLE formats that depend on the site of the
#               current object and the selected locale.
#----------------------------------------------------------------------------------------
# example      #ListPrice[money]          -> 72,95 
#               #Price[money:USD]          -> $72.95
#               #ListPrice[money_microdata]-> <span itemprop="price" content="459.00">459.00</span>
#                                             <span itemprop="priceCurrency" content="GBP"></span>
#               #Position[integer]         -> 1.020
#               #Weight[float]             -> 12,5
#               #ExampleFloat              -> 12345.6789
#               #ExampleFloat[float]       -> 12.345,68
#               #SmallExampleFloat         -> 12.34
#               #SmallExampleFloat[float]  -> 12,34
#               #Now[date]                 -> 01.12.2004
#               #Now[date_tz]              -> 01.12.2004
#               #Now[time]                 -> 11:31
#               #Now[time_tz]              -> 11:31
#               #Now[datetime]             -> 01.12.2004 11:31
#               #Now[datetime_utc]         -> 01.12.2004 11:31 UTC +0000
#               #Now[datetime:short]       -> 01.12.2004 11:31
#               #Now[datetime:medium]      -> 01.12.2004 11:31:01
#               #Now[datetime:long]        -> 01. Dezember 2004 11:31:01 +0100
#==============================================================================
package DE_EPAGES::Object::API::TLE::SiteFormatter;

use strict;
use DE_EPAGES::Core::API::Error qw( ExistsError GetError );
use DE_EPAGES::Core::API::PerlTools qw( AddDuration );
use DE_EPAGES::Core::API::TestParameter qw( TestInteger );
use DE_EPAGES::Core::API::Warning qw( Warning );
use DE_EPAGES::Object::API::Factory qw (LoadObject LoadRootObject);

#========================================================================================
# function     RegisterHandlerProc
# state        private
#----------------------------------------------------------------------------------------
# syntax       $RegisterHandlerProc( $hParams );
#----------------------------------------------------------------------------------------
# description  this function register this attribute handle on the processor via hook
#               function on hook TLEProcessorRegistration
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments keys :
#               <ul>
#                   <li>Processor - tle processor - object
#               </ul> | ref.hash
#========================================================================================
sub RegisterHandlerProc {
    my ($Params) = @_;
    my $TLEProcessor = $Params->{'Processor'};
    my $Handler = __PACKAGE__->new();
    $Handler->register($TLEProcessor);
}

#========================================================================================
# function     new
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler = new DE_EPAGES::TLE::API::SpaceFormatter;
# example      my $SpaceFormatter  = new DE_EPAGES::TLE::API::SpaceFormatter;
#----------------------------------------------------------------------------------------
# description  create a new handler
#----------------------------------------------------------------------------------------
# return       $Handler | blessed object | object
#========================================================================================
sub new { return bless {}, shift; }

#========================================================================================
# function     register
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler->register( $Processor );
#----------------------------------------------------------------------------------------
# description  this function register this handle on the processor
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
#========================================================================================
sub register {
    my $self = shift;
    my ($Processor) = @_;

    $Processor->registerHandler('FormatHandler', $self, 'integer');
    $Processor->registerHandler('FormatHandler', $self, 'float');
    $Processor->registerHandler('FormatHandler', $self, 'money');
    $Processor->registerHandler('FormatHandler', $self, 'roundmoney');
    $Processor->registerHandler('FormatHandler', $self, 'roundmoney_plain');
    $Processor->registerHandler('FormatHandler', $self, 'money_plain');
    $Processor->registerHandler('FormatHandler', $self, 'date');
    $Processor->registerHandler('FormatHandler', $self, 'date_tz');
    $Processor->registerHandler('FormatHandler', $self, 'datetime');
    $Processor->registerHandler('FormatHandler', $self, 'time');
    $Processor->registerHandler('FormatHandler', $self, 'time_tz');
    $Processor->registerHandler('FormatHandler', $self, 'datetime_utc');
    $Processor->registerHandler('FormatHandler', $self, 'money_microdata');

    $Processor->registerHandler('VariableHandler', $self, 'ExampleFloat');
    $Processor->registerHandler('VariableHandler', $self, 'SmallExampleFloat');

    $Processor->registerHandler('FunctionHandler', $self, 'SITELOCALE');
    $Processor->registerHandler('FunctionHandler', $self, 'ADD_DURATION');
    return;
}

# possible formats: #Price[money] #Price[roundmoney] #Price[money:USD:en_US:4] #Factor[float:4]
sub Format {
    my $self = shift;
    my ($Processor, $Format, $Value, $TLEName, $raParameter) = @_;
    return undef unless defined $Value;
    return $Value if $Value eq '';

    my $CurrencyID = undef;
    $CurrencyID = shift @$raParameter if $Format eq 'money';
    $CurrencyID = shift @$raParameter if $Format eq 'roundmoney';
    $CurrencyID = shift @$raParameter if $Format eq 'roundmoney_plain';

    my $LocaleID = defined $raParameter->[0] && $raParameter->[0] =~ /^\w{2}_\w{2}$/ ? shift @$raParameter : undef;
    $CurrencyID = undef if (defined $CurrencyID && $CurrencyID !~ /\S/);
    $LocaleID   = undef if (defined $LocaleID   && $LocaleID   !~ /\S/);

    my $ObjectID = $Processor->tle('INPUT.ObjectID');
    my $Object = defined $ObjectID ? LoadObject($ObjectID) : $Processor->tle('Object');
    die 'no object found' unless defined $Object;
    my $Site = $Object->getSite();

    $LocaleID = $Processor->tle('Locale') unless defined $LocaleID;
    $LocaleID = $Processor->tle('INPUT.Locale') if !defined $LocaleID || ref $LocaleID;
    $LocaleID = 'en_US' unless defined $LocaleID;

    if (not defined $LocaleID) {
        warn "tle value $TLEName was not formatted, no locale found.";
        return $Value;
    }

    my $Formatter = $Site->siteLocale( $LocaleID );
    my $FormattedValue = eval {
        if ($Format eq 'money' || $Format eq 'roundmoney' || $Format eq 'roundmoney_plain' || $Format eq 'money_microdata') {
            $CurrencyID = $Processor->tle('CurrencyID') unless defined $CurrencyID;
            $CurrencyID = $Processor->tle('INPUT.Currency') unless defined $CurrencyID;
            $Formatter->format($Value, $Format, $CurrencyID, @$raParameter);
        }
        else {
            $Formatter->format($Value, $Format, @$raParameter);
        }
    };
    if( ExistsError() ) {
        # EPG-20394: if the error "round() overflow" was not catched, the application
        # server would get into an undefined state
        Warning( 'FormatError', {
            Value => $Value,
            Message => GetError()->shortMessage,
        });
        return $Value;
    }
    else {
        return $FormattedValue;
    }
}

sub tle {
    my $self = shift;
    my ($Processor, $TLEName) = @_;

    return 12.34 if $TLEName eq 'SmallExampleFloat' ; ## no critic qw(ProhibitMagicNumbers)
    return 12345.6789 if $TLEName eq 'ExampleFloat' ; ## no critic qw(ProhibitMagicNumbers)
}

sub existsTLE {
    my $self = shift;
    return 1;
}

#========================================================================================
# function     SITELOCALE
# state        public
#----------------------------------------------------------------------------------------
# syntax       #SITELOCALE(#LocaleID)
# example      #SITELOCALE(#INPUT.Locale)
#----------------------------------------------------------------------------------------
# description  Returns the site locale object, which includes sepators, daynames, ....
#----------------------------------------------------------------------------------------
# input        #LocaleID | locale id | string
#========================================================================================
sub SITELOCALE ($$$) { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $raParams) = @_;

    return undef unless defined $raParams->[0];
    return LoadRootObject()->siteLocale($raParams->[0])->tleHash;
}

#========================================================================================
# function     ADD_DURATION
# state        public
#----------------------------------------------------------------------------------------
# syntax       #ADD_DURATION(#DateTime, #Duration)
# example      #ADD_DURATION(#Now, 1)
# example      #ADD_DURATION(#Now, -1)
#----------------------------------------------------------------------------------------
# description  Adds a duration (in days) to a datetime object and returns the new one.
#----------------------------------------------------------------------------------------
# input        #DateTime | date time | DateTime object
# input        #Duration | duration | integer
#========================================================================================
sub ADD_DURATION ($$$) { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $raParams) = @_;
    return undef unless defined $raParams->[0];
    return $raParams->[0] unless defined $raParams->[1];

    my $DateTime = $raParams->[0];
    my $Duration = $raParams->[1];
    TestInteger('Duration', $Duration);

    return AddDuration($DateTime->clone, 'days', $Duration);
}

1;
