use strict;
use warnings;
use Getopt::Long;
use Time::HiRes qw( time );

use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw( RunOnStore );
use DE_EPAGES::Object::API::Factory qw( LoadRootObject );
use DE_EPAGES::Core::API::File qw ( WriteFile );

sub Main {

    local $| = 1;

    my $Password;
    my $StoreName;
    my $ShopAlias;
    my $ExportYear;
    my $FilePath;
    my $Help;

    GetOptions(
        'help'        => \$Help,
        'passwd=s'    => \$Password,
        'storename=s' => \$StoreName,
        'shop=s'      => \$ShopAlias,
        'year=s'      => \$ExportYear,
        'file=s'      => \$FilePath,
    );

    usage() if $Help;

    unless ( defined $ExportYear ) {
        print STDERR "Missing parameter -year.\n\n";
        usage();
    }

    unless ( defined $FilePath ) {
        print STDERR "Missing parameter -file.\n\n";
        usage();
    }

    if ( (defined $ShopAlias && !defined $StoreName) || (defined $StoreName && !defined $ShopAlias) ) {
        print STDERR "-storename and -shop must be passed together, or not at all.\n\n";
        usage();
    }

    my $time1 = time;

    my $aStores;

    $aStores = [$StoreName] if $StoreName;
    RunOnStore(
        Store      => 'Site',
        DBPassword => undef,
        Sub        => sub {
          my $Root = LoadRootObject();

          $aStores = $Root->get('Stores');
        },
    ) if !defined $StoreName;

    my $Csv = ",Jan,Feb,Mar,Apr,May,Jun,Jul,Aug,Sep,Oct,Nov,Dec";
    foreach my $Store ( @$aStores ) {
      my $StoreAlias = ref \$Store eq 'SCALAR' ? $Store : $Store->alias;
      next if $StoreAlias eq 'Site';

      RunOnStore(
          Store      => $StoreAlias,
          DBPassword => undef,
          Sub        => sub {
            my $Root = LoadRootObject();

            my $aShops;
            if ( defined $ShopAlias ) {
              $aShops = [$Root->child('Shops')->child($ShopAlias)];
            } else {
              $aShops = $Root->get('Shops');
            }

            foreach my $Shop ( @$aShops ) {
              next unless $Shop->featureMaxValue('OpenAITokens') > 0;

              my $hOpenAITokensUsage = $Shop->get('OpenAITokensUsage');

              my $SAlias = $Shop->alias;
              my $ShopName = $Shop->get('NameOrAlias', 2);
              $Csv .= "\n$ShopName ($SAlias)";

              for my $Month ( 1..12 ) {
                my $Key = $ExportYear.$Month;

                my $TotalTokens = 0;
                if ( exists $hOpenAITokensUsage->{$Key} ) {
                  my $InputTokens = $hOpenAITokensUsage->{$Key}->{'InputTokens'} + 0;
                  my $OutputTokens = $hOpenAITokensUsage->{$Key}->{'OutputTokens'} + 0;

                  $TotalTokens = $InputTokens + $OutputTokens;
                }

                $Csv .= ",$TotalTokens";
              }

            }
          },
      );
    }

    WriteFile( $FilePath, \$Csv, { binary => 1, create_directory => 0 } );

    my $time2 = time;
    printf "need %.1f seconds\n", $time2 - $time1;
}

sub usage {
    print <<END_USAGE;
Description:
    Exports token usage data for all months of a specific year across all shops (or for a specific shop)

Usage:
    perl $0 [options] [flags] datafiles
options:
    -passwd     database user password                          ()
    -year       year that you want data exported for            (required)
    -file       file name/path to write to                      (required)
    -storename  name of store                                   (optional, must be passed with -shop)
    -shop       alias of shop                                   (optional, must be passed with -storename)
flags:
    -help       show the command line options
Example:
    perl $0 -year 2024 -file tokenusageexport.csv
END_USAGE
    exit 2;
}

RunScript('Sub' => \&Main);
