#========================================================================================
# package      DE_EPAGES::Order::API::Attributes::LineItemContainer_Ext
# state        private
# description  Attribute interface for LineItemContainer.
#========================================================================================
package DE_EPAGES::Order::API::Attributes::LineItemContainer_Ext;
use base qw( DE_EPAGES::Object::API::BaseAttribute);

use strict;
use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Core::API::TestParameter qw ( TestObject );
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias );
use DE_EPAGES::Object::API::Table::Locale qw ( GetLanguageByLocaleID );
use DE_EPAGES::Shop::API::Address qw ( GetTaxCountryIDForShop IsEUAddress );
use DE_EPAGES::Trigger::API::Trigger qw( TriggerHook );

##no critic qw(ProhibitCascadingIfElse ProhibitExcessComplexity)

sub ErrorDefinition {
    return {
        'NoLineItemPrice' => 'LineItem #Path has no price',
    };
}

sub getAttributes {
    my $self = shift;
    my ($Object, $aNames) = @_;
    my $hInfo = {};
    foreach my $Name (@$aNames) {
        if ($Name eq 'LanguageID') {
            my $LocaleID = $Object->get('LocaleID');
            $hInfo->{$Name} = GetLanguageByLocaleID($LocaleID);
        }
        elsif ($Name eq 'PaymentMethods') {
            $hInfo->{$Name} = $self->_paymentMethods($Object);
        }
        elsif ($Name eq 'ShippingMethods') {
            $hInfo->{$Name} = $self->_shippingMethods($Object);
        }
        elsif ($Name eq 'ShopID') {
            $hInfo->{$Name} = $Object->getSite()->id;
        }
        elsif ($Name eq 'IsTaxClassRemoved') {
            $hInfo->{$Name} = $self->_GetIsTaxClassRemovedFromLineItems($Object);
        }
        elsif ($Name eq 'IsTaxClassParentBackup') {
            $hInfo->{$Name} = $self->_GetIsTaxClassParentBackup($Object);
        }
        elsif ($Name eq 'LineItemsSum') {
            $hInfo->{$Name} = $self->_getItemsSum($Object, 'LineItems');
        }
        elsif ($Name eq 'LineItemStaticsSum') {
            $hInfo->{$Name} = $self->_getItemsSum($Object, 'Statics');
        }
        elsif ($Name eq 'SalesDiscountsSum') {
            $hInfo->{$Name} = $self->_getItemsSum($Object, 'SalesDiscounts');
        }
        elsif ($Name eq 'SalesDiscountsSumWithoutDownloadProducts') {
            $hInfo->{$Name} = $self->_getItemsSum($Object, 'SalesDiscounts', { 'WithoutDownloadProducts' => 1 });
        }
        elsif ($Name eq 'PrePaymentsSum') {
            $hInfo->{$Name} = $self->_getItemsSum($Object, 'PrePayments');
        }
        elsif ($Name eq 'DiscountsSum') {
            $hInfo->{$Name} = $self->_getItemsSum($Object, 'Discounts');
        }
        elsif ($Name eq 'DiscountsSumWithoutDownloadProducts') {
            $hInfo->{$Name} = $self->_getItemsSum($Object, 'Discounts', { 'WithoutDownloadProducts' => 1 });
        }
        elsif ($Name eq 'LineItemsSubTotal') {
            $hInfo->{$Name} = $Object->get('LineItemsSum');
        }
        elsif ($Name eq 'LineItemsSubTotalWithoutDownloadProducts') {
            $hInfo->{$Name} = $self->_getItemsSum($Object, 'LineItems', { 'WithoutDownloadProducts' => 1 });
        }
        elsif ($Name eq 'LineItemsSubTotalDiscountable') {
            $hInfo->{$Name} = $self->_getItemsSumDiscountable($Object, 'LineItems');
        }
        elsif ($Name eq 'OriginalPriceSubTotal') {
            $hInfo->{$Name} = $self->_getItemsSumByOriginalPrice($Object, 'LineItems');
        }
        elsif ($Name eq 'PrepaymentTotalBase') {
            $hInfo->{$Name} = $self->_getItemsSumForPrepayment($Object, 'LineItems');
        }
        elsif ($Name eq 'PrepaymentTotal') {
            $hInfo->{$Name} = $self->_getItemsSumByPrepayment($Object, 'LineItems');
        }
        elsif ($Name eq 'SalesDiscountsSubTotal') {
            $hInfo->{$Name} = $Object->get('LineItemsSubTotal') + $Object->get('SalesDiscountsSum');
        }
        elsif ($Name eq 'SalesDiscountsSubTotalWithoutDownloadProducts') {
            $hInfo->{$Name} = $Object->get('LineItemsSubTotalWithoutDownloadProducts') + $Object->get('SalesDiscountsSumWithoutDownloadProducts');
        }
        elsif ($Name eq 'SalesDiscountsSubTotalDiscountable') {
            $hInfo->{$Name} = $Object->get('LineItemsSubTotalDiscountable') + $Object->get('SalesDiscountsSum');
        }
        elsif ($Name eq 'DiscountsSubTotal') {
            $hInfo->{$Name} = $Object->get('SalesDiscountsSubTotal') + $Object->get('DiscountsSum');
        }
        elsif ($Name eq 'DiscountsSubTotalWithoutDownloadProducts') {
            $hInfo->{$Name} = $Object->get('SalesDiscountsSubTotalWithoutDownloadProducts') + $Object->get('DiscountsSumWithoutDownloadProducts');
        }
        elsif ($Name eq 'ShippingSubTotal') {
            my $Shipping = $Object->get('Shipping');
            my $ShippingPrice = defined $Shipping ? $Shipping->get('LineItemPrice') : 0 ;
            $hInfo->{$Name} = $Object->get('DiscountsSubTotal') + (defined $ShippingPrice ? $ShippingPrice :0);
        }
        elsif ($Name eq 'PaymentSubTotal') {
            my $Payment = $Object->get('Payment');
            my $PaymentPrice = defined $Payment ? $Payment->get('LineItemPrice') : 0 ;
            $hInfo->{$Name} = $Object->get('ShippingSubTotal') + (defined $PaymentPrice ? $PaymentPrice :0);
        }
        elsif ($Name eq 'PaymentDiscountSubTotal') {
            my $PaymentDiscount = $Object->get('PaymentDiscount');
            my $PaymentDiscountPrice = defined $PaymentDiscount ? $PaymentDiscount->get('LineItemPrice') : 0 ;
            $hInfo->{$Name} = $Object->get('PaymentSubTotal') + (defined $PaymentDiscountPrice ? $PaymentDiscountPrice :0);
        }
        elsif ($Name eq 'TaxMatrixID') {
            my $TaxArea = $Object->get('TaxArea');
            $hInfo->{$Name} = defined $TaxArea
                ? $TaxArea->get('TaxMatrix')->id
                : $Object->getSite->get('TaxMatrix')->id;
        }
        elsif ($Name eq 'TaxMatrixDigitalID') {
            my $TaxArea = $Object->get('TaxAreaDigital');
            if(defined $TaxArea){
                my $TaxMatrixDigital = $TaxArea->get('TaxMatrixDigital');
                $hInfo->{$Name} = $TaxMatrixDigital->id if defined $TaxMatrixDigital;
            } else{
                my $ShopTaxMatrixDigital = $Object->getSite->get('TaxMatrixDigital');
                $hInfo->{$Name} = $ShopTaxMatrixDigital->id if defined $ShopTaxMatrixDigital;
            }
        }
        elsif ($Name eq 'TaxationAddressID') {
            # get the tax relevant address for taxation
            # this can be the billing address or shipping address, dependend on the elements of the address

            my $Address = _TaxationAddress($Object->businessContainer);
            $hInfo->{$Name} = defined $Address ? $Address->id : undef;
        }
        elsif ($Name eq 'ShippingCountryID') {
            $hInfo->{$Name} = _ShippingAddressInfo($Object->businessContainer)->{'CountryID'};
        }
        elsif ($Name eq 'BillingCountryID') {
            $hInfo->{$Name} = _BillingAddressInfo($Object->businessContainer)->{'CountryID'};
        }
        elsif ($Name eq 'TaxableLineItems') {
            my $LineItemTaxClass = LoadClassByAlias('LineItemTax');
            my $LineItemTaxDigitalClass = LoadClassByAlias('LineItemTaxDigital');
            $hInfo->{$Name} = [grep { not $_->instanceOf($LineItemTaxClass) and not $_->instanceOf($LineItemTaxDigitalClass) }
                                   @{LoadClassByAlias('LineItem')->getObjects('ByLineItemContainerID', [$Object->id])}];
        }
        elsif ($Name eq 'NoTax') {
            $hInfo->{$Name} = _NoTax($Object->businessContainer);
        }
        elsif ($Name eq 'NoTaxIgnoreTaxModel') {
            $hInfo->{$Name} = _NoTaxByVATID($Object->businessContainer);
        }
        elsif ($Name eq 'NeedPayment') {
            $hInfo->{$Name} = $Object->needPayment > 0 ? 1 : 0 ;
        }
        elsif ($Name eq 'NeedShipping') {
            $hInfo->{$Name} = $Object->needShipping > 0 ? 1 : 0;
        }
        elsif ($Name eq 'ShowTaxRate') {
            $hInfo->{$Name} = !($Object->get('NoTax') || $Object->getSite->get('VATreleased')) ? 1 : 0;
        }
        elsif ($Name eq 'BasketOrOrderID') {
            my $BasketOrOrder = $Object->parent;
            if( (not $BasketOrOrder->instanceOf('CustomerOrder')) or (not $BasketOrOrder->instanceOf('Basket')) ) {
                    $BasketOrOrder = undef;
                    foreach my $Parent (@{$Object->get('ParentObjects')}){
                        $BasketOrOrder = $Parent if ($Parent->instanceOf('CustomerOrder') or $Parent->instanceOf('Basket'));
                        last if $BasketOrOrder;
                    }
            }
            $hInfo->{$Name} = $BasketOrOrder->id;
        }
    }
    return $hInfo;
}

sub _sum {
    my $self = shift;
    my ($aLineItems, $WithoutDownloadProducts) = @_;

    my $Sum = 0;
    foreach my $LineItem (@$aLineItems) {
        my $Product;
        $Product = $LineItem->get('Product') if $LineItem->class->alias eq 'LineItemProduct';
        next if $WithoutDownloadProducts and defined $Product and $Product->get('IsDownloadProduct');

        my $Price = $LineItem->get('LineItemPrice');
        Error('NoLineItemPrice', { Path => $LineItem->pathString }) unless defined $Price;
        $Sum += $Price;
    }

    return $Sum;
}

sub _getItemsSum {
    my $self = shift;
    my ($Object, $ItemsName, $hParams) = @_;

    my $Sum = $self->_sum($Object->get($ItemsName), $hParams->{'WithoutDownloadProducts'});
    return $Object->formatter->round($Sum);
}

sub _getItemsSumDiscountable {
    my $self = shift;
    my ($Object, $ItemsName) = @_;

    my $Sum = $self->_sum( [ grep { !$_->get('IsStatic') } @{$Object->get($ItemsName)} ]);
    return $Object->formatter->round($Sum);
}

sub _getItemsSumByOriginalPrice {
    my $self = shift;
    my ($Object, $ItemsName) = @_;

    my $Sum = 0;
    foreach my $LineItem ( @{$Object->get($ItemsName)}) {
        my $OriginalPrice = $LineItem->get('OriginalLineItemPrice');
        $Sum += $OriginalPrice ? $OriginalPrice : $LineItem->get('LineItemPrice');
    }
    return $Object->formatter->round($Sum);
}

sub _getItemsSumForPrepayment {
    my $self = shift;
    my ($Object, $ItemsName) = @_;

    my $Sum = 0;
    foreach my $LineItem ( @{$Object->get($ItemsName)}) {
        my $OriginalPrice = $LineItem->get('OriginalLineItemPrice');
        next unless defined $OriginalPrice;
        $Sum += $OriginalPrice;
    }
    return $Object->formatter->round($Sum);
}

sub _getItemsSumByPrepayment {
    my $self = shift;
    my ($Object, $ItemsName) = @_;

    my $Sum = $self->_sum( [ grep { defined $_->get('OriginalLineItemPrice')} @{$Object->get($ItemsName)} ]);
    return $Object->formatter->round($Sum);
}

sub _paymentMethods {
    my $self = shift;
    my ($Object) = @_;

    # find payment methods that can be used with the selected shipping method
    my $ShippingLineItem = $Object->get('Shipping');
    my $ShippingMethod = defined $ShippingLineItem ? $ShippingLineItem->get('ShippingMethod') : undef;
    my $aPaymentMethods = defined $ShippingMethod
        ? $ShippingMethod->get('PaymentMethods')
        : $Object->getSite()->get('PaymentMethodsWithoutShipping');

    # options for canAddToBasket
    my $hOptions = $Object->get([qw(CurrencyID TaxModel LocaleID)]);
    $hOptions->{'LineItemContainer'} = $Object;

    my @sortPaymentMethods =
        sort { $a->get('Position')<=>$b->get('Position') }
        grep { $_->canAddToBasket($hOptions) }
        @$aPaymentMethods;

    TriggerHook( 'LineItemContainerPaymentMethods', {
        LineItemContainer => $Object,
        PaymentMethods => \@sortPaymentMethods,
    });

    return \@sortPaymentMethods;
}

sub _shippingMethods {
    my $self = shift;
    my ($Object) = @_;

    # options for canAddToBasket
    my $hOptions = $Object->get([qw(CurrencyID TaxModel)]);
    $hOptions->{'LineItemContainer'} = $Object;

    my $Shop = $Object->getSite();
    my $aShippingMethods = [grep { $_->canAddToBasket($hOptions) } @{$Shop->get('ShippingMethods')}];

    TestObject('ContainerShippingMethod',$_,'DE_EPAGES::Order::API::Object::ContainerShippingMethod') foreach  @{$aShippingMethods};

    # allow line items to select shipping methods
    $aShippingMethods = $_->grepAvailableShippingMethods($aShippingMethods) foreach @{$Object->get('LineItems')};

    my @sortShippingMethods =
        sort { $a->get('Position')<=>$b->get('Position') }
        @$aShippingMethods;

    TriggerHook( 'LineItemContainerShippingMethods', {
        LineItemContainer => $Object,
        ShippingMethods => \@sortShippingMethods,
    });

    return \@sortShippingMethods;
}

sub _GetIsTaxClassRemovedFromLineItems {
    my $self = shift;
    my ($Object) = @_;

    foreach my $Items ( qw ( LineItems)) { #Shipping Payment
        my $aItem  = $Object->get($Items);
        if (ref $aItem eq 'ARRAY') {
            return 1 if grep { not defined $_->get('TaxClass') } @$aItem;
        } else {
            return 1 unless defined $aItem->get('TaxClass');
        }
    }
    return 0;
}

sub _GetIsTaxClassParentBackup {
    my $self = shift;
    my ($Object) = @_;

    foreach my $Items ( qw (LineItems)) { #Shipping Payment
        my $aItem  = $Object->get($Items);
        foreach my $LineItem (@$aItem) {
            my $TaxClassDigital;
            my $TaxClass;
            if (defined $LineItem->get('TaxClassDigital')) {
                $TaxClassDigital = $LineItem->get('TaxClassDigital')->parent;
                #return if TaxClassDigital is from a backuped taxmatrix
                return 1 if $TaxClassDigital->parent->alias eq 'TaxMatrixBackup';
            }
            if (defined $LineItem->get('TaxClass')) {
                $TaxClass = $LineItem->get('TaxClass')->parent;
                #return if TaxClass is from a backuped taxmatrix
                return 1 if $TaxClass->parent->alias eq 'TaxMatrixBackup';
            }
        }
    }
    return 0;
}

sub _TaxationAddress {
    my ($Container) = @_;
    my $ContainerParent = $Container->parent;

    # ignore Classes like PackingSlip because they do not have a BillingAddress
    my $hAddressInfo = {};
    if ($ContainerParent->instanceOf('CustomerOrder') or
        $ContainerParent->instanceOf('Basket')) {
        $hAddressInfo = $ContainerParent->get(['ShippingAddress', 'BillingAddress']);

        # fallback to customer or user addresses if we have not container addresses
        if((not defined $hAddressInfo->{'ShippingAddress'}) and (not defined $hAddressInfo->{'BillingAddress'})) {
            my $UserOrCustomer = $ContainerParent->get('User') // $ContainerParent->get('Customer');
            # Customer object has only a BillingAddress but Users have both
            if(defined $UserOrCustomer ){
                my $aAddressesToGet = $UserOrCustomer->instanceOf('Customer') ? ['BillingAddress'] : ['ShippingAddress', 'BillingAddress'] ;
                $hAddressInfo = $UserOrCustomer->get($aAddressesToGet);
            }
        }
    }

    # test if we have a ShippingAddress. If so we compare the ShippingAddress with the BillingAddress
    # and use the BillingAddress if both are equal and only the VAT ID is missing (UNITY-8634)
    my $Address;
    if (defined $hAddressInfo->{'ShippingAddress'}){
        #if we have a VATID on the ShippingAddress then we use this address
        if($hAddressInfo->{'ShippingAddress'}->get('VATID')){
            $Address =  $hAddressInfo->{'ShippingAddress'};
        } else {
            # no VATID on the shipping address, so we check if  BA and SA are equal
            # if so we use the BA with the maybe existing VATID otherwise use the SA
            my $aIgnoreThisElements = ['VATID', 'EMail'];
            $Address = $hAddressInfo->{'ShippingAddress'}->CompareAddress($hAddressInfo->{'BillingAddress'}, $aIgnoreThisElements) ?
                $hAddressInfo->{'BillingAddress'} : $hAddressInfo->{'ShippingAddress'} ;
        }

    } else {
        $Address = $hAddressInfo->{'BillingAddress'};
    }

    return $Address;

}

sub _ShippingAddressInfo {
    my ($Container) = @_;

    my $Address = _TaxationAddress($Container);

    my $hInfo = {};
    if (defined $Address) {
        $hInfo->{'CountryID'} = $Address->get('CountryID');
        $hInfo->{'AddressIsEU'} = IsEUAddress($Address);
        $hInfo->{'VATID'} = $Address->get('VATID');
        $hInfo->{'Address'} = $Address;
    }

    return $hInfo;
}

sub _BillingAddressInfo {
    my ($Container) = @_;
    my $ContainerParent = $Container->parent;
        # ignore Classes like PackingSlip because they do not have a BillingAddress
    my $Address;
    if ($ContainerParent->instanceOf('CustomerOrder') or
        $ContainerParent->instanceOf('Basket')) {
        $Address = $ContainerParent->get('BillingAddress');

        if(not defined $Address) {
            my $User = $ContainerParent->get('User');
            $Address = $User->get('BillingAddress') if defined $User;
            if(not defined $Address) {
                my $Customer = $ContainerParent->get('Customer');
                $Address = $Customer->get('BillingAddress') if defined $Customer;
            }
        }
    }

    my $hInfo = {};

    if (defined $Address) {
        $hInfo->{'CountryID'} = $Address->get('CountryID');
        $hInfo->{'AddressIsEU'} = IsEUAddress($Address);
        $hInfo->{'VATID'} = $Address->get('VATID');
        $hInfo->{'Address'} = $Address;
    }

    return $hInfo;

}

sub _NoTax {
    my ($Container) = @_;

    return _NoTaxByVATID($Container);

}

sub _NoTaxByVATID {
    my ($Container) = @_;
    my $ContainerParent = $Container->parent;
    my $Shop = $ContainerParent->getSite();
    if ($ContainerParent->instanceOf('OrderDocument')) {
        $ContainerParent = $ContainerParent->get('Order');
        $Container = $ContainerParent->container;
    }

    # get shop country id from shop address
    my $ShopCountryID = GetTaxCountryIDForShop($Shop);
    my $IsEUShopCountry = IsEUAddress($Shop->get('Address'));

    # get information about the tax relevant address. This is always the shipping address!
    my $hTaxationCountryInfo = _ShippingAddressInfo($Container) ;
    my $TaxationCountryID = $hTaxationCountryInfo->{'CountryID'} // $ShopCountryID;
    my $TaxationIsEUCountry = $hTaxationCountryInfo->{'AddressIsEU'};

    #GetLog()->debug("Use countryid $TaxationCountryID for tax-calculation");
    #GetLog()->debug('ShopCountry: ' . ($ShopCountryID // '') . ', Country: ' . ($TaxationCountryID // ''));

    # do we have a country information ?
    if (defined $ShopCountryID && defined $TaxationCountryID) {

        # Shop country is not the shipping country
        # and both countries are EU countries
        if ( ($TaxationIsEUCountry && $IsEUShopCountry) and ($TaxationCountryID != $ShopCountryID)) {
            GetLog()->debug('Country for shipping is not shopcountry and both are EU-countries');
            # for taxation address not shopaddress and both in EU: return true if customer has VATID
            if ($TaxationCountryID && $TaxationCountryID != $ShopCountryID && $TaxationIsEUCountry) {
                # test VATID from customers taxation address
                my $VATID = $hTaxationCountryInfo->{'VATID'};
                GetLog()->debug('Country for taxation is not shopcountry and both are EU-countries, VATID is: ' . ($VATID ? $VATID : 'undef'));
                return 1 if defined $VATID && $VATID ne '';
            }
        }
    }
    GetLog()->debug('_NoTaxIgnoreTaxModel returns 0');

    # we have not a VATID on the taxation relevant address
    return 0;
}

1;
