#========================================================================================
# package      DE_EPAGES::PayPalPPCP::API::TLE::PayPalPPCPHandlerHandler
# state        public
#----------------------------------------------------------------------------------------
# description Implements a TLE Function which returns preformatted PayPalPPCP widget HTML code
#========================================================================================
package DE_EPAGES::PayPalPPCP::API::TLE::PayPalPPCPHandler;

use strict;
use JSON;
use MIME::Base64;
use URI::Escape qw(uri_escape_utf8);
use DE_EPAGES::Presentation::API::DereferrerToken qw (GenerateDereferrerToken);
use DE_EPAGES::WebInterface::API::ApplicationServer qw (GetAppServer);
use DE_EPAGES::Payment::API::Table::ShippingPaymentMap qw (
    ExistsShippingPaymentMap
);
use DE_EPAGES::PayPalPPCP::API::Constants qw (@FEATURE_PAYPALPPCP_S_METHODS $DISABLED_APMS);

#========================================================================================
# function     RegisterHandlerProc
# state        private
#----------------------------------------------------------------------------------------
# syntax       $RegisterHandlerProc( $hParams );
#----------------------------------------------------------------------------------------
# description  this function registers this attribute handle on the processor via hook
#               function on hook TLEProcessorRegistration
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments keys :
#               <ul>
#                   <li>Processor - tle processor - object
#               </ul> | ref.hash
#========================================================================================
sub RegisterHandlerProc {
    my ($Params) = @_;

    __PACKAGE__->new()->register( $Params->{'Processor'} );
    return;
}

#========================================================================================
# function     new
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler = new DE_EPAGES::PayPalPPCP::API::TLE::PayPalPPCPHandler;
#----------------------------------------------------------------------------------------
# description  create a new handler
#----------------------------------------------------------------------------------------
# return       $Handler | blessed object | object
#========================================================================================
sub new { bless {}, shift;}

#========================================================================================
# function     tle
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Value = $Handler->tle( $Processor, $TLEName);
# example      my $Value = $Handler->tle( $Processor, 'Name');
#----------------------------------------------------------------------------------------
# description  This function is called by tle processor to get the value of a tle
#               variable.
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
# input        $TLEName | tle name | string
# return       $Value | return value | integer
#========================================================================================
sub tle {
    my $self = shift;
    my ($Processor, $TLEName) = @_;

    my $Object = $Processor->tle('Object');
    my $Shop = $Object->getSite();

    return undef unless $Shop->instanceOf('Shop');

    my $Servlet = GetAppServer()->currentServlet();
    my $IsMobile = $Servlet->vars('IsMobile') ? 1 : 0;

    if( $TLEName eq 'ExtractedPayPalPPCPParameters' ) {
        my $ClassAlias = 'PaymentMethodPayPalPPCP';
        my $ViewAction = $Servlet->vars('ViewAction') // $Servlet->vars('INPUT.ViewAction') // $Servlet->params('ViewAction');
        my $IsMBOSettingsPage = ($ViewAction eq 'MBO-ViewSettings' || $ViewAction eq 'UnityMBO-ViewSettings') && $Object->class->alias eq $ClassAlias ? 1 : 0;

        my ($PaymentMethod) = grep {$_->class->alias eq $ClassAlias && ($_->get('IsActivated') || $IsMBOSettingsPage) } @{$Shop->get('PaymentMethods')};

        return undef if ! $PaymentMethod;
        return undef if ! $PaymentMethod->hasAppCredentials();

        my $CreateOrderUrl  = $self->_getOrderActionURL($Shop, 'Create', $IsMobile);
        my $ApproveOrderUrl = $self->_getOrderActionURL($Shop, 'Approve', $IsMobile);

        my $hPaymentMethodInfo = $PaymentMethod->get(['ClientID', 'MerchantInfoJSON', 'EnablePayLater', 'EnablePayLaterBanners', 'EnableCartButtons']);
        my $PartnerBNCode = $PaymentMethod->getBNCode();

        #-- Minimum/Maximum amounts for the payment method. The basket has to be between these amounts for the buttons to show.
        my $aShopCurrencies = $Shop->get('Currencies');
        my $hMinMaxAllCurrencies = {};
        foreach my $hCurrencyInfo ( @$aShopCurrencies ) {
            next if ! $hCurrencyInfo->{'IsActivated'};
            next if ! $hCurrencyInfo->{'IsUsed'};

            $hMinMaxAllCurrencies->{$hCurrencyInfo->{'CurrencyID'}} = {
                MinPriceValue => $PaymentMethod->getMinPriceValue($hCurrencyInfo->{'CurrencyID'}),
                MaxPriceValue => $PaymentMethod->getMaxPriceValue($hCurrencyInfo->{'CurrencyID'}),
            };
        }

        my $ShippingMethods = $Shop->get('ShippingMethods');
        my $ahLanguages = $Shop->get('Languages');
        my $hIncompatibleShippingMethodNames = {};
        my $hShippingMethodLandRestrictions = {}; #-- Region/country restrictions per shipping method, used for 'onShippingChange'

        #-- Shipping/Payment method dependencies. If the cart has a shipping method that is NOT available with PPCP, we want to know about it
        foreach my $ShippingMethod (@$ShippingMethods) {
            #-- This means, this shipping method is NOT compatible with our PPCP payment method
            if ( ! ExistsShippingPaymentMap($ShippingMethod->id, $PaymentMethod->id) ) {
                foreach my $hInfo ( @$ahLanguages ) {
                    my $NameOrAlias = $ShippingMethod->get('NameOrAlias', $hInfo->{'LanguageID'});
                    $hIncompatibleShippingMethodNames->{ encode_base64(uri_escape_utf8($NameOrAlias), '') } = 1;
                }
            } else {
              #-- This shipping method is compatible
              my $Region = $ShippingMethod->get('Region');
              next if ! $Region;
              my $aCountries = $Region->get('Countries');
              $hShippingMethodLandRestrictions->{$ShippingMethod->id} = {};
              $hShippingMethodLandRestrictions->{$ShippingMethod->id}->{$_->{'Code2'}} = 1 foreach @$aCountries;
            }
        }

        #-- Payment Method region restriction
        my $aPaymentCountryRestrictions = [];
        my $Region = $PaymentMethod->get('Region');
        my $aCountries = $Region ? $Region->get('Countries') : [];
        push @$aPaymentCountryRestrictions, $_->{'Code2'} foreach @$aCountries;

        return {
            'ClientID'          => $hPaymentMethodInfo->{ 'ClientID'},
            'MerchantID'        => $hPaymentMethodInfo->{'MerchantInfoJSON'}->{'merchant_id'},
            'CurrencyID'        => $Shop->get('CurrencyID'),
            'PartnerBNCode'     => $PartnerBNCode,
            'EnablePayLater'    => $hPaymentMethodInfo->{ 'EnablePayLater'},
            'EnablePayLaterBanners'    => $hPaymentMethodInfo->{ 'EnablePayLaterBanners'},
            'ApprovePPCPToken'  => $self->_generatePickupToken($ApproveOrderUrl),
            'ApproveOrderUrl'   => $ApproveOrderUrl,
            'CreateOrderUrl'    => $CreateOrderUrl,
            'MinMaxAllCurrencies' => to_json($hMinMaxAllCurrencies),
            'InCompatibleShippingMethods' => to_json($hIncompatibleShippingMethodNames),
            'ShippingMethodLandRestrictions' => to_json($hShippingMethodLandRestrictions),
            'EnableCartButtons' => $hPaymentMethodInfo->{ 'EnableCartButtons'},
            'PaymentMethod'     => $PaymentMethod->tleHash,
            'PaymentMethodRegionRestrictions' => scalar @$aPaymentCountryRestrictions > 0 ? $aPaymentCountryRestrictions : undef,
        };
    }

    #-- https://developer.paypal.com/limited-release/fraudnet/integrate/
    if ( $TLEName eq 'PayPalPPCPFraudNetFParameter' ) {
        if ( $Object->instanceOf('Basket') ) {
            my $PaymentLine = $Object->container->get('Payment');
            if ( $PaymentLine->instanceOf('LineItemPaymentPayPalPPCP') ) {
                my $PaymentMethod = $PaymentLine->get('PaymentMethod');
                my $FParameter = $PaymentMethod->getFraudNetFParameter($PaymentLine);
                return $FParameter;
            }
        }

        return '';
    }

    #-- https://developer.paypal.com/sdk/js/configuration/#integration-date
    if ( $TLEName eq 'PayPalPPCPIntegrationDate' ) {
        return '2022-03-24';
    }

    if ( $TLEName eq 'PayPalPPCP_S_Methods' ) {
        return \@FEATURE_PAYPALPPCP_S_METHODS;
    }

    if ( $TLEName eq 'PayPalPPCP_DISABLED_APMS' ) {
        return $DISABLED_APMS;
    }

    return undef;
}

#========================================================================================
# function     existsTLE
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Value = $Handler->existsTLE( $Processor, $TLEName);
# example      my $Value = $Handler->existsTLE( $Processor, 'Products_Count');
#----------------------------------------------------------------------------------------
# description  This function is called by the tle processor to test if a TLE variable
#               is available.
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
# input        $TLEName | tle name | string
# return       $Value | yes or no | bool
#========================================================================================
sub existsTLE {
    my $self = shift;
    my ($Processor, $TLEName) = @_;

    return (defined $self->tle($Processor, $TLEName)) ? 1 : 0;
}

#========================================================================================
# function     PPCP_CanPUI
# state        public
#----------------------------------------------------------------------------------------
# syntax       #PPCP_CanPUI(#PaymentMethod)
# example      #PPCP_CanPUI(#PaymentMethod)
#----------------------------------------------------------------------------------------
# description  Check if this payment method supports PUI
#----------------------------------------------------------------------------------------
# input        #PaymentMethod | A PPCP Payment Method | object
# return       1/0 | Whether PUI is supported | boolean
#========================================================================================
sub PPCP_CanPUI { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($PaymentMethod, $Container) = @$aParams;

    return 0 unless $PaymentMethod && $PaymentMethod->instanceOf('PaymentMethodPayPalPPCP');

    return $PaymentMethod->canUsePUI($Container);
}

#========================================================================================
# function     GetBNCode
# state        public
#----------------------------------------------------------------------------------------
# syntax       #GetBNCode(#PaymentMethod)
# example      #GetBNCode(#PaymentMethod)
#----------------------------------------------------------------------------------------
# description  Get the BN Code for this ppcp payment method
#----------------------------------------------------------------------------------------
# input        #PaymentMethod | A PPCP Payment Method | object
# return       1/0 | Whether PUI is supported | boolean
#========================================================================================
sub GetBNCode { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($PaymentLine) = @$aParams;

    my $PaymentMethod = $PaymentLine->get('PaymentMethod');
    return '' unless $PaymentMethod && $PaymentMethod->instanceOf('PaymentMethodPayPalPPCP');

    return $PaymentMethod->getBNCode($PaymentLine);
}

sub register {
    my $self = shift;
    my ($Processor) = @_;

    $Processor->registerHandler( 'VariableHandler', $self, 'ExtractedPayPalPPCPParameters');
    $Processor->registerHandler( 'VariableHandler', $self, 'PayPalPPCPFraudNetFParameter');
    $Processor->registerHandler( 'VariableHandler', $self, 'PayPalPPCPIntegrationDate');
    $Processor->registerHandler( 'VariableHandler', $self, 'PayPalPPCP_S_Methods');
    $Processor->registerHandler( 'VariableHandler', $self, 'PayPalPPCP_DISABLED_APMS');
    $Processor->registerHandler( 'FunctionHandler', $self, 'PPCP_CanPUI' );
    $Processor->registerHandler( 'FunctionHandler', $self, 'GetBNCode' );

    return;
}

sub _getOrderActionURL {
    my $self = shift;
    my ($Shop, $ActionName, $IsMobile) = @_;

    #-- Use the right .sf/.mobile extension
    my $Url = $Shop->get($IsMobile?'WebServerScriptUrlMobileSSL':'WebServerScriptUrlSSL');

    return "$Url?ViewAction=JSONViewResponse&ChangeAction=JSON" . $ActionName . "PayPalPPCPOrder";
}

sub _generatePickupToken {
    my $self = shift;
    my ( $PickupTokenUrl ) = @_;

    return GenerateDereferrerToken($PickupTokenUrl);
}

1;

