#==============================================================================
# §package      DE_EPAGES::PayPalPPCP::Hooks::OrderProcess
# §state        private
#==============================================================================
package DE_EPAGES::PayPalPPCP::Hooks::OrderProcess;

use DE_EPAGES::Object::API::Factory qw( LoadClassByAlias );
use DE_EPAGES::Order::API::LineItemContainer qw( UpdateTaxModel SaveShipping );
use DE_EPAGES::Presentation::API::Session qw( SetSessionAttribute );
use DE_EPAGES::Core::API::Error qw (Error);
use DE_EPAGES::Core::API::PerlTools qw (ElementOf);
use DE_EPAGES::Order::UI::Basket;

use strict;
use DE_EPAGES::PayPalPPCP::API::Constants qw ( $FS_PAYPAL $FS_PAYLATER $FS_CARD $INVOICE_FS $ORDER_STATUS_CREATED $ORDER_STATUS_VOIDED $ORDER_STATUS_APPROVED $ORDER_STATUS_COMPLETED $ORDER_STATUS_PAYER_ACTION_REQUIRED @FEATURE_PAYPALPPCP_S_METHODS );

#========================================================================================
# §function     OnUI_ViewMultiCheckoutPayment
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       OnUI_ViewMultiCheckoutPayment($hParams);
#----------------------------------------------------------------------------------------
# §description  Executed when viewing the payment page of the individual steps checkout
#----------------------------------------------------------------------------------------
# §input        $hParams | hook arguments, used keys are
#               <ul>
#                   <li>Servlet - servlet - object</li>
#               </ul> | ref.hash.*
#========================================================================================
sub OnUI_ViewMultiCheckoutPayment {
    my ($hParams) = @_;
}

#========================================================================================
# §function     OnUI_AfterSaveMultiCheckoutPayment
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       OnUI_AfterSaveMultiCheckoutPayment($hParams);
#----------------------------------------------------------------------------------------
# §description  Executed after clicking 'Next' on the payment page of the individual steps checkout
#----------------------------------------------------------------------------------------
# §input        $hParams | hook arguments, used keys are
#               <ul>
#                   <li>Servlet - servlet - object</li>
#               </ul> | ref.hash.*
#========================================================================================
sub OnUI_AfterSaveMultiCheckoutPayment {
    my ($hParams) = @_;

    my $Servlet = $hParams->{'Servlet'};

    #-- If this hook was recalled from a classic hook, this parameter is our own
    my $PPCPClassic = $hParams->{'PPCPClassic'};
    my $IsMobile = $Servlet->vars('IsMobile') ? 1 : 0;

    my $Basket = $Servlet->object;
    return if $Basket->get('OrderGUID');

    my $Container = $Basket->container;
    return if ! defined $Container;

    my $PaymentLine = $Container->get('Payment');
    return if ! defined $PaymentLine;
    return unless $PaymentLine->instanceOf('LineItemPaymentPayPalPPCP');

    my $PaymentMethod = $PaymentLine->get('PaymentMethod');
    my $Shop = $PaymentMethod->getSite();

    my $Form = $Servlet->form;

    my $FundingSource = $Form->value('FundingSource');
    if ( $FundingSource && $FundingSource =~ /\S/ ) {

        #-- Verify again that PUI can be used, just in case
        if ( $FundingSource eq $INVOICE_FS ) {
            $FundingSource = undef if ! $PaymentMethod->canUsePUI($PaymentMethod, $PaymentLine->container);
        }

        $PaymentLine->set({
            'FundingSource' => $FundingSource,
        });
    } else {
        $PaymentLine->set({
            'FundingSource' => undef,
        });
    }

    #-- Check that the currently selected funding source is valid for the 'S' package
    if ( $FundingSource && $Shop->get('FeatureMaxValue')->{'PayPalPPCP_S'} && ! ElementOf($FundingSource, @FEATURE_PAYPALPPCP_S_METHODS) ) {
        $FundingSource = undef;
        $PaymentLine->set({
            'FundingSource' => $FundingSource,
        });
    }

    $PaymentLine->set({
        'PayViaPUI' => $FundingSource && $FundingSource eq $INVOICE_FS ? 1 : 0,
    });

    #-- This is initially created by the calling function, but at that point, 'PayViaPUI' has not been saved yet, so our Register/AfterRegister CheckoutProcess hooks will not have run correctly
    #-- Therefore, run it again by (re)creating the BasketOrderProcess object, which triggers off those hooks
    if ( ! $PPCPClassic ) {
      my $BasketOrderProcess = DE_EPAGES::Order::API::Object::BasketOrderProcess->new($Servlet);
      $Servlet->vars('BasketOrderProcess',$BasketOrderProcess->tleHash);
    }

    #-- Verify birthday/phone
    if ( $FundingSource && $FundingSource eq $INVOICE_FS ) {
        my $AddressClass = LoadClassByAlias('Address');

        my $HasError = 0;
        my $hInfo;
        foreach my $AttrAlias ( qw/Birthday Phone/ ) {
            my $Attribute = $AddressClass->attribute($AttrAlias);

            my $hFieldDef = { Name => $AttrAlias, Form => 'Billing' . 'Address' };
            $hFieldDef->{Mandatory} = 1;
            $hFieldDef->{Type} = 'reg_date' if $AttrAlias eq 'Birthday';
            $hFieldDef->{Type} = 'string' if $AttrAlias eq 'Phone';
            $hFieldDef->{MaxValue} = $Attribute->get('Length') if $AttrAlias eq 'Phone' && $Attribute->get('Length');

            my $OrgValue = $Form->value( $AttrAlias );

            my $Val = $Form->field( $Shop, $OrgValue, $hFieldDef );

            if ( ! defined $Val ) {
                $HasError = 1;
            } else {
                $hInfo->{$AttrAlias} = $Val;
            }
        }

        #-- I think other handlers may have already added their form errors, so in order
        #-- for ours to take precedence, we need to clear all existing form errors??
        if ( $HasError ) {
            $Form->resetFormErrors();
            $Form->executeFormError({
                'ViewAction' => $PPCPClassic ? $IsMobile ? 'MobileSF-View' : 'View' : 'ViewMultiCheckoutPayment',
                'Name' => 'MAND_FIELDS',
                'Reason' => 'PUI_INVALID_MAND_FIELDS',
                'Form' => 'SavePUI',
            });
        } else {
            if ( $PPCPClassic ) {
              #-- If we add a billing address without a country, PUI will no longer be available when reloading the cart, since we check
              #-- if the billing address is from Germany
              #-- And we also get an error when navigating to the address page after PUI because the address hook will recognise that the
              #-- address is incomplete, but is that so bad?
              $hInfo->{'CountryID'} = 276;
              #-- If Base and no address, create one
              DE_EPAGES::Order::UI::Basket->saveAddresses($Servlet, {
                'Billing' => $hInfo,
              });
            }

            my $BillingAddress = $Basket->get('BillingAddress');
            $BillingAddress->set($hInfo) if $BillingAddress;
        }
    }
}

#========================================================================================
# §function     OnUI_ViewMultiCheckoutConfirmation
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       OnUI_ViewMultiCheckoutConfirmation($hParams);
#----------------------------------------------------------------------------------------
# §description  Executed when viewing the checkout confirmation page of the individual steps checkout
#----------------------------------------------------------------------------------------
# §input        $hParams | hook arguments, used keys are
#               <ul>
#                   <li>Servlet - servlet - object</li>
#               </ul> | ref.hash.*
#========================================================================================
sub OnUI_ViewMultiCheckoutConfirmation {
    my ($hParams) = @_;

    my $Servlet = $hParams->{'Servlet'};

    my $Basket = $Servlet->object;
    my $Shop = $Basket->getSite();

    return if $Basket->get('OrderGUID');

    my $Container = $Basket->container;
    return if ! defined $Container;

    my $PaymentLine = $Container->get('Payment');
    return if ! defined $PaymentLine;
    return if ! $PaymentLine->instanceOf('LineItemPaymentPayPalPPCP');

    my $PaymentMethod = $PaymentLine->get('PaymentMethod');

    my $hOrderInfo = $PaymentLine->get('OrderInfoJSON');
    if ( $hOrderInfo && %$hOrderInfo ) {
        _updatePPCPOrder($PaymentLine, $PaymentMethod);
    }

    #-- If we are coming from the express button
    my $Fppcp = $Servlet->form->value('Fppcp');
    if ( $Fppcp ) {
        my $CurrentGrandTotal = $Container->get('GrandTotal');

        #-- Because the address that comes from PayPal may not be compatible with region restrictions, if set up
        #-- This will do nothing if the existing shipping method is viable, if not, it will replace it with the next viable
        #-- A message is also shown indicating a change in the basket
        SaveShipping($Container, undef);

        #=================================
        #-- Taken from Order::UI::OrderProcess::SaveMultiCheckoutAddress
        #=================================
        # update TaxModel, if address was changed
        my $TaxModelChanged = UpdateTaxModel($Container);
        if (defined $TaxModelChanged) {
            SetSessionAttribute( $Servlet->intSessionID, 'TaxModelChanged', 1) unless $Shop->get('VATreleased');
        }

        #-- Recalculate, our shipping may have changed, or the tax model may have changed, whatever
        $Container->recalculate();

        my $TaxChanged;
        my $TaxChangedDigital;

        # test if we have a TaxMethod
        if($Shop->get('TaxMethod')){
            $Shop->get('TaxMethod')->updateContainerTaxArea($Container);
        }else {
            # update tax area, if address changed
            $TaxChanged = $Basket->updateTaxArea();
            $TaxChangedDigital = $Basket->updateTaxAreaDigital();
        }

        $Servlet->params('TaxChanged', 1) if $TaxChanged;
        $Servlet->params('TaxChangedDigital', 1) if $TaxChangedDigital;
        #=================================
        #-- Taken from Order::UI::OrderProcess::SaveMultiCheckoutAddress
        #=================================

        #-- When checking out for the first time using the express button, the cart's grand total is dependent on the dt (delivery threshold) country switch, if it is enabled
        #-- and since there is no existing address in the cart, the code above will not register 'TaxChanged'
        #-- Since it is difficult and time consuming to compare with the current value of the dt country switch, just compare the grand totals
        #-- Yeah it's not entirely accurate because it will also trigger this message if they just add/remove products in the meantime, for e.g., but for the most part ...
        my $NewGrandTotal = $Container->get('GrandTotal');
        if ( $CurrentGrandTotal != $NewGrandTotal ) {
            $Servlet->params('TaxChanged', 1);
            $Servlet->params('TaxChangedDigital', 1);
        }
    } else {
        #-- Normal checkout got us here
        ################################################
        ################################################

        return if $PaymentLine->get('PayViaPUI');

        #-- If for whatever reason, this order is 'completed', the payment has already been captured
        if ( _isPPCPOrderCompleted($PaymentLine) ) {
            return;
        }

        $hOrderInfo = $PaymentLine->get('OrderInfoJSON');

        if ( _shouldResetPPCPOrder($hOrderInfo, $PaymentLine) ) {
            $PaymentLine->set({
                'OrderInfoJSON' => undef,
                'OrderPayload'  => undef,
                'TransID2'      => undef,
            });
            $hOrderInfo = undef;
        }

        if ( ! $hOrderInfo || ! %$hOrderInfo ) {
            #-- If this call fails, OrderInfoJSON will be undefined and we show a message asking the customer to refresh the page, basically calling this function again
            my $hJsonResponse = $PaymentMethod->createPPCPOrder($PaymentLine);
            $PaymentLine->set({
                'OrderInfoJSON' => $hJsonResponse,
                'OrderPayload'  => defined $hJsonResponse ? $PaymentMethod->getOrderPayload($PaymentLine) : undef,
                'TransID2'      => $hJsonResponse->{'id'}
            }) if $hJsonResponse;

            return;
        }

        #-- At this point, we have order information from PayPal, AND it is not 'completed'

        #-- Is this order not valid anymore?
        #-- Patch it
        if ( ! _isPPCPOrderStillValid($PaymentLine) ) {
            #-- Let OnUI_AfterSaveMultiCheckoutConfirmation patch it
        }
    }
}

#========================================================================================
# §function     OnUI_AfterSaveMultiCheckoutConfirmation
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       OnUI_AfterSaveMultiCheckoutConfirmation($hParams);
#----------------------------------------------------------------------------------------
# §description  Executed after saving the basket confirmation page of the checkout
#               Because our PP button now resides on this page, this sub essentially takes
#               the place of our former explicit PP view/change action.
#
#               We go through many of the same checks as in OnUI_ViewMultiCheckoutConfirmation,
#               but we just return inside these checks here, as that view action will be executed again
#               in such cases.
#----------------------------------------------------------------------------------------
# §input        $hParams | hook arguments, used keys are
#               <ul>
#                   <li>Servlet - servlet - object</li>
#               </ul> | ref.hash.*
#========================================================================================
sub OnUI_AfterSaveMultiCheckoutConfirmation {
    my ($hParams) = @_;

    my $Servlet = $hParams->{'Servlet'};

    my $Basket = $Servlet->object;

    return if $Basket->get('OrderGUID');

    my $Container = $Basket->container;
    return if ! defined $Container;

    my $PaymentLine = $Container->get('Payment');
    return if ! defined $PaymentLine;
    return if ! $PaymentLine->instanceOf('LineItemPaymentPayPalPPCP');

    return if $PaymentLine->get('PayViaPUI');

    my $PaymentMethod = $PaymentLine->get('PaymentMethod');

    my $hOrderInfo = $PaymentLine->get('OrderInfoJSON');
    if ( $hOrderInfo && %$hOrderInfo ) {
        _updatePPCPOrder($PaymentLine, $PaymentMethod);
    }

    #-- If for whatever reason, this order is 'completed', the payment has already been captured, let 'SaveMultiCheckoutConfirmation' create the order
    if ( _isPPCPOrderCompleted($PaymentLine) ) {
        return;
    }

    #-- If this hook was recalled from a classic hook, this parameter is our own
    my $PPCPClassic = $hParams->{'PPCPClassic'};
    my $IsMobile = $Servlet->vars('IsMobile') ? 1 : 0;

    my $ErrorAction = $PPCPClassic ? $IsMobile ? 'MobileSF-ViewOffer' : 'ViewOffer' : 'ViewMultiCheckoutConfirmation';

    $hOrderInfo = $PaymentLine->get('OrderInfoJSON');
    if ( _shouldResetPPCPOrder($hOrderInfo, $PaymentLine) ) {
        #-- Just run the check here and ERROR it back to 'SaveMultiCheckoutConfirmation'
        Error('INVALID_FORM', {
            'ViewVars'=> {
                'FormError'    => 0,
                'ViewObjectID' => $Basket->id,
            },
            'ViewAction' => $ErrorAction,
        });
    }

    if ( ! $hOrderInfo || ! %$hOrderInfo ) {
        #-- Just run the check here and ERROR it back to 'SaveMultiCheckoutConfirmation'
        Error('INVALID_FORM', {
            'ViewVars'=> {
                'FormError'    => 0,
                'ViewObjectID' => $Basket->id,
            },
            'ViewAction' => $ErrorAction,
        });
    }

    #-- At this point, we have order information from PayPal, AND it is not 'completed'

    #-- Is this order not valid anymore?
    #-- Patch it
    if ( ! _isPPCPOrderStillValid($PaymentLine) ) {
        my $Patched = $PaymentMethod->patchPPCPOrder($PaymentLine);
        if ( $Patched ) {
            #-- Even if we fail to retrieve updated order info in the next line, we can at least update this attribute so that we don't try and patch again
            $PaymentLine->set({
                'OrderPayload' => $PaymentMethod->getOrderPayload($PaymentLine),
            });

            my $hJsonResponse = $PaymentMethod->getPPCPOrder($PaymentLine);
            $PaymentLine->set({
                'OrderInfoJSON' => $hJsonResponse,
                'TransID2'      => $hJsonResponse->{'id'},
            }) if $hJsonResponse;
        } else {
            #-- Show the payment page and ask the customer to refresh the page so that we can try again since we must patch in order to proceed
            Error('INVALID_FORM', {
                'ViewVars'=> {
                    'FormError'    => 0,
                    'ApiCallFailed'=> 1,
                    'ViewObjectID' => $Basket->id,
                },
                'ViewAction' => $ErrorAction,
            });
        }
    }

    #-- At this point, our order should be current with the basket/order we are operating on
    #-- If the status is 'APPROVED', we don't need to ask the customer to go through the payment at PP again, just capture it?
    $hOrderInfo = $PaymentLine->get('OrderInfoJSON');
    my $DoCaptureAnyway = 0;

    #-- google pay pre-capture check, check that SCA, if activated, has completed
    if ( exists $hOrderInfo->{'payment_source'}->{'google_pay'}->{'card'}->{'authentication_result'} ) {
      my $LiabilityShift = $hOrderInfo->{'payment_source'}->{'google_pay'}->{'card'}->{'authentication_result'}->{'liability_shift'};

      if ( $LiabilityShift eq 'NO' || $LiabilityShift eq 'UNKNOWN' ) {
        #-- It was not successful, ask the customer to try again
        Error('INVALID_FORM', {
            'ViewVars'=> {
                'FormError'    => 0,
                'ApiCallFailed'=> 1,
                'ViewObjectID' => $Basket->id,
            },
            'ViewAction' => $ErrorAction,
        });
      }
    }

    if ( $hOrderInfo->{'status'} eq $ORDER_STATUS_APPROVED ) {
        $DoCaptureAnyway = 1;
    }

    #-- Capture?
    my $CAP = $Servlet->form->value('CAP');
    if ( (defined $CAP && $CAP == 1) || $DoCaptureAnyway ) {
        my $hJsonResponse = $PaymentMethod->capturePPCPOrder($PaymentLine);
        # set the OrderInfoJSON and the PP order id for finding the order in an easy way (used for webhooks)
        $PaymentLine->set({
            'OrderInfoJSON' => $hJsonResponse,
            'TransID2'      => $hJsonResponse->{'id'},
        }) if $hJsonResponse;

        if ( _isPPCPOrderCompleted($PaymentLine) ) {
            $PaymentLine->updateLineItemFromPPCP();
            return;
        } else {
            #-- Do some error handling. Here, the order must at least be in 'APPROVED' status, or we wouldn't have tried to capture.
            Error('INVALID_FORM', {
                'ViewVars'=> {
                    'FormError'    => 0,
                    'CaptureFailed'=> 1,
                    'ViewObjectID' => $Basket->id,
                },
                'ViewAction' => $ErrorAction,
            });
        }
    }
}

#=============================
#   Check for situations in which the PayPal Order is no longer valid
#=================================
sub _shouldResetPPCPOrder {
    my ($hOrderInfo, $PaymentLine) = @_;

    #-- Voided, recreate the order @ paypal
    #-- https://developer.paypal.com/api/orders/v2/#orders_create
    #-- OR
    #-- Scenario: Completely new basket, the 'express' button is clicked on the cart page, and then the window is immediately closed BEFORE paying.
    #--           What is important is that the order object at PayPal is created.
    #-- This order has 'application_context'=>'shipping_preference' set to 'GET_FROM_FILE', which means address selection is allowed at PayPal
    #-- When we go to patch this order, we can't seem to update this parameter, and will allow address selection. We don't want this because if
    #-- we get to this code in this subroutine, it will be via standard checkout and not express
    #-- Ergo, check if we are in this state and clear the order
    if ( $hOrderInfo && %$hOrderInfo ) {
        my $hOrderPayLoad = $PaymentLine->get('OrderPayload');

        my $ShippingPreference = $hOrderPayLoad->{'payment_source'}->{'paypal'}->{'experience_context'}->{'shipping_preference'};
        if ( $hOrderInfo->{'status'} eq $ORDER_STATUS_VOIDED
            || ( ($hOrderInfo->{'status'} eq $ORDER_STATUS_CREATED || $hOrderInfo->{'status'} eq $ORDER_STATUS_PAYER_ACTION_REQUIRED) && $ShippingPreference && $ShippingPreference eq 'GET_FROM_FILE' ) ) {

            return 1;
        }

        #-- See API/Object/PaymentMethodPayPalPPCP.pm, sub getOrderPayload
        #-- We only pass a 'payment_source' in this case, when it is NOT an APM. This means that address selection is allowed at PayPal.
        #-- If we are now paying via one of these methods below, but the original paypal order was created for an APM,
        #-- we need to recreate it so that we DON'T allow address selection at PayPal, as this will confuse the customer, and could mess with
        #-- shipping calculations.
        if ( !$PaymentLine->get('FundingSource') || ElementOf($PaymentLine->get('FundingSource'), ($FS_PAYPAL, $FS_PAYLATER, $FS_CARD)) ) {
          if ( ! exists $hOrderInfo->{'payment_source'}->{'paypal'} ) {
            return 1;
          }
        }
    }

    return 0;
}

#=============================
#   Request an update for the PP order
#=================================
sub _updatePPCPOrder {
    my ($PaymentLine, $PaymentMethod) = @_;

    my $hJsonResponse = $PaymentMethod->getPPCPOrder($PaymentLine, 1);

    if ( $hJsonResponse ) {
        #-- This is a successful response
        if ( $hJsonResponse->{'id'}
                && $hJsonResponse->{'status'}
                && $hJsonResponse->{'intent'}
        ) {
            $PaymentLine->set({
                'OrderInfoJSON' => $hJsonResponse,
                'TransID2'      => $hJsonResponse->{'id'}
            });
        } else {
            #-- We have a json structure, presumably with errors
            if ( $hJsonResponse->{'name'} eq 'RESOURCE_NOT_FOUND' ) {
                #-- Need to recreate this order, since it can no longer be found at PayPal
                #-- Hopefully this condition never arises for 'completed' orders
                $PaymentLine->set({
                    'OrderInfoJSON' => undef,
                    'OrderPayload'  => undef,
                    'TransID2'      => undef,
                });
            }
        }
    }
}

#=============================
#   Check to see if the order has the 'completed' status, which means a payment has been captured at PayPal
#=================================
sub _isPPCPOrderCompleted {
    my ($PaymentLine) = @_;

    return 0 unless $PaymentLine->get('OrderInfoJSON');
    return 1 if $PaymentLine->get('OrderInfoJSON')->{'status'} eq $ORDER_STATUS_COMPLETED;

    return 0;
}

#=============================
#   Check to see if the order information we sent to PayPal is still valid
#   Most importantly, the total amount
#=====================================
sub _isPPCPOrderStillValid {
    my ($PaymentLine) = @_;

    my $hOrderInfo = $PaymentLine->get('OrderInfoJSON');
    my $hOrderPayload = $PaymentLine->get('OrderPayload');

    return 0 if ! $hOrderInfo || ! $hOrderPayload;
    #-- Check if the totals, currency, etc. are still valid for this basket

    my $PaymentMethod = $PaymentLine->get('PaymentMethod');
    my $CurrentPayLoad = $PaymentMethod->getOrderPayload($PaymentLine);

    #-- Data::Dumper modifies variable types (e.g. integer/string)! Be careful with this as there was an obscure interaction with the JSON module:
    #-- { 'a' => 1 }, the value of 'a' is an integer, but after logging it with Data::Dumper, it becomes { 'a' => '1' }
    #-- Technically, this is still a 1, and perl could care less, but JSON does!
    #-- This is especially important since we are comparing a JSON string below, and '{"a":1}' is different to '{"a":"1"}'

    #-- Basically, DON'T log these variables before comparing them
    my $Json = JSON->new;
    if ( $Json->canonical->encode($hOrderPayload) eq $Json->canonical->encode($CurrentPayLoad) ) {
        return 1;
    }

    return 0;
}

1;
