#========================================================================================
# §package      DE_EPAGES::Product::API::CSV::ProductImportHandler;
# §base         DE_EPAGES::ShopCSVExportImport::API::CSV::ShopObjectImportHandler
# §state        public
#----------------------------------------------------------------------------------------
# §description  This module implements a handler to import products
# §example      use DE_EPAGES::Product::API::CSV::ProductImportHandler;
#               DE_EPAGES::Product::API::CSV::ProductImportHandler->importCSV($FileName, $hFileFormat, $Shop);
#========================================================================================
package DE_EPAGES::Product::API::CSV::ProductImportHandler;
use base DE_EPAGES::ShopCSVExportImport::API::CSV::ShopObjectImportHandler;

use strict;
use HTTP::Request;
use DE_EPAGES::Core::API::CSV qw ( ParseCSV );
use DE_EPAGES::Core::API::File qw ( ExistsDirectory CreateDirectory fsplit IsFileNewer ExistsFile GetFileSize DeleteFile );
use DE_EPAGES::Core::API::Image qw ( TestImageFormat );
use DE_EPAGES::Core::API::Error qw ( ExistsError );
use DE_EPAGES::Core::API::Net qw ( GetContentLength GetFile );
use DE_EPAGES::Core::API::PerlTools qw( GenerateGUID unique ElementOf );
use DE_EPAGES::Core::API::Unicode qw ( Unicode2Filename );
use DE_EPAGES::Content::API::ShortUrl qw ( GetShortUrlByClass );
use DE_EPAGES::Object::API::Constants qw( POSITION_STEP );
use DE_EPAGES::Object::API::Factory qw( LoadObject LoadRootObject );
use DE_EPAGES::Presentation::API::Constants qw (MAX_IMAGE_UPLOAD_SIZE);
use DE_EPAGES::Product::API::Constants qw ( UNITY_ATTRIBUTES_FROM_MASTERPRODUCT_WRITTEN_ON_SUBS_ON_CREATE
                                            UNITY_NOT_UPDATE_ATTRIBUTES_ON_SUBS_ON_UPDATE_MASTER CONVERSION_FACTOR_BYTE );
use DE_EPAGES::Product::API::Product qw( ScaleImageSizes );
use DE_EPAGES::Product::API::Shop qw (GetProductImageSizes);
use DE_EPAGES::Core::API::Url qw ( StringIsURL );
use DE_EPAGES::Core::API::Warning qw( Warning );
use DE_EPAGES::WebInterface::API::Config qw ( WebInterfaceConfig );

use DE_EPAGES::Product::API::CSV::ProductAnalyzer;

# variable to hold warnings for unity
our @IMAGE_MISSING_WARNINGS =();
our @IMAGE_DOWNLOAD_WARNINGS =();

# indicates if we have an Unity import (identified by the headerline [UnityAlias] entry)
# see also function headerColumn()
our $IsUnityImport = undef;

#ERRORDEFINITION
# Be sure to add your warning in the cartridge's Constants.pm list!!!
sub ErrorDefinition {
    return {
        'ONE_OR_MORE_REFERENCED_IMAGE_MISSING' => 'One or more referenced images not found',
    };
}

#========================================================================================
# §function     analyzeAttributeLine
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       my $hAttributeValues = $self->analyzeAttributeLine();
#----------------------------------------------------------------------------------------
# §description  Analyzes the current line in the CSV file and fills the object data hash
#               with information for the actual import.
#               Uses $self->topID('Item') as input, which contains the list of values in
#               the current line.
#----------------------------------------------------------------------------------------
# §return       $hAttributeValues | properties of a object of CSV file with following keys
#               <ul>
#                 <li>DirectAttributes  - direct attributes of the object with following keys
#                   <ul>
#                       <li>Alias - alias of the object - string</li>
#                       <li>Class - class object of the object - object</li>
#                       <li>Parent - parent object of the object - object</li>
#                       <li> more direct attributes which should set via $Object->set method,
#                            or can used by $Class->insert($hAttributeValues->{'DirectAttributes'});</li>
#                   </ul> - ref.hash
#                 </li>
#                 <li>DoNotImport -  0 or 1, set to 1 if there are errors to set the parsed values - boolean</li>
#                 <li>Object - object if it is found or undef if it doesn't exist - object</li>
#                 <li>LocalizedAttributes -localized attributes hash with language id as key - ref.hash</li>
#                 <li>PriceAttributes - price attributes hash with CurrencyID/TaxModel as key e.g. ('EUR/1') -ref.hash</li>
#                 <li>VariationValueIDs - variation value ids which are used at the subproduct - ref.array</li>
#                 <li>PreDefMultiValueIDs - value ids for a multiple-choice option attribute - ref.hash.array.int</li>
#               </ul>
#               | ref.hash
# §example      hAttributeValues='HASH
#                   'DirectAttributes'  =>'HASH'
#                       'Alias' =>'js_3200705005'
#                       'AvailabilityDate'  => undef
#                       'Class' =>'DE_EPAGES::Product::API::Object::ProductType=HASH'
#                       'ImageLarge'        =>'js_3200705005.jpg'
#                       'ImageMedium'       =>'js_3200705005_m.jpg'
#                       'ImageMediumSmall'  =>'js_3200705005_ms.jpg'
#                       'ImageMediumLarge'  =>'js_3200705005_ml.jpg'
#                       'ImageSmall'        =>'js_3200705005_s.jpg'
#                       'IsAvailable'       =>'1'
#                       'IsBundleProduct'   =>'0'
#                       'IsDailyPrice'  =>'0'
#                       'IsDefault' =>'0'
#                       'IsNew' =>'0'
#                       'IsVisible' =>'1'
#                       'Length'    => undef
#                       'Manufacturer'  =>'Tatonka'
#                       'ManufacturerSKU'   => undef
#                       'MinOrder'  =>'1'
#                       'OrderUnit' =>'DE_EPAGES::Object::API::Object::UnitOfMeasurement=HASH'
#                       'Parent'    =>'DE_EPAGES::Product::API::Object::ProductFolder=HASH'
#                       'PriceQuantity' =>'1'
#                       ....
#                   'DoNotImport'   =>'0'
#                   'LocalizedAttributes'   =>'HASH'
#                       '1' =>'HASH'
#                           'Description'   =>'Wanderrucksack mit effektiver R?el?'
#                           'Name'  =>'Tatonka Vento'
#                       '2' =>'HASH'
#                           'Description'   =>'Hiking Backpack with effective back ventilation'
#                           'Name'  =>'Tatonka Vento'
#                   'Object'    =>'DE_EPAGES::Product::API::Object::Product=HASH'
#                   'PriceAttributes'   =>'HASH'
#                       'EUR/1' =>'HASH'
#                           'CurrencyID'    =>'EUR'
#                           'Price' =>'69.95'
#                           'TaxModel'  =>'1'
#                       'USD/1' =>'HASH'
#                           'CurrencyID'    =>'USD'
#                           'Price' =>'69.95'
#                           'TaxModel'  =>'1'
#                   'VariationValueIDs' =>'ARRAY'
#                           [0] =>  '3407'
#                           [1] =>  '3403'
#========================================================================================
sub analyzeAttributeLine {
    my $self = shift;

    my $aColumns = $self->topID('Item');

    # Singleton
    if (not defined $self->{'ProductAnalyzer'}) {
        $self->{'ProductAnalyzer'} = DE_EPAGES::Product::API::CSV::ProductAnalyzer->new( $self );
    }
    return $self->{'ProductAnalyzer'}->analyze( $aColumns );
}

#========================================================================================
# §function     modifyUsedAttributes
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->modifyUsedAttributes($hUsedAttributes, $ColumnIndex);
#----------------------------------------------------------------------------------------
# §description  add keys:
#                 $hUsedAttributes->{$ColumnIndex}{'IsSuperProductAliasColumn'} = $hUsedAttributes->{$ColumnIndex}{'Alias'}  eq 'SuperProduct'  ? 1 : 0;
#                 $hUsedAttributes->{'SuperProductAliasColumn'} with value=$ColumnIndex  if $hUsedAttributes->{$ColumnIndex}{'IsSuperProductAliasColumn'};
#----------------------------------------------------------------------------------------
# §input        $hUsedAttributes | used attributes | ref.hash
# §input        $ColumnIndex | current column | integer
#========================================================================================
sub modifyUsedAttributes {
    my $self = shift;
    my ( $hUsedAttributes, $ColumnIndex) = @_;
    $self->SUPER::modifyUsedAttributes($hUsedAttributes, $ColumnIndex);

    $hUsedAttributes->{$ColumnIndex}{'IsSuperProductAliasColumn'} = $hUsedAttributes->{$ColumnIndex}{'Alias'}  eq 'SuperProduct'  ? 1 : 0;
    $hUsedAttributes->{$ColumnIndex}{'IsEcoParticipationCategoryColumn'} = $hUsedAttributes->{$ColumnIndex}{'Alias'}  eq 'EcoParticipationCategory'  ? 1 : 0;
    $hUsedAttributes->{$ColumnIndex}{'IsIsDownloadProductColumn'} = $hUsedAttributes->{$ColumnIndex}{'Alias'}  eq 'IsDownloadProduct'  ? 1 : 0;
    $hUsedAttributes->{$ColumnIndex}{'IsIsBundleProductColumn'} = $hUsedAttributes->{$ColumnIndex}{'Alias'}  eq 'IsBundleProduct'  ? 1 : 0;
    $hUsedAttributes->{'SuperProductAliasColumn'}  = $ColumnIndex if $hUsedAttributes->{$ColumnIndex}{'IsSuperProductAliasColumn'};
    $hUsedAttributes->{'EcoParticipationCategoryColumn'}  = $ColumnIndex if $hUsedAttributes->{$ColumnIndex}{'IsEcoParticipationCategoryColumn'};
    $hUsedAttributes->{'IsDownloadProductColumn'}  = $ColumnIndex if $hUsedAttributes->{$ColumnIndex}{'IsIsDownloadProductColumn'};
    $hUsedAttributes->{'IsBundleProductColumn'}  = $ColumnIndex if $hUsedAttributes->{$ColumnIndex}{'IsIsBundleProductColumn'};
    return;
}

#========================================================================================
# §function     throwImageMissingWarningUnity
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->throwImageMissingWarningUnity($Object);
#----------------------------------------------------------------------------------------
# §description  Throws a general warning if a referenced image file is missing
#               A Warning is generated only once for the lifetime of the ProductImportHandler
#               and not on each call
#----------------------------------------------------------------------------------------
# §input        $Object | object which is imported | object
#========================================================================================
sub throwImageMissingWarningUnity {
    my $self = shift;
    my ( $Object ) = @_;

    Warning(
        'ONE_OR_MORE_REFERENCED_IMAGE_MISSING', {
            ObjectAlias   => $Object->get('Alias'),
            ObjectID      => $Object->id,
        }
    ) unless scalar(@IMAGE_MISSING_WARNINGS);

    push @IMAGE_MISSING_WARNINGS, $self->lineNumber;

    return;
}

#========================================================================================
# §function     throwImageDownloadWarningUnity
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->throwImageDownloadWarningUnity($Object);
#----------------------------------------------------------------------------------------
# §description  Throws a general warning if a referenced image is not downloadable
#               because of filesize to large or other request problems etc.
#               A Warning is generated only once for the lifetime of the ProductImportHandler
#               and not on each call
#----------------------------------------------------------------------------------------
# §input        $Object | object which is imported | object
#========================================================================================
sub throwImageDownloadWarningUnity {
    my $self = shift;
    my ( $Object ) = @_;

    Warning(
        'ONE_OR_MORE_REFERENCED_IMAGE_NOT_DOWNLOADABLE', {
            ObjectAlias   => $Object->get('Alias'),
            ObjectID      => $Object->id,
        }
    ) unless scalar(@IMAGE_DOWNLOAD_WARNINGS);

    push @IMAGE_DOWNLOAD_WARNINGS, $self->lineNumber;

    return;
}

#========================================================================================
# §function     resetFlagsAndWarnings
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->resetFlagsAndWarnings();
#----------------------------------------------------------------------------------------
# §description  reset the gloabl warnings counter.
#========================================================================================
sub resetFlagsAndWarnings {
    @IMAGE_MISSING_WARNINGS = ();
    @IMAGE_DOWNLOAD_WARNINGS = ();
    $IsUnityImport = undef;
}

#========================================================================================
# §function     baseAttributeClassAlias
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->baseAttributeClassAlias();
#----------------------------------------------------------------------------------------
# §description  return Base AttributeClass Alias, here : 'ProductClass'
#----------------------------------------------------------------------------------------
# §return        $BaseAttributeClassAlias | attribute class alias of normal attributes | string
#========================================================================================
sub baseAttributeClassAlias {
    my $self = shift;
    return 'ProductClass';
}

#========================================================================================
# §function     importSpecialData
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->importSpecialData($Object, $hAttributeValues);
#----------------------------------------------------------------------------------------
# §description  possibility to import special object data, called by importObject
#               saves Product Prices and super product SelectedVariations
#----------------------------------------------------------------------------------------
# §input        $Object | object which is imported | object
# §input        $hAttributeValues    | attribute values | ref.hash
#========================================================================================
sub importSpecialData {
    my $self = shift;
    my ( $Product, $hAttributeValues ) = @_;

    #product special
    # update prices
    foreach my $EntryKey (keys %{$hAttributeValues->{'PriceAttributes'}}) {
        my $Entry = $hAttributeValues->{'PriceAttributes'}{$EntryKey};

        my $Value = $Entry->{'Price'};
        $Value = undef if defined $Value and $Value !~ /\S/;

        if ($Entry->{'Alias'} eq 'ManufacturerPrices') {
            $Product->setManufacturerPrice( $Entry->{'CurrencyID'}, $Entry->{'TaxModel'}, $Value );
        } elsif ($Entry->{'Alias'} eq 'EcoParticipations') {
            $Product->setEcoParticipation($Entry->{'CurrencyID'}, $Entry->{'TaxModel'}, $Value );
        } elsif ($Entry->{'Alias'} eq 'Deposits') {
            $Product->setDeposit($Entry->{'CurrencyID'}, $Entry->{'TaxModel'}, $Value );
        } else {
            $Product->setListPrice($Entry->{'CurrencyID'}, $Entry->{'TaxModel'}, $Value );
        }
    }

    #update super product
    my $SuperProduct = $hAttributeValues->{'DirectAttributes'}{'SuperProduct'};
    if (defined $SuperProduct ) {
        my $VariationValueIDs = $hAttributeValues->{'VariationValueIDs'};
        my @ExistingSelectedVariations = map { $_->{'SelectedID'}}  @{$SuperProduct->get('SelectedVariations')};
        my @UniqueValues = unique (@ExistingSelectedVariations,@$VariationValueIDs);
        $SuperProduct->set( { 'SelectedVariations' => [map { {'SelectedID' => $_} } @UniqueValues ] } );
    }

    # update PreDefMulti attributes
    if( defined $hAttributeValues->{'PreDefMultiValueIDs'} ) {
        foreach my $Name ( keys %{$hAttributeValues->{'PreDefMultiValueIDs'}} ) {
            my @Values = map { LoadObject($_) } @{$hAttributeValues->{'PreDefMultiValueIDs'}->{$Name}};
            $Product->set({ $Name => \@Values });
        }
    }
    return;
}

#========================================================================================
# §function     headerColumn
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       my $hHeaderInfo = $self->headerColumn($HeaderString, $Index);
#----------------------------------------------------------------------------------------
# §description  Replaces attribute information for compatibility with import files from
#               previous ePages versions.
#========================================================================================
my %COMPAT_MAP = (
    'LongDescription' => 'Text',
    'ListPrice' => 'ListPrices',
    'ManufacturerPrice' => 'ManufacturerPrices',
    'EcoParticipation' => 'EcoParticipations',
    'Deposit' => 'Deposits',
    'UnityAlias' => 'Alias',
);

sub headerColumn {
    my $self = shift;
    my ($HeaderString, $Index) = @_;


    # test if we have a field [UnityAlias] in the headerline. If so we import unity export files.
    # If we have an unity import file, set global variable and replace [UnityAlias] by [ALias]]
    $IsUnityImport = 1 if $HeaderString =~ /\[UnityAlias\]/;

    # replace obsolete attribute aliases
    $HeaderString =~ s/\[\Q$_\E\b/\[$COMPAT_MAP{$_}/ foreach keys %COMPAT_MAP;

    return $self->SUPER::headerColumn( $HeaderString, $Index );
}

#========================================================================================
# §function     modifyObject
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $self->modifyObject($Product, $hOldAttributes, $hDirectAttributes, $hOldLocalizedAttributes, $hLocalizedAttributes );
#----------------------------------------------------------------------------------------
# §description  Modifies the object after its creation/modification by the import.
#               Called by importObject.
#----------------------------------------------------------------------------------------
# §input        $Product | object which is imported | object
# §input        $hOldAttributes | attributes of the object before the import.
#               May be undef if the product is created new and not modified.
#               | ref.hash or undefined
# §input        $hDirectAttributes | current Product's direct attribute values | ref.hash
# §input        $hOldLocalizedAttributes | localized attributes of the object before the import.
#               May be undef if the product is created new and not modified.
#               | ref.hash or undefined
# §input        $hLocalizedAttributes | current Product's localized attribute values | ref.hash
#----------------------------------------------------------------------------------------
# §return       -
#========================================================================================
sub modifyObject {
    my $self = shift;
    my ($Product, $hOldAttributes, $hDirectAttributes, $hOldLocalizedAttributes, $hLocalizedAttributes ) = @_;

    if($Product) {
        my $Shop = $self->object;
        my $IsUnity = $Shop->get('IsUnityRegistered');

        my @Attributes = (qw(ImageLarge ImageMedium ImageSmall ImageHotDeal));

        # for unity we import only ImageLarge and generate all other images from this image
        @Attributes = (qw(ImageLarge)) if $IsUnity;

        my $GeneratedAttributes = {};
        foreach my $AttributeName (@Attributes) {
            if(defined $hDirectAttributes->{$AttributeName}) {
                my @TargetAttributes;
                # make sure we don't overwrite attributes
                foreach my $TargetAttribute (@Attributes) {
                    unless($GeneratedAttributes->{$TargetAttribute} || defined($hDirectAttributes->{$TargetAttribute}) || $TargetAttribute eq $AttributeName) {
                        $GeneratedAttributes->{$TargetAttribute} = 1;
                        push(@TargetAttributes, $TargetAttribute);
                    }
                }

                # we generate this images for unity and ignore the values in the import file
                @TargetAttributes = (qw(ImageMedium ImageSmall ImageHotDeal ImageMediumSmall ImageMediumLarge)) if $IsUnity;

                $self->processImportedImageAttribute($Product, $hOldAttributes, $hDirectAttributes, $AttributeName, 1, undef, \@TargetAttributes);
             }
        }

        unless ($IsUnity){
            # generate MediumSmall and MediumLarge images
            foreach my $MediumSource (qw(ImageMedium ImageLarge ImageSmall ImageHotDeal)) {
                if(defined($hDirectAttributes->{$MediumSource})) {
                    $self->processImportedImageAttribute($Product, $hOldAttributes, $hDirectAttributes, $MediumSource, 1, undef, [qw(ImageMediumSmall ImageMediumLarge)]);
                    last;
                }
            }
        }

        if($hDirectAttributes->{'ImagesSlideShowString'} ) {
            $self->_processImagesSlideShowStringAttribute($Product, $hOldAttributes, $hDirectAttributes, 'ImagesSlideShowString');
        }

        # ony if we have not Unity we set the  short url if empty
        my $hLanguages = $self->languages;
        unless ($IsUnity) {
            my $Parent = $Product->parent;
            if ($Parent->id != $Shop->id) {
                my $Class = $Product->class;
                foreach my $LanguageID (keys %$hLanguages) {
                    my $URI = $Product->get('URI', $LanguageID);
                    if ((not defined $URI) || $URI eq '') {
                        $URI = GetShortUrlByClass($Parent, $Class, $Product->get('NameOrAlias', $LanguageID), $LanguageID);
                        $Product->set({'URI' => $URI}, $LanguageID) if defined $URI and $URI ne '';
                    }
                }
            }
        }

        # some checks for EPG-21172
        foreach my $LanguageID ( keys %$hLocalizedAttributes ) {
            my $hInfo = $hLocalizedAttributes->{$LanguageID};

            # instruction manual
            my $NewManual = $hInfo->{'InstructionManual'};
            if (defined $NewManual) {
                my $File = $Product->get('PublicPath') . "/$NewManual";
                unless (ExistsFile($File)) {
                    my $Manual = $hOldLocalizedAttributes->{$LanguageID}{'InstructionManual'};
                    $hInfo->{'InstructionManual'} = $Manual;
                    Warning(
                        'REFERENCED_FILE_MISSING', {
                            Line          => $self->lineNumber,
                            File          => $NewManual,
                            ObjectAlias   => $Product->get('Alias'),
                            ObjectID      => $Product->id,
                            AttributeName => 'InstructionManual'
                        }
                    );

                    $Product->set( $hInfo, $LanguageID );
                }
            }
        }

        # preview file
        my $OldPreviewFile = $hOldAttributes->{'Preview'};
        my $PreviewFile = $hDirectAttributes->{'Preview'};

        if (defined $PreviewFile) {
            my $File = $Product->get('PublicPath') . "/$PreviewFile";
            unless (ExistsFile($File)) {
                $hDirectAttributes->{'Preview'} = $OldPreviewFile;
                Warning(
                    'REFERENCED_FILE_MISSING', {
                        Line          => $self->lineNumber,
                        File          => $PreviewFile,
                        ObjectAlias   => $Product->get('Alias'),
                        ObjectID      => $Product->id,
                        AttributeName => 'Preview'
                    }
                );
                $Product->set( { 'Preview' => $OldPreviewFile } );
            }
        }

        # import language dependent attribute ProductDataSheet
        # create allow list for Unicode2Filename because we allow links for uploading content
        my $AllowedChars = ':/';
        foreach my $LanguageID ( keys %$hLocalizedAttributes ) {
            my $hInfo = $hLocalizedAttributes->{$LanguageID};
            foreach my $AttributeName (qw (_ProductDataSheet)) {

                # Remove possible whitespace in file name because the REST API does not like it.
                my $NewAttributeFileName = defined $hInfo->{$AttributeName} ?  Unicode2Filename($hInfo->{$AttributeName}, $AllowedChars) : $hInfo->{$AttributeName};
                $self->_importAttributeFile($Product, $AttributeName, $NewAttributeFileName, $LanguageID);
            }
        }

        # import attribute EnergyLabelSourceFile
        foreach my $AttributeName (qw(EnergyLabelSourceFile)) {
            if(defined $hDirectAttributes->{$AttributeName}) {;
                # Remove possible whitespace in file name because the REST API does not like it.
                my $NewAttributeFileName = defined $hDirectAttributes->{$AttributeName} ?  Unicode2Filename($hDirectAttributes->{$AttributeName}, $AllowedChars) : $hDirectAttributes->{$AttributeName};
                $self->_importAttributeFile($Product, $AttributeName, $NewAttributeFileName);
            }
        }

        # test if we have a energy efficency class > than the EnergyLabelHighestEfficiencyClass
        if(defined $hDirectAttributes->{'EnergyLabelsString'}) {
            my $aSystemLabels = ParseCSV(LoadRootObject()->get('EnergyLabelsString'));

            my $aParsed = ParseCSV($hDirectAttributes->{'EnergyLabelsString'});
            if (defined $hDirectAttributes->{'EnergyLabelHighestEfficiencyClass'}) {
                #test if EnergyLabelHighestEfficiencyClass < product class
                my $Highest = $hDirectAttributes->{'EnergyLabelHighestEfficiencyClass'};

                my $cnt = 0;
                my $labelidx = 0;
                my $maxclassidx = 0;

                foreach my $Value (@$aSystemLabels) {
                    $labelidx = $cnt if($Value eq $aParsed->[0]);
                    $maxclassidx = $cnt if($Value eq $Highest);
                    $cnt++;
                }

                my $Value;

                # highest is > than A, use A+++
                if ($maxclassidx > 3) {
                    $Value = $aSystemLabels->[0];
                } else {
                    $Value = $maxclassidx > $labelidx ? $aParsed->[0] : $Highest;
                }

                $Product->set({'EnergyLabelHighestEfficiencyClass' => $Value});

                #-- set EnergyLabelsString as this can be done just after the import of the source file
                $Product->set({ 'EnergyLabelsString' => $hDirectAttributes->{'EnergyLabelsString'} });
            }
        }


        # only for Unity test if we have all attribute values from the master set on the subproduct
        if($Product->get('SuperProduct') and $IsUnity){

            my $aAttributeNames;

            foreach my $AttributeName (UNITY_ATTRIBUTES_FROM_MASTERPRODUCT_WRITTEN_ON_SUBS_ON_CREATE) {
                # do not overwrite some attributes on subproducts...
                next if ElementOf($AttributeName,UNITY_NOT_UPDATE_ATTRIBUTES_ON_SUBS_ON_UPDATE_MASTER);
                push @$aAttributeNames,$AttributeName;
            }

            my $hAttributeValues = $Product->get('SuperProduct')->get($aAttributeNames);
            $Product->set($hAttributeValues);


            # set some special attributes
            my $hSpecialAttributes = {'HasSubOwnPrices'=> 1};
            $Product->set($hSpecialAttributes);
        }


        $self->SUPER::modifyObject($Product, $hOldAttributes, $hDirectAttributes);

        if ($IsUnity){
            foreach my $Image (qw(ImageMediumSmall ImageMediumLarge)){
                if(StringIsURL($Product->get($Image))){
                    $Product->set({$Image => undef});
                }
            }
            # if we have a unity shop and an epages6 import file we add the text from the epages6 LongDescription [Text]
            # field to the end of the description [Description] field separated by <br />
            unless($IsUnityImport){
                foreach my $LanguageID (keys %$hLanguages) {
                    my $ImportEpages6Text = $hLocalizedAttributes->{$LanguageID}->{'Text'};
                    next unless $ImportEpages6Text;
                    my $Description = $Product->get('Description', $LanguageID);
                    my $NewDescriptionPlusText = defined $Description ? $Description.'<br />'.$ImportEpages6Text : $ImportEpages6Text;
                    $Product->set({'Description' => $NewDescriptionPlusText}, $LanguageID);
                }
            }
        }
    } #end if product

    return;
}

sub _importAttributeFile {
    my $self = shift;
    my ($Product, $AttributeName, $NewAttributeValue, $LanguageID) = @_;

    my $Site              = $Product->getSite();
    my $IsUnity           = $Site->get('IsUnityRegistered');
    my $ProductPublicPath = $Product->get('PublicPath');
    my $MediaGallery      = $Site->get('MediaGallery');

    if ($IsUnity) {
        $MediaGallery = $Site->child('UnityMediaImportFiles');
        # for unity we test if we have an url image attribute
        # and download then the image from the given url
        $NewAttributeValue = $self->_processImageURL($Product, $Site, $NewAttributeValue);
    }

    # test if defined and not a URL
    if (defined $NewAttributeValue and not StringIsURL($NewAttributeValue)) {
        # correct slashes and spaces
        $NewAttributeValue = $self->cleanupAttribute( $NewAttributeValue );

        # test if file exists in Product public path
        my $FileNamePublic = $ProductPublicPath . "/$NewAttributeValue";

        unless (ExistsFile($FileNamePublic)) {
            # test if file exists in the media galery with the given path
            my (undef, $NewFileName, $NewFileExtension,$MGNewFileAbs) = $self->findImageFile($MediaGallery, $NewAttributeValue);
            # file was found in the MG
            if(defined $MGNewFileAbs and ExistsFile($MGNewFileAbs)){
                # copy file to product public path, if found in the media gallery
                # and correct new attribute if has a path and save only filename to attribute
                # base class handler does copyfile
                $self->SUPER::processMediaGalleryImage($Product, undef, $MGNewFileAbs, $NewFileName.$NewFileExtension);
                # in case we imported an URL string in a Unity shop we have to overwrite it with the actual file name
                my $CurrentAttributeValue = $Product->get($AttributeName,$LanguageID);
                my $UnityOverwriteURLAttribute = $IsUnity && StringIsURL($CurrentAttributeValue);
                # set attribute before if not already done
                $Product->set({$AttributeName => $NewFileName.$NewFileExtension },$LanguageID) if ($CurrentAttributeValue ne $NewFileName.$NewFileExtension || !$CurrentAttributeValue || $UnityOverwriteURLAttribute);
                # strip any paths from product image attributes (undo csv import changes)
                $self->fixObjectFileAttributes( $Product, [$AttributeName],$LanguageID);
            } else {
                # referenced file not found
                Warning(
                    'REFERENCED_FILE_MISSING', {
                        Line          => $self->lineNumber,
                        File          => $NewAttributeValue,
                        ObjectAlias   => $Product->get('Alias'),
                        ObjectID      => $Product->id,
                        AttributeName => $AttributeName
                    }
                );
            }
        }
    }

    return;
}

#========================================================================================
# §function     _processImagesSlideShowStringAttribute
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $self->_processImagesSlideShowStringAttribute(
#                    $Product, $hOldAttributes, $hDirectAttributes, $AttributeName);
#----------------------------------------------------------------------------------------
# §description  EPG-9416: Handler method to fetch product slideshow images from the
#               MediaGallery. Different implementation of base classes'
#               processImportedImageAttribute method.
#----------------------------------------------------------------------------------------
# §input        $Product | Product object which was imported | object
# §input        $hOldAttributes | attributes of the object before the import.
#               May be undef if the product is created new and not modified.
#               | ref.hash or undefined
# §input        $hDirectAttributes | current Product's direct attribute values | ref.hash
# §input        $AttributeName | the attribute to process | string
#========================================================================================
sub _processImagesSlideShowStringAttribute {
    my $self = shift;
    my ($Product, $hOldAttributes, $hDirectAttributes, $AttributeName) = @_;

    my $Site = $Product->getSite();
    my $ProductPublicPath = $Product->get('PublicPath');
    my $MediaGallery = $Site->get('MediaGallery');
    my $IsUnity = $Site->get('IsUnityRegistered');

    $MediaGallery = $Site->child('UnityMediaImportFiles') if $IsUnity;

    # do the same stuff as for ImageLarge, but allow multiple file names
    my @FileNames = split(/[,;]/, $hDirectAttributes->{$AttributeName});
    my $hNewImages;
    my $Position = 0;

    # for all images
    foreach my $SlideShowImage (@FileNames)
    {

        # for unity we test if we have an url image attribute
        # and download then the image from the given url
        $SlideShowImage = $self->_processImageURL($Product, $Site, $SlideShowImage) if $IsUnity;

        ##no critic qw(ProhibitDeepNests)
        if(!StringIsURL($SlideShowImage) )
        {
            $SlideShowImage = $self->cleanupAttribute($SlideShowImage);

            # ********* FIND NEW IMAGE FILE *********
            my (undef, $NewSlideShowImageFileName, $NewSlideShowImageExtension,
                $MGNewSlideShowImageAbs) = $self->findImageFile($MediaGallery, $SlideShowImage);

            # ********* CURRENT IMAGE FILE VALUE FOR THE OBJECT *********
            # look if particular image is in the old list
            my $OldSlideShowImage = undef;
            my $FullSlideShowFileName = $NewSlideShowImageFileName . $NewSlideShowImageExtension;
            if(defined $FullSlideShowFileName and
               defined $hOldAttributes and
               defined $hOldAttributes->{$AttributeName} and
               $hOldAttributes->{$AttributeName} =~ /\Q$FullSlideShowFileName\E/x) {
               $OldSlideShowImage = $NewSlideShowImageFileName.$NewSlideShowImageExtension;
            }

            if(defined $MGNewSlideShowImageAbs and ExistsFile($MGNewSlideShowImageAbs))
            {
                # ********* CASE 1: NEW IMAGE FILE EXISTS *********
                # check if product already has that "ImagesSlideShowString" image
                if(not defined $OldSlideShowImage)
                {
                    # ********* CASE 1.1: NO OLD IMAGE FILE *********
                    if (_TestImageFormat($FullSlideShowFileName)) {
                        $self->processMediaGalleryImage(
                            $Product,
                            $hDirectAttributes,
                            $MGNewSlideShowImageAbs,
                            $FullSlideShowFileName,
                            1,      #automatic generation of other image sizes
                            undef,  #no scale values supplied
                            1       #ignore update product attributes for slideshow
                        );
                    }
                }
                else
                {
                    if($OldSlideShowImage eq $FullSlideShowFileName)
                    {
                        # ********* CASE 1.2: OLD IMAGE FILE WITH SAME NAME AS NEW IMAGE FILE *********
                        if(not ExistsFile($ProductPublicPath . '/' . $OldSlideShowImage) or
                            IsFileNewer($ProductPublicPath . '/' . $OldSlideShowImage, $MGNewSlideShowImageAbs))
                        {
                            # is newer, so continue processing.
                            if (_TestImageFormat($FullSlideShowFileName)) {
                                $self->processMediaGalleryImage(
                                    $Product,
                                    $hDirectAttributes,
                                    $MGNewSlideShowImageAbs,
                                    $FullSlideShowFileName,
                                    1,      #automatic generation of other image sizes
                                    undef,  #no scale values supplied
                                    1       #ignore update product attributes for slideshow
                                );
                            }
                        }
                        else
                        {
                            # ********* CASE 1.2.2: NEW FILE IS NOT NEWER *********
                        }
                    }
                    else
                    {
                        # ********* CASE 1.3: NEW IMAGE FILE HAS OTHER NAME AS OLD IMAGE FILE *********
                        # replace old image file
                        if (_TestImageFormat($FullSlideShowFileName)) {
                            $self->processMediaGalleryImage(
                                $Product,
                                $hDirectAttributes,
                                $MGNewSlideShowImageAbs,
                                $FullSlideShowFileName,
                                1,      #automatic generation of other image sizes
                                undef,  #no scale values supplied
                                1       #ignore update product attributes for slideshow
                            );
                        }
                    }
                }
                $hNewImages->{$FullSlideShowFileName} = $Position;
            }
            else
            {
                # ********* CASE 2: NEW IMAGE FILE NOT FOUND IN MEDIAGALLERY *********
                if(ExistsFile($ProductPublicPath . '/' . $FullSlideShowFileName))
                {
                    # ********* CASE 2.1: IMAGE FOUND IN PRODUCT IMAGE GALLERY *********
                    $hNewImages->{$FullSlideShowFileName} = $Position;
                }
                else
                {
                    # ********* CASE 2.2: NO IMAGE FOUND AT ALL **********
                    $IsUnity ? $self->throwImageMissingWarningUnity($Product) :
                               $self->throwImageMissingWarning($Product, $SlideShowImage, $AttributeName);
                }
            } # if ExistsFile
        }
        else # isURL
        {
            # just add the URL entry
            $hNewImages->{$SlideShowImage} = $Position unless $IsUnity ;
        }

        $Position+=POSITION_STEP;
        ##use critic
    } # Foreach FileNames

    # update Product if changed, the Product's ImagesSlideShow Hash is rebuilt automatically
    if(scalar(keys %$hNewImages) > 0) {
        my @SortedImages = sort { $hNewImages->{$a} cmp $hNewImages->{$b} } keys %$hNewImages;
        my $NewImagesString = join(';', @SortedImages);
        $Product->set( { $AttributeName => $NewImagesString } );
    } else {
        # test if images in ImagesSlideShowString exists before we set the ImagesSlideShowString
        my $aExistingImages = [];
        foreach my $ImageName (@FileNames){
            push @$aExistingImages, $ImageName if ExistsFile($ProductPublicPath . '/' . $ImageName);
        }
        $Product->set( { $AttributeName => join(';', @$aExistingImages) });
    }
}

#========================================================================================
# §function     processMediaGalleryImage
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $hNewAttributes = $self->processMediaGalleryImage(
#                   $Object,
#                   $hDirectAttributes,
#                   $SourceFileName,
#                   $TargetFileName
#               );
# §example      $hDirectAttributes = $self->processMediaGalleryImage(
#                   $Object,
#                   $hDirectAttributes,
#                   $MGNewImageLargeAbs,
#                   $Product->get('PublicPath') . '/' . $NewImageLargeFileName.$NewImageLargeExtension
#               );
#----------------------------------------------------------------------------------------
# §description  Copies the product image from the given source and generates its
#               scaled variations. Updates the given attribute hash with the image names.
#----------------------------------------------------------------------------------------
# §input        $Object | Category object which was imported | object
# §input        $hDirectAttributes  | direct attribute values | ref.hash
# §input        $SourceFileName     | full file path and name of the source image | string
# §input        $TargetFileName     | full file path and name for the target image | string
# §input        $ScaleImages | boolean value if automatic generation of other image sizes
#               should be applied.
# §input        $hScaleDimension | Width and Height for the scaled image. unused here. |
#               ref.hash
# §input        $DoNotUpdateAttributes | boolean value whether product attributes should be
#               updated | boolean
# §input        $TargetAttributes | additional attributes to generate images for | ref.array.string
#========================================================================================
sub processMediaGalleryImage {
    my $self = shift;
    my ($Object, $hDirectAttributes, $SourceFileName, $TargetFileName,
        $ScaleImages, $hScaleDimension, $DoNotUpdateAttributes, $TargetAttributes) = @_;

    $TargetAttributes //= [qw(ImageSmall ImageMedium ImageMediumSmall ImageMediumLarge ImageHotDeal)];

    # base class handler does only copyfile from source to target
    $self->SUPER::processMediaGalleryImage($Object, $hDirectAttributes, $SourceFileName, $TargetFileName);

    if($ScaleImages) {
        # Process scale images

        # Names
        my $hParams = {};
        $hParams->{'SCALE_NAME_m'}  = 'ImageMedium'      if(ElementOf('ImageMedium', @$TargetAttributes));
        $hParams->{'SCALE_NAME_ms'} = 'ImageMediumSmall' if(ElementOf('ImageMediumSmall', @$TargetAttributes));
        $hParams->{'SCALE_NAME_ml'} = 'ImageMediumLarge' if(ElementOf('ImageMediumLarge', @$TargetAttributes));
        $hParams->{'SCALE_NAME_s'}  = 'ImageSmall'       if(ElementOf('ImageSmall', @$TargetAttributes));
        $hParams->{'SCALE_NAME_h'}  = 'ImageHotDeal'     if(ElementOf('ImageHotDeal', @$TargetAttributes));
        $hParams->{'SCALE_NAME_xs'} = 'ImageThumbnail';

        # Sizes from Constants
        my $hProductImageSizes = GetProductImageSizes($Object->getSite());
        foreach my $ScaleKey (keys %$hProductImageSizes) {
            $hParams->{'SCALE_WIDTH_'.$ScaleKey} = $hProductImageSizes->{$ScaleKey}->[0];
            $hParams->{'SCALE_HEIGHT_'.$ScaleKey} = $hProductImageSizes->{$ScaleKey}->[1];
        }

        my $hImageFiles = ScaleImageSizes($Object->get('PublicPath') . '/' . $TargetFileName, $hParams);

        # update product
        if(not $DoNotUpdateAttributes) {
            foreach my $AttString (@$TargetAttributes) {
                $Object->set( { $AttString => $hImageFiles->{$AttString} } ) if(defined($hImageFiles->{$AttString}));
            }
        }
    }
}

#========================================================================================
# §function     processImportedImageAttribute
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->processImportedImageAttribute(
#                   $Object, $hOldAttributes, $hDirectAttributes,
#                   $AttributeName, $ScaleImages
#               );
# §example      $self->processImportedImageAttribute(
#                   $Product, $hOldAttributes, $hDirectAttributes, 'ImageSmall', 0
#               );
#----------------------------------------------------------------------------------------
# §description  EPG-9416: Handler method to fetch object images from the MediaGallery.
#----------------------------------------------------------------------------------------
# §input        $Object | Product object which was imported | object
# §input        $hOldAttributes | attributes of the object before the import.
#               May be undef if the product is created new and not modified.
#               | ref.hash or undefined
# §input        $hDirectAttributes | current Category's direct attribute values | ref.hash
# §input        $AttributeName | Name of the attribute | string
# §input        $ScaleImages | boolean value if automatic generation of other image sizes
#               should be applied. The parameter is just supplied to the
#               processMediaGalleryImage method, which implemention may differ in derived
#               classes.
# §input        $hScaleDimension | Width and Height for the scaled image | ref.hash
# §input        $TargetAttributes | additional attributes to generate images for | ref.array.string
#========================================================================================
sub processImportedImageAttribute {
    my $self = shift;
    my ( $Object, $hOldAttributes, $hDirectAttributes, $AttributeName, $ScaleImages, $hScaleDimension, $TargetAttributes ) = @_;

    my $Site             = $Object->getSite();
    my $IsUnity          = $Site->get('IsUnityRegistered');

    # for unity we test if we have an url image attribute
    # and download then the image from the given url
    $hDirectAttributes->{$AttributeName} = $self->_processImageURL($Object, $Site, $hDirectAttributes->{$AttributeName}) if $IsUnity;

    ##no critic qw(ProhibitDeepNests)
    if ( !StringIsURL( $hDirectAttributes->{$AttributeName} ) ) {
        my $ObjectPublicPath = $Object->get('PublicPath');
        my $MediaGallery     = $Site->get('MediaGallery');

        $MediaGallery = $Site->child('UnityMediaImportFiles') if $IsUnity;

        # ********* CLEANUP ATTRIBUTE STRING *********
        my $NewImageLarge = $self->cleanupAttribute( $hDirectAttributes->{$AttributeName} );

        # ********* FIND NEW IMAGE FILE *********
        my ( undef, $NewImageLargeFileName, $NewImageLargeExtension, $MGNewImageLargeAbs )
            = $self->findImageFile( $MediaGallery, $NewImageLarge );
        my $FullNewImageLargeFileName = $NewImageLargeFileName . $NewImageLargeExtension;

        # ********* CURRENT IMAGE FILE VALUE FOR THE OBJECT *********
        my $OldImageFile = defined $hOldAttributes ? $hOldAttributes->{$AttributeName} : undef;

        if ( defined $MGNewImageLargeAbs and ExistsFile($MGNewImageLargeAbs) ) {

            # ********* CASE 1: NEW IMAGE FILE EXISTS *********
            if ( not defined $OldImageFile ) {
                # ********* CASE 1.1: NO OLD IMAGE FILE *********
                if (_TestImageFormat($FullNewImageLargeFileName)) {
                    $self->processMediaGalleryImage( $Object, $hDirectAttributes, $MGNewImageLargeAbs,
                        $FullNewImageLargeFileName,
                        $ScaleImages, $hScaleDimension,
                        0, $TargetAttributes );
                    $Object->set( { $AttributeName => $FullNewImageLargeFileName } );
                }
            }
            else {
                if ( $OldImageFile eq $FullNewImageLargeFileName ) {

                    # ********* CASE 1.2: OLD IMAGE FILE WITH SAME NAME AS NEW IMAGE FILE *********
                    if ( not ExistsFile( $ObjectPublicPath . '/' . $OldImageFile )
                        or IsFileNewer( $ObjectPublicPath . '/' . $OldImageFile, $MGNewImageLargeAbs ) ) {

                        # ********* CASE 1.2.1: OLD FILE MISSING OR NEW FILE IS NEWER *********
                        # file is new, so continue processing
                        if (_TestImageFormat($FullNewImageLargeFileName)) {
                            $self->processMediaGalleryImage( $Object, $hDirectAttributes, $MGNewImageLargeAbs,
                                $FullNewImageLargeFileName,
                                $ScaleImages, $hScaleDimension,
                                0, $TargetAttributes);
                        }
                    }
                    else {

                        # ********* CASE 1.2.2: NEW FILE IS NOT NEWER *********
                        # file is not newer, so keep the current one
                        $Object->set( { $AttributeName => $NewImageLargeFileName . $NewImageLargeExtension } );

                        foreach my $Attribute (@$TargetAttributes) {
                            $Object->set( { $Attribute => $hOldAttributes->{$Attribute} }) if $hOldAttributes->{$Attribute};
                        }
                    }
                }
                else {

                    # ********* CASE 1.3: NEW IMAGE FILE HAS OTHER NAME AS OLD IMAGE FILE *********
                    # replace old image file
                    if (_TestImageFormat($FullNewImageLargeFileName)) {
                        $self->processMediaGalleryImage( $Object, $hDirectAttributes, $MGNewImageLargeAbs,
                            $FullNewImageLargeFileName,
                            $ScaleImages, $hScaleDimension,
                            0, $TargetAttributes );
                        $Object->set( { $AttributeName => $FullNewImageLargeFileName } );
                    }
                }
            }
            # replace url with filename because the url was set before
            if($IsUnity and $AttributeName eq 'ImageLarge'){
                my $OldAttributeValue = $Object->get($AttributeName);
                $Object->set({'ImagesSlideShowString' => undef});
                 if( defined $OldAttributeValue and StringIsURL($OldAttributeValue)){
                    $Object->set({$AttributeName => $hDirectAttributes->{$AttributeName}});
                }
            }
            # strip any paths from product image attributes (undo csv import changes)
            $self->fixObjectFileAttributes( $Object, [$AttributeName] );
        }
        else {
            # ********* CASE 2: NEW IMAGE FILE NOT FOUND *********
            if ( !ExistsFile( $ObjectPublicPath . '/' . $FullNewImageLargeFileName ) ) {
                # ********* CASE 2.1: NO OLD IMAGE FILE OR OLD IMAGE FILE DOES NOT EXIST *********
                    $IsUnity ? $self->throwImageMissingWarningUnity($Object) :
                               $self->throwImageMissingWarning($Object, $hDirectAttributes->{$AttributeName}, $AttributeName);

                # restore old images due to EPG-11767
                $self->_restoreOldAttributeValues($Object, $hOldAttributes)
            }
        }
    } else {
        # isURL and Unity
        $self->_restoreOldAttributeValues($Object, $hOldAttributes) if $IsUnity;
    }

    return;
}

# for unity we test if we have an url image attribute
# and download then the image from the given url
# afterwards we use only the image name as attribute and
# process a normal image from the unity media gallery
sub _processImageURL {
    my $self = shift;
    my ($Product, $Site, $FileURL) = @_;

    return $FileURL unless StringIsURL($FileURL);

    # Storage for downloaded images
    my $MediaGallery = $Site->child('UnityMediaImportFiles');

    # use time for creating a directory
    my $TempDirName = GenerateGUID();
    my $StoragePath = $MediaGallery->get('PublicPath')."/". $TempDirName ;

    # create directory for downloads and give a warning if already exists
    if (ExistsDirectory($StoragePath)) {
        $self->throwImageDownloadWarningUnity();
    } else{
        CreateDirectory($StoragePath);
    }

    # get file from URL and store to MediaGallery
    my (undef, $FileName, $Extension) = fsplit($FileURL);
    my $StoragePathFile = $StoragePath."/".$FileName.$Extension;

    my $hResult = GetContentLength($FileURL);

    my $ContentLength = $hResult->{'ContentLength'};
    my $FileDownloaded = 0;
    if ( ! defined $ContentLength ) {
        my $ResponseCode = GetFile($FileURL, $StoragePathFile);

        if ( $ResponseCode == 0 ) {
            #-- Read the file and get its file size
            if ( ExistsFile($StoragePathFile) ) {
              $ContentLength = GetFileSize($StoragePathFile);
              $FileDownloaded = 1;
            }
        }
    }

    if (defined $ContentLength ){
        # in kB !
        my $MaxFileSize = MAX_IMAGE_UPLOAD_SIZE // undef;
        $MaxFileSize = WebInterfaceConfig()->get('GLOBAL', 'MAXCONTENTLENGTH') unless defined $MaxFileSize;
        # test size

        if(($MaxFileSize * CONVERSION_FACTOR_BYTE) < $ContentLength) {
            $self->throwImageDownloadWarningUnity($Product);
            DeleteFile($StoragePathFile) if $FileDownloaded;
            return $FileURL;
        }

        return $TempDirName."/".$FileName.$Extension if $FileDownloaded;

        my $ResponseCode = GetFile($FileURL, $StoragePathFile);
        # give back the file name for normal media gallery handling
        return $TempDirName."/".$FileName.$Extension unless $ResponseCode;

        # we have a $ResponseCode != 0 give the url back
        $self->throwImageDownloadWarningUnity($Product);
        return $FileURL;

    } else {
        # error contentlength
        $self->throwImageDownloadWarningUnity($Product);
        return $FileURL;
    }
}

# restore old images due to EPG-11767
sub _restoreOldAttributeValues {
    my $self = shift;
    my ($Object, $hOldAttributes) = @_;

    # restore old images due to EPG-11767
    my %RestoreAttributeValues;
    for my $Attribute ( qw(ImageLarge ImageMedium ImageMediumSmall ImageMediumLarge ImageSmall ImageHotDeal) ) {
        my $Old = $hOldAttributes->{$Attribute};
        $RestoreAttributeValues{$Attribute} = $Old;
    }
    $Object->set( \%RestoreAttributeValues );

    return;
}

#========================================================================================
sub _TestImageFormat {
    my ($FileName) = @_;

    # returns true if image file extension is safe
    eval { TestImageFormat('TargetFile', $FileName) };
    return ExistsError ? 0 : 1;
}

1;
