#========================================================================================
# §package      DE_EPAGES::Product::API::ProductPatch
# §state        public
#----------------------------------------------------------------------------------------
# §description  Implements JSON-PATCH (RFC 6902) for the product object.
#========================================================================================
package DE_EPAGES::Product::API::ProductPatch;
use base DE_EPAGES::WebService::API::JSONPatch;

use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::PerlTools qw (trim JsonTrue);
use DE_EPAGES::Core::API::TestParameter qw (TestString TestInteger TestFloat TestArray TestHash);
use DE_EPAGES::Object::API::Factory qw ( LoadObject LoadRootObject ExistsObjectByGUID );
use DE_EPAGES::Object::DAL::UnitOfMeasurement qw ( DAL_GetInfoUnitOfMeasurementByAbbreviation );
use DE_EPAGES::PriceList::API::PriceList qw (GetUnityPriceList); ## no critic (RequireCorrectImports)
use DE_EPAGES::Product::API::Product qw (TestUPCEAN);
use DE_EPAGES::Product::API::ProductImageService;
use DE_EPAGES::Shop::API::ShopTaxClass;
use DE_EPAGES::XML::API::DateTimeFormatXML;
use DE_EPAGES::Core::API::File qw ( WriteFile );
use DE_EPAGES::Core::API::Unicode qw ( Unicode2Filename );

use MIME::Base64 qw ( decode_base64 );
use List::Util qw (reduce);
use LWP::UserAgent;
use strict;

## no critic (ProhibitUnusedPrivateSubroutines ProhibitLongLines ProhibitMagicNumbers)

#========================================================================================
# §function     new
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       new( $Product );
#----------------------------------------------------------------------------------------
# §description  Constructor
#----------------------------------------------------------------------------------------
# §input        $Product | A Product | DE_EPAGES::Product::API::Object::Product
# §input        $LanguageId | A LanguageID | integer
# §input        $CurrencyId | A CurrencyId | integer
# §return       $self | self | DE_EPAGES::Product::API::ProductPatch
#========================================================================================
sub new {
    my $class = shift;
    my ($Product, $LanguageId, $CurrencyId) = @_;
    my $self  = bless {
        'Product' => $Product,
        'LanguageId' => $LanguageId,
        'CurrencyId' => $CurrencyId,
    }, $class;
    return $self;
}

sub _add { ##no critic qw(Subroutines::ProhibitExcessComplexity)
    my $self = shift;
    my ($Path, $Value) = @_;

    my $Product = $self->{'Product'};
    my $LanguageId = $self->{'LanguageId'};
    my $CurrencyId = $self->{'CurrencyId'};

    if ($Path eq '/productNumber') {
        $Product->set({ 'Alias' => $Value }, $LanguageId);
        return;
    }

    if ($Path eq '/name') {
        $Product->set({ 'Name' => $Value }, $LanguageId);
        return;
    }

    if ($Path eq '/visible') {
        $Product->set({ 'IsVisible' => JsonTrue($Value) });
        return;
    }

    if ($Path eq '/shortDescription') {
        $Product->set({ 'Description' => $Value }, $LanguageId);
        return;
    }

    if ($Path eq '/description') {
        $Product->set({ 'Text' => $Value }, $LanguageId);
        return;
    }

    if ($Path eq '/deliveryPeriod') {
        $Product->set({ 'DeliveryPeriod' => $Value });
        return;
    }

    if ($Path eq '/manufacturer') {
        $Product->set({ 'Manufacturer' => $Value });
        return;
    }

    if ($Path eq '/upc' || $Path eq '/ean') {
        TestUPCEAN('Value', $Value);
        $Product->set({ 'UPCEAN' => $Value });
        return;
    }

    if ($Path eq '/productCode') {
        TestString('Value', $Value);
        $Product->set({'UPCEAN' => $Value});
        return;
    }

    if ($Path eq '/essentialFeatures') {
        $Product->set({ 'MainCharacteristics' => $Value }, $LanguageId);
        return;
    }

    if ($Path eq '/searchKeywords') {
        TestArray('Value', $Value);
        my $Keywords = $self->_convertKeywordsArrayToString($Value);
        $Product->set({ 'Keywords' => $Keywords }, $LanguageId);
        return;
    }

    if ($Path eq '/priceInfo/price/amount') {
        my $Shop = $Product->getSite();
        my $TaxModel = $Shop->get('TaxModel');
        $Product->setListPrice($CurrencyId, $TaxModel, $Value);
        return;
    }

    if ($Path eq '/priceInfo/manufacturerPrice/amount') {
        my $Shop = $Product->getSite();
        my $TaxModel = $Shop->get('TaxModel');
        $Product->setManufacturerPrice($CurrencyId, $TaxModel, $Value);
        return;
    }

    if ($Path eq '/priceInfo/depositPrice/amount') {
        my $Shop = $Product->getSite();
        my $TaxModel = $Shop->get('TaxModel');
        $Product->setDeposit($CurrencyId, $TaxModel, $Value);
        return;
    }

    if ($Path eq '/productImage') {
        my $ProductImageService = $self->_createProductImageService($Product);
        $ProductImageService->setSlideShowImageAsProductImage($Value);
        return;
    }

    if ($Path eq '/stocklevel') {
        TestFloat('Value', $Value);
        my $CurrentStockLevel = $Product->get('StockLevel');
        if (not defined $CurrentStockLevel) {
            $CurrentStockLevel = 0;
        }
        my $NewStockLevel = $CurrentStockLevel + $Value;
        $Product->set({ 'StockLevel' => $NewStockLevel });
        return;
    }

    if ($Path eq '/minStocklevel') {
        TestFloat('Value', $Value);
        $Product->set({ 'StockLevelAlert' => $Value });
        return;
    }

    if ($Path eq '/manufacturerProductNumber') {
        $Product->set({ 'ManufacturerSKU' => $Value });
        return;
    }

    if ($Path eq '/productLength') {
        TestFloat('Value', $Value);
        $Product->set({ 'Length' => $Value });
        return;
    }

    if ($Path eq '/productWidth') {
        TestFloat('Value', $Value);
        $Product->set({ 'Width' => $Value });
        return;
    }

    if ($Path eq '/productHeight') {
        TestFloat('Value', $Value);
        $Product->set({ 'Height' => $Value });
        return;
    }

    if ($Path eq '/deliveryWeight') {
        return unless (
            defined($Value) &&
                defined($Value->{'amount'}) &&
                defined($Value->{'unit'})
        );
        my $aUnitOfMeasurementInfo = DAL_GetInfoUnitOfMeasurementByAbbreviation($Value->{'unit'});
        Error("delivery weight unit is not valid: $Value->{'unit'}")
            if not scalar @{$aUnitOfMeasurementInfo} or $aUnitOfMeasurementInfo->[0]->{'Type'} ne 'mass';

        $Product->set({ 'Weight' => $Value->{'amount'}, 'WeightUnit' => LoadObject($aUnitOfMeasurementInfo->[0]->{'UnitID'}) });
        return;
    }

    if ($Path eq '/energyLabelInfo') {
        TestHash('Value', $Value);

        my $FileName = $Value->{'file'}->{'fileName'};
        $Product->set({
            'EnergyLabelHighestEfficiencyClass' => $Value->{'EnergyLabelHighestEfficiencyClass'},
            'EnergyLabels' => $Value->{'EnergyLabels'},
            'EnergyLabelSourceFile' => $FileName,
        });

        my $ProductImageService = $self->_createProductImageService($Product);
        my $Image = decode_base64($Value->{'file'}->{'contents'});

        $ProductImageService->uploadAndScaleImage($FileName, $Image);

        return;
    }

    # set or replace customer specific prices
    if ($Path eq '/customerSpecificPriceInfo') {
        TestArray('Value', $Value);

        #1st test structure and then insert afterwards
        foreach my $hEntry (@$Value) {
            TestString('customerGroup', $hEntry->{'customerGroup'});
            TestArray('priceInfo', $hEntry->{'priceInfo'});
            foreach my $hPriceInfo (@{$hEntry->{'priceInfo'}}) {
                TestInteger('quantity', $hPriceInfo->{'quantity'});
                TestFloat('price', $hPriceInfo->{'price'});
                TestFloat('discountPercentage', $hPriceInfo->{'discountPercentage'}) if $hPriceInfo->{'discountPercentage'};
            }
        }

        # test if we have already a customer group for adding values
        my $Shop = $Product->getSite();
        my $InputTaxModel = $Shop->get('InputTaxModel');
        foreach my $hEntry (@$Value) {
            # add/replace only if we have a existing customer group!
            if ($Shop->folder('Groups')->existsChild($hEntry->{'customerGroup'})) {
                my $KG = $Shop->folder('Groups')->child($hEntry->{'customerGroup'});
                my $CustomerGroupPriceList = @{$KG->get('PriceLists')}[0];
                my $ahScalePricesAndDiscounts = [];
                push @$ahScalePricesAndDiscounts, { 'Price' => $_->{'price'}, 'Discount' => $_->{'discountPercentage'} // 0, 'Quantity' => $_->{'quantity'} } foreach @{$hEntry->{'priceInfo'}};
                $CustomerGroupPriceList->setScalePrices($Product, $InputTaxModel, $ahScalePricesAndDiscounts);
                $CustomerGroupPriceList->setQuantityDiscounts($Product, $InputTaxModel, $ahScalePricesAndDiscounts);
            }
            else {
                Error('customer group not exists: ' . $hEntry->{'customerGroup'});
            }
        }

        return;

    }

    if($Path eq '/isNew') {
        $Product->set({'IsNew' => JsonTrue($Value)});
        return;
    }

    #-- YYYY-MM-DDTHH:MM::SSZ
    if($Path eq '/newnessDate') {
        TestString('Value', $Value);
        my $DF = DE_EPAGES::XML::API::DateTimeFormatXML->new;
        my $Dt = $DF->parse_datetime($Value);
        $Dt->set(hour=>0,minute=>00,second=>00);
        $Product->set({'NewnessDate' => $Dt});
        return;
    }

    #-- { "type":"YouTube", "id":"xyz" }
    if($Path eq '/video') {
        TestHash('Value', $Value);
        my $Type = $Value->{'type'};
        if ( $Type ne 'YouTube' && $Type ne 'Vimeo' ) {
          Error('Unsupported video type: ' .  $Type);
        }

        my $ID = $Value->{'id'};
        $ID =~ s/\s//g;
        $ID =~ s/"//g;
        if ( !$ID || $ID eq '' ) {
          Error('Video id must be provided');
        }

        $Product->set({
          'ProductVideoProvider' => "[{\"VideoType\":\"$Type\",\"VideoID\":\"$ID\"}]",
        }, $LanguageId);

        return;
    }

    if($Path eq '/isManufacturerPriceRRP') {
        $Product->set({'IsManufacturerPriceRRP' => JsonTrue($Value)});
        return;
    }

    #-- { "refAmount":1, "refUnit":"piece", "refContentAmount":2 }
    if($Path eq '/referencePriceInfo') {
        TestHash('Value', $Value);
        if ( !defined $Value->{'refAmount'} || !defined $Value->{'refUnit'} || !defined $Value->{'refContentAmount'} ) {
          Error('refAmount, refUnit, refContentAmount are all required fields');
        }

        #-- The Min/Max values as enforced by the MBO. Actually, the MBO supports 6dp, but we run into precision issues with 6dp.
        my $refAmount = $Value->{'refAmount'};
        my $refUnit = $Value->{'refUnit'};
        my $refContentAmount = $Value->{'refContentAmount'};
        TestInteger('refAmount', $refAmount, 1, 2099999999);
        TestString('refUnit', $refUnit);
        TestFloat('refContentAmount', $refContentAmount, 0.000001, 2099999999.99999);

        my $aUnits = LoadRootObject()->get('Units');
        my $RefUnitObject;
        foreach ( @$aUnits ) {
          if ( $_->alias eq $refUnit ) {
            $RefUnitObject = $_ ;
            last;
          }
        }

        Error("Invalid reference unit $refUnit") unless $RefUnitObject;

        $Product->set({
          'RefAmount' => $refAmount,
          'RefUnit' => $RefUnitObject,
          'RefContentAmount' => int($refContentAmount*100000)/100000, #-- truncate to 5dp, NOT round
        });

        return;
    }

    if($Path eq '/isShippingLinkVisible') {
        $Product->set({'IsShippingLinkVisible' => JsonTrue($Value)});
        return;
    }

    #-- { "isCustomizable":true, "customizableTextLength":50, "customizableHeadlineText":"Ex" }
    if($Path eq '/customisationInfo') {
        TestHash('Value', $Value);
        if ( !defined $Value->{'isCustomizable'} || !defined $Value->{'customizableTextLength'} || !defined $Value->{'customizableHeadlineText'} ) {
          Error('isCustomizable, customizableTextLength, customizableHeadlineText are all required fields');
        }

        my $isCustomizable = JsonTrue($Value->{'isCustomizable'});
        my $customizableTextLength = $Value->{'customizableTextLength'};
        my $customizableHeadlineText = $Value->{'customizableHeadlineText'};
        TestInteger('isCustomizable', $isCustomizable);
        TestInteger('customizableTextLength', $customizableTextLength);
        TestString('customizableHeadlineText', $customizableHeadlineText);

        $Product->set({
          'IsCustomizable' => $isCustomizable,
          'CustomizableTextLength' => $customizableTextLength,
          'CustomizableHeadlineText' => $customizableHeadlineText,
        }, $LanguageId);

        return;
    }

    if($Path eq '/title') {
        TestString('title', $Value);
        $Product->set({'Title' => $Value}, $LanguageId);
        return;
    }

    #-- { "fileName":"abc.pdf", "url":"https://www.example.com/file.pdf" }
    if($Path eq '/productDataSheet') {
        TestHash('Value', $Value);

        my $FileName = $Value->{'fileName'};
        my $Url = $Value->{'url'};
        TestString('fileName', $FileName);
        TestString('url', $Url);

        $FileName = Unicode2Filename($FileName);

        my $Ua = LWP::UserAgent->new;
        my $Req = HTTP::Request->new(GET => $Url);
        my $Res = $Ua->request($Req);

        my $FileContents;
        if ($Res->is_success) {
          $FileContents = $Res->decoded_content;
        } else {
          Error('Unable to retrieve file');
        }

        my $FullPath = $Product->get('PublicPath').'/'.$FileName;
        WriteFile( $FullPath, \$FileContents, { binary => 1, create_directory => 0 } );

        $Product->set({
          'ProductDataSheet' => $FileName,
        }, $LanguageId);

        return;
    }

    if($Path eq '/isVATInfoVisible') {
        $Product->set({'IsVATInfoVisible' => JsonTrue($Value)});
        return;
    }

    if($Path eq '/shippingMethodsRestrictedTo') {
        TestArray('Value', $Value);

        my $Shop = $Product->getSite();
        my $aShippingMethods;

        foreach my $GUID ( @$Value ) {
          Error('No shipping method found') unless ExistsObjectByGUID($GUID);

          my $ShippingMethod = $Shop->loadObjectByGuid($GUID, 'ShippingMethod');
          push @$aShippingMethods, $ShippingMethod;
        }

        $Product->set({
          'ShippingMethods' => $aShippingMethods,
        });

        return;
    }

    if($Path eq '/bulkPriceInfo') {
        TestHash('Value', $Value);
        TestArray('priceInfo', $Value->{'priceInfo'});

        my $Shop = $Product->getSite();
        my $InputTaxModel = $Shop->get('InputTaxModel');
        my $UnityBulkPriceList = GetUnityPriceList($Shop, 1);
        my $ahScalePricesAndDiscounts = [];
        foreach my $hPriceInfo (@{$Value->{'priceInfo'}}) {
            TestInteger('quantity', $hPriceInfo->{'quantity'});
            TestFloat('price', $hPriceInfo->{'price'});
            next if $hPriceInfo->{'quantity'} == 1; # ignore qty=1 because of the listprice
            push @$ahScalePricesAndDiscounts, { 'Price' => $hPriceInfo->{'price'}, 'Quantity' => $hPriceInfo->{'quantity'} };
        }
        $UnityBulkPriceList->setScalePrices($Product, $InputTaxModel, $ahScalePricesAndDiscounts);

        return;
    }

    #-- { "priceQuantity":1, "orderUnit":"piece", "minOrder":2, "intervalOrder":2 }
    if($Path eq '/orderUnitInfo') {
        TestHash('Value', $Value);
        my $priceQuantity = $Value->{'priceQuantity'};
        my $orderUnit = $Value->{'orderUnit'};
        my $minOrder = $Value->{'minOrder'};
        my $intervalOrder = $Value->{'intervalOrder'};
        if ( !defined $priceQuantity || !defined $orderUnit || !defined $minOrder || !defined $intervalOrder ) {
          Error('priceQuantity, orderUnit, minOrder, intervalOrder are all required fields');
        }

        #-- Why these max values? Tallies with what the MBO enforces >.<
        TestFloat('priceQuantity', $priceQuantity, 0.001, 2099999999.999);
        TestString('orderUnit', $orderUnit);
        TestFloat('minOrder', $minOrder, 0.001, 2099999999.999);
        TestFloat('intervalOrder', $intervalOrder, 0.001, 2099999999.999);

        my $aUnits = LoadRootObject()->get('Units');
        my $OrderUnitObject;
        foreach ( @$aUnits ) {
          if ( $_->alias eq $orderUnit ) {
            $OrderUnitObject = $_ ;
            last;
          }
        }

        Error("Invalid reference unit $orderUnit") unless $OrderUnitObject;

        #-- Round to 3 dp, as enforced by the MBO
        $Product->set({
          'PriceQuantity' => int($priceQuantity*1000)/1000, #-- truncate to 3dp, NOT round
          'OrderUnit' => $OrderUnitObject,
          'MinOrder' => int($minOrder*1000)/1000, #-- truncate to 3dp, NOT round
          'IntervalOrder' => int($intervalOrder*1000)/1000, #-- truncate to 3dp, NOT round
        });

        return;
    }

    Error("Adding in $Path is not implemented!");
}

sub _remove {
    my $self = shift;
    my ($Path) = @_;

    my $Product = $self->{'Product'};
    my $LanguageId = $self->{'LanguageId'};
    my $CurrencyId = $self->{'CurrencyId'};

    if($Path eq '/bulkPriceInfo') {
        my $UnityBulkPriceList = GetUnityPriceList($Product->getSite());
        $_->delete() foreach @{$UnityBulkPriceList->scalePrices( $Product )};
        return;
    }

    if($Path eq '/name') {
        $Product->set({'Name' => undef}, $LanguageId);
        return;
    }

    if($Path eq '/shortDescription') {
        $Product->set({'Description' => undef}, $LanguageId);
        return;
    }

    if($Path eq '/description') {
        $Product->set({'Text' => undef}, $LanguageId);
        return;
    }

    if($Path eq '/deliveryPeriod') {
        $Product->set({'DeliveryPeriod' => undef});
        return;
    }

    if($Path eq '/manufacturer') {
        $Product->set({'Manufacturer' => undef});
        return;
    }

    if($Path eq '/upc' || $Path eq '/ean') {
        $Product->set({'UPCEAN' => undef});
        return;
    }

    if($Path eq '/productCode') {
        $Product->set({'UPCEAN' => undef});
        return;
    }

    if($Path eq '/essentialFeatures') {
        $Product->set({'MainCharacteristics' => undef}, $LanguageId);
        return;
    }

    if($Path eq '/searchKeywords') {
        $Product->set({'Keywords' => undef}, $LanguageId);
        return;
    }

    if($Path eq '/priceInfo/price/amount') {
        my $Shop = $Product->getSite();
        my $TaxModel = $Shop->get('TaxModel');
        $Product->setListPrice($CurrencyId, $TaxModel, undef);
        return;
    }

    if($Path eq '/priceInfo/manufacturerPrice/amount') {
        my $Shop = $Product->getSite();
        my $TaxModel = $Shop->get('TaxModel');
        $Product->setManufacturerPrice($CurrencyId, $TaxModel, undef);
        return;
    }

    if($Path eq '/priceInfo/depositPrice/amount') {
        my $Shop = $Product->getSite();
        my $TaxModel = $Shop->get('TaxModel');
        $Product->setDeposit($CurrencyId, $TaxModel, undef);
        return;
    }

    if($Path eq '/stocklevel') {
        $Product->set({'StockLevel' => undef});
        return;
    }

    if($Path eq '/minStocklevel') {
        $Product->set({'StockLevelAlert' => undef});
        return;
    }

    if($Path eq '/manufacturerProductNumber') {
        $Product->set({'ManufacturerSKU' => undef});
        return;
    }

    if($Path eq '/productLength') {
        $Product->set({'Length' => undef});
        return;
    }

    if($Path eq '/productWidth') {
        $Product->set({'Width' => undef});
        return;
    }

    if($Path eq '/productHeight') {
        $Product->set({'Height' => undef});
        return;
    }
    if($Path eq '/newnessDate') {
        $Product->set({'NewnessDate' => undef});
        return;
    }
    if($Path eq '/video') {
        $Product->set({
          'ProductVideoProvider' => undef,
        }, $LanguageId);
        return;
    }
    if($Path eq '/referencePriceInfo') {
        $Product->set({
          'RefAmount' => undef,
          'RefUnit' => undef,
          'RefContentAmount' => undef,
        });

        return;
    }
    if($Path eq '/customisationInfo') {
        $Product->set({
          'IsCustomizable' => 0,
          'CustomizableTextLength' => undef,
          'CustomizableHeadlineText' => undef,
        }, $LanguageId);

        return;
    }
    if($Path eq '/energyLabelInfo') {
        my $FileName = $Product->get('EnergyLabelSourceFile');
        DE_EPAGES::Core::API::File::DeleteFile($Product->get('PublicPath').'/'.$FileName) if $FileName;
        $Product->set({
            'EnergyLabelHighestEfficiencyClass'     => undef,
            'EnergyLabels'                          => undef,
            'EnergyLabelSourceFile'                 => undef,
        });
        return;
    }
    if($Path eq '/productDataSheet') {
        my $FileName = $Product->get('ProductDataSheet', $LanguageId);
        DE_EPAGES::Core::API::File::DeleteFile($Product->get('PublicPath').'/'.$FileName) if $FileName;
        $Product->set({
            'ProductDataSheet'     => undef,
        }, $LanguageId);
        return;
    }
    if($Path eq '/shippingMethodsRestrictedTo') {
      $Product->set({
          'ShippingMethods'     => undef,
      });
      return;
    }
    Error("Removing $Path is not implemented!");
}

sub _replace {
    my $self = shift;
    my ($Path, $Value) = @_;

    my $Product = $self->{'Product'};
    my $Shop = $Product->getSite();

    if($Path eq '/bulkPriceInfo') {
        my $InputTaxModel = $Shop->get('InputTaxModel');
        my $UnityBulkPriceList = GetUnityPriceList($Shop, 1);
        my $ahExistingScalePrices =  $UnityBulkPriceList->getScalePrices( $Product, $InputTaxModel, undef, 1 );
        my $ahScalePricesToSet = [];
        # replace values in the existing pricelist
        foreach my $hPriceInfo (@{$Value->{'priceInfo'}}) {
            TestInteger('quantity', $hPriceInfo->{'quantity'});
            next if $hPriceInfo->{'quantity'} == 1; # ignore qty=1 because of the listprice
            foreach my $hExistingEntry (@$ahExistingScalePrices){
                next unless $hExistingEntry->{'Quantity'} == $hPriceInfo->{'quantity'};
                $hExistingEntry->{'Price'} = $hPriceInfo->{'price'};
            }
        }

        # filter out empty prices and save price list
        foreach my $hEntry (@$ahExistingScalePrices){
            next unless $hEntry->{'Price'};
            push @$ahScalePricesToSet, $hEntry;
        }

        $UnityBulkPriceList->setScalePrices($Product, $InputTaxModel, $ahScalePricesToSet);

        return;
    }

    if($Path eq '/priceInfo/taxClass/taxClassId') {
        my $TaxClass = $self->_loadTaxClass($Value, $Shop);

        if(defined $TaxClass) {
            $self->_setTaxClass($TaxClass);
        } else {
            Error('TaxClass could not be found!');
        }

        return;
    }

    if($Path eq '/producttype') {
        TestString('Value', $Value);

        # test if the given product type GUID is valid and can be set on the product
        # if we have variations on the product we can not set the new producttype
        Error('Product has variation products with current product type!') if  $Product->get('IsMaster');

        my $ProductType = $Shop->loadObjectByGuid($Value, 'ProductType');
        my $ProductTypeFolder = $Shop->folder('ProductTypes');

        Error("ProductType $Value not exists!") unless $ProductTypeFolder->existsChild($ProductType->alias);

        # do change only, when class was really changed
        if ($Product->class->id != $ProductType->id) {
            $Product->set({'Class' => $ProductType});
        }

        return;
    }

    if($Path eq '/stocklevel') {
        TestFloat('Value', $Value);
        $Product->set({'StockLevel' => $Value});

        return;
    }

    Error("Replacing $Path is not implemented!");
}

sub _loadTaxClass {
    my $self = shift;
    my ($GUID, $Shop) = @_;

    my $ShopTaxClass = $self->_createShopTaxClass($Shop);
    my $TaxClass = $ShopTaxClass->loadTaxClassbyGUID($GUID);

    return $TaxClass;
}

sub _setTaxClass {
    my $self = shift;
    my ($TaxClass) = @_;

    my $Product = $self->{'Product'};

    if($TaxClass->instanceOf('TaxClass') || $TaxClass->instanceOf('TaxClassUS')) {
        $Product->set({'TaxClass' => $TaxClass});

        # Note: if there is a TaxClassDigital, the TaxClass of a product will be ignored!
        $Product->set({'TaxClassDigital' => undef});
    }

    if($TaxClass->instanceOf('TaxClassDigital')) {
        $Product->set({'TaxClassDigital' => $TaxClass});
    }

    return;
}

sub _parseBoolean {
    my $self = shift;
    my ($Boolean) = @_;

    if ( $Boolean eq 'true' ) {
        return 1;
    } elsif ( $Boolean eq 'false' ) {
        return 0;
    } else {
        Error("Could not parse Boolean $Boolean");
    }
}

sub _convertKeywordsArrayToString {
    my $self = shift;
    my($aKeywords) = @_;

    my @TrimmedKeywords = map {trim($_)} @$aKeywords; # remove leading and trailing spaces
    my @Keywords = grep {length($_) > 0} @TrimmedKeywords; # remove empty strings
    my $KeywordString = reduce {"$a, $b"} @Keywords; # create string with comma-separated list of keywords

    return $KeywordString;
}

sub _createProductImageService {
    my $self = shift;
    my ($Product) = @_;
    return DE_EPAGES::Product::API::ProductImageService->new($Product);
}

sub _createShopTaxClass {
    my $self = shift;
    my ($Shop) = @_;
    return DE_EPAGES::Shop::API::ShopTaxClass->new($Shop);
}

1;
