#========================================================================================
# package      DE_EPAGES::Product::API::TLE::ProductHandler
# state        public
#----------------------------------------------------------------------------------------
# description  Implements TLE functions for products.
#========================================================================================
package DE_EPAGES::Product::API::TLE::ProductHandler;
use base DE_EPAGES::TLE::API::LoopHandler;

use strict;
use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Core::API::Object::NumberFormat;
use DE_EPAGES::Core::API::PerlTools qw ( ElementOf );
use DE_EPAGES::Core::API::TestParameter qw ( TestInteger TestObject TestString );
use DE_EPAGES::Core::API::Url qw ( UrlEncodeValue );
use DE_EPAGES::Customer::API::Shop qw ( GetCustomerForUser );
use DE_EPAGES::Object::API::Factory qw ( LoadObject );
use DE_EPAGES::Product::API::Product qw ( UsedVariationValues );
use DE_EPAGES::Product::API::Table::SelectedVariation qw (
    GetInfoSelectedVariationsByProductID
);
use DE_EPAGES::Product::DAL::Product qw ( DAL_GetVariationAttributesBySuperProductID );
use DE_EPAGES::Shop::API::Constants qw ( TAX_MODEL_GROSS );

#========================================================================================
# function     RegisterHandlerProc
# state        private
#----------------------------------------------------------------------------------------
# syntax       $RegisterHandlerProc( $hParams );
#----------------------------------------------------------------------------------------
# description  this function registers this attribute handle on the processor via hook
#               function on hook TLEProcessorRegistration
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments keys :
#               <ul>
#                   <li>Processor - tle processor - object
#               </ul> | ref.hash
#========================================================================================
sub RegisterHandlerProc {
    my ($Params) = @_;
    GetLog->debug( 'ProductHandler.RegisterHandlerProc' );

    __PACKAGE__->new()->register( $Params->{'Processor'} );
    return;
}

#========================================================================================
# function     new
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler = new DE_EPAGES::Product::API::TLE::ProductHandler;
# example      my $PrefixHandler  = new DE_EPAGES::Product::API::TLE::ProductHandler;
#----------------------------------------------------------------------------------------
# description  create a new handler
#----------------------------------------------------------------------------------------
# return       $Handler | blessed object | object
#========================================================================================
sub new { bless {}, shift; }

#========================================================================================
# function     register
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler->register( $Processor );
#----------------------------------------------------------------------------------------
# description  this function register this handle on the processor
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
#========================================================================================
sub register {
    my $self = shift;
    my ($Processor) = @_;
    GetLog->debug( 'ProductHandler.register' );

    $Processor->registerHandler('FunctionHandler', $self, 'IsVariationAttribute');
    $Processor->registerHandler('FunctionHandler', $self, 'LOOPPosVariationTypeValues');
    $Processor->registerHandler('FunctionHandler', $self, 'REFERENCEPRICE');
    $Processor->registerHandler('FunctionHandler', $self, 'LISTPRICE');
    $Processor->registerHandler('FunctionHandler', $self, 'BUNDLEDLISTPRICESUM');
    $Processor->registerHandler('FunctionHandler', $self, 'MANUFACTURERPRICE');
    $Processor->registerHandler('FunctionHandler', $self, 'ECOPARTICIPATION');
    $Processor->registerHandler('FunctionHandler', $self, 'DEPOSIT');
    $Processor->registerHandler('FunctionHandler', $self, 'DISPLAYPRICE');
    $Processor->registerHandler('FunctionHandler', $self, 'HASPRICE');
    $Processor->registerHandler('FunctionHandler', $self, 'CANDISPLAYFROMPRICE');
    $Processor->registerHandler('FunctionHandler', $self, 'ProductURL');
    $Processor->registerHandler('FunctionHandler', $self, 'AppointmentDayList');
    $Processor->registerHandler('FunctionHandler', $self, 'DISPLAYPRICETAGINFOGROSS');
    $Processor->registerHandler('FunctionHandler', $self, 'DISPLAYFREETAXINFO');
    $Processor->registerHandler('FunctionHandler', $self, 'APPOINTMENTMENT_PRICE');
    $Processor->registerHandler('FunctionHandler', $self, 'APPOINTMENT_PAYMENTS');
    $Processor->registerHandler('FunctionHandler', $self, 'EXIST_APPOINTMENT_PAYMENT');
    $Processor->registerHandler('FunctionHandler', $self, 'PREPAYMENT');
    $Processor->registerHandler('FunctionHandler', $self, 'APPOINTMENTMENT_PREPAYMENT');
    $Processor->registerHandler('FunctionHandler', $self, 'PRICESAVINGS');
    $Processor->registerHandler('FunctionHandler', $self, 'GetVisibleUserDefinedAttributes');
    $Processor->registerHandler('FunctionHandler', $self, 'UnavailableVariations');
    $Processor->registerHandler('FunctionHandler', $self, 'InvisibleVariations');
    $Processor->registerHandler('FunctionHandler', $self, 'ManufacturerPriceForCheapestVariation');
    $Processor->registerHandler('FunctionHandler', $self, 'LOWESTREFERENCEPRICEFORMASTER');
    $Processor->registerHandler('FunctionHandler', $self, 'PRICERANGEFORVARIATIONS');

    return;
}

#========================================================================================
# function     IsVariationAttribute
# state        public
#----------------------------------------------------------------------------------------
# syntax       #IsVariationAttribute(#Product, #Attribute)
# syntax       #LOOP(#UserDefinedAttributeValues)
#                 #IF(#IsVariationAttribute(#Object, #Attribute))
#                   #Attribute.NameOrAlias = #Value
#                 #ENDIF
#               #ENDIF
#----------------------------------------------------------------------------------------
# description  Tests if the attribute is a variation attribute of product.
#----------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #Attribute | attribute object | object
#========================================================================================
sub IsVariationAttribute { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $Attribute ) = @$aParams;

    my $SuperProduct = $Product->superProduct;

    my $IsVariationAttribute = 0;
    my $VariationTypeValues = defined $SuperProduct ? $SuperProduct->get('VariationTypeValues') : $Product->get('VariationTypeValues');
    foreach my $VariationType (@$VariationTypeValues) {
        if ( $VariationType->{'Attribute'}->id == $Attribute->id ) {
            $IsVariationAttribute = 1;
            last;
        }
    }
    return $IsVariationAttribute;
}

#========================================================================================
# function     LOOPPosVariationTypeValues
# state        public
#----------------------------------------------------------------------------------------
# syntax       #BLOCK("LOOPPosVariationTypeValues", #Product) template #ENDBLOCK
# example      #BLOCK("LOOPPosVariationTypeValues", #Object)
#                 #Attribute.NameOrAlias
#                 #LOOP(#PreDefAttributes)
#                   #ValueOrAlias #IF(NOT #IsUsed)(not used)#ENDIF
#                 #ENDLOOP
#               #ENDBLOCK
#----------------------------------------------------------------------------------------
# description  Loops the template code for each variation type.
#               The TLE variables #Attribute and #PreDefAttributes can be used within
#               the block.
#               #Attibute contains the attribute object.
#               #PreDefAttributes contains a list of all available values of the
#               variation type. #IsUsed indicates if the value is used by the product.
#----------------------------------------------------------------------------------------
# input        #Product | product object | object
#========================================================================================
sub LOOPPosVariationTypeValues { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams, $cTemplate) = @_;
    my ($Product ) = @$aParams;

    my $Attributes = $Product->class->get('UserDefinedAttributes');
    my @PosVariationAttributes = grep {$_->get('Type') eq 'PreDefString' or $_->get('Type') eq 'PreDefLocalizedString' } @$Attributes;

    my $rExistingValues = GetInfoSelectedVariationsByProductID($Product->id);
    my @SelectVariationIDs = map { $_->{'PreDefAttributeID'} } @$rExistingValues;
    my $UsedVariationIDs = UsedVariationValues($Product);

    my @VariationTypes;
    foreach my $Attribute (@PosVariationAttributes) {
        my $PreDefValues = $Attribute->get('PreDefAttributes');
        foreach my $PreDefValue (@$PreDefValues ) {
            my $ID = $PreDefValue->id;
            $PreDefValue = $PreDefValue->tleHash;
            $PreDefValue->{'IsUsed'} = ElementOf ($ID , @$UsedVariationIDs);
            $PreDefValue->{'IsSelected'} = ElementOf ($ID , @SelectVariationIDs);
        }
        push @VariationTypes, { 'Attribute'        => $Attribute->tleHash,
                                'PreDefAttributes' => $PreDefValues
                              }
    }
    return $self->LOOP($Processor,[\@VariationTypes], $cTemplate);
}

#========================================================================================
# function     REFERENCEPRICE
# state        public
#----------------------------------------------------------------------------------------
# syntax       #REFERENCEPRICE(#Product, #Price, #CurrencyID)
# example      #LOCAL("CurrencyID", #INPUT.Currency)
#                 #LOCAL("ListPrice", #LISTPRICE(#Object, #INPUT.Currency, #Shop.TaxModel))
#                   #LOCAL("ReferencePrice",#REFERENCEPRICE(#Object, #ListPrice, #CurrencyID))
#                     #IF(#ReferencePrice != #ListPrice)
#                       #RefAmount#RefUnit.Abbreviation = #ReferencePrice[money]
#                     #ENDIF
#                   #LOCAL
#                 #ENDLOCAL
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Calculates the price of a product relative to the reference amout.
#               The following formula is used:
#               <br/>#Price * (#Product.RefAmount / #Product.RefContentAmount)
#----------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #Price | price | float
#========================================================================================
sub REFERENCEPRICE { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $Price, $CurrencyID) = @$aParams;

    return $Product->getReferencePrice($CurrencyID, $Price);
}


#========================================================================================
# function     LOWESTREFERENCEPRICEFORMASTER
# state        public
#----------------------------------------------------------------------------------------
# syntax       #LOWESTREFERENCEPRICEFORMASTER(#Product, #CurrencyID, #TaxModel)
# example      #LOCAL("CurrencyID", #INPUT.Currency)
#                 #LOCAL("ListPrice", #LISTPRICE(#Object, #INPUT.Currency, #Shop.TaxModel))
#                   #LOCAL("LowestReferencePrice",#LOWESTREFERENCEPRICEFORMASTER(#Object, #CurrencyID, $TaxModel))
#                     #IF(#LowestReferencePrice != #ListPrice)
#                       #RefAmount#RefUnit.Abbreviation = #ReferencePrice[money]
#                     #ENDIF
#                   #LOCAL
#                 #ENDLOCAL
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Calculates the reference price of the variation with the lowest list price
#               The following formula is used:
#               <br/>#Price * (#Product.RefAmount / #Product.RefContentAmount)
#               Must be used on SuperProduct
#----------------------------------------------------------------------------------------
# input        #SuperProduct | product object | object
# input        #CurrencyID | currency id (e.g. 'USD') | string
# input        #TaxModel | tax calculation model (0=net, 1=gross) | int
#========================================================================================
sub LOWESTREFERENCEPRICEFORMASTER { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($SuperProduct, $CurrencyID, $TaxModel) = @$aParams;

    TestObject('Product', $SuperProduct, 'DE_EPAGES::Product::API::Object::Product');
    TestString('CurrencyID', $CurrencyID, 3, 3);
    TestInteger('TaxModel', $TaxModel, 0, 1);

    my $hPriceParams = {
        CurrencyID  => $CurrencyID,
        TaxModel    => $TaxModel,
    };
    my $SubProduct = $SuperProduct->subProductWithLowestPrice($hPriceParams);
    if (defined $SubProduct) {
        my $Price = $SubProduct->getListPrice($CurrencyID, $TaxModel);
        my $ReferencePrice = $SubProduct->getReferencePrice($CurrencyID, $Price);

        return $ReferencePrice;
    }

    return undef;
}

#========================================================================================
# function     PRICERANGEFORVARIATIONS
# state        public
#----------------------------------------------------------------------------------------
# syntax       #PRICERANGEFORVARIATIONS(#Object, #CurrencyID, #Shop.TaxModel, #Session.User.Object)
# example      #LOCAL("PriceRange",#PRICERANGEFORVARIATIONS(#Object, #CurrencyID, #Shop.TaxModel, #Session.User.Object))
#                 #PriceRange.LowestPrice
#                 #PriceRange.HighestPrice
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Returns the highest and lowest price of a product for a given currency and tax model
#               regardless to the fact wheather the product is available or not.
#----------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #Currency | currency code, e.g. "EUR" | string
# input        #TaxModel | tax calculation model (0=net, 1=gross) | int
# input        #User | (optional) user object | object
#========================================================================================
sub PRICERANGEFORVARIATIONS { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($SuperProduct, $CurrencyID, $TaxModel, $User) = @$aParams;

    TestObject('Product', $SuperProduct, 'DE_EPAGES::Product::API::Object::Product');
    TestString('CurrencyID', $CurrencyID, 3, 3);
    TestInteger('TaxModel', $TaxModel, 0, 1);
    TestObject('User', $User, 'DE_EPAGES::Permission::API::Object::User') if defined $User;

    my $Shop = $SuperProduct->getSite();
    my $Customer = undef;

    if (defined $User and not $User->get('IsAnonymous')) {
        $Customer = GetCustomerForUser ($User, $Shop);
        $TaxModel = $Customer->get('TaxModel') if defined $Customer;
    }

    my $LowestPrice = $SuperProduct->fromPrice({
        'CurrencyID' => $CurrencyID,
        'TaxModel'   => $TaxModel,
        'Customer'   => $Customer,
        'Quantity'   => 1,
        'IgnoreCanAddToBasket' => 1,
        'User'       => $User,
    });

    my $HighestPrice = $SuperProduct->toPrice({
        'CurrencyID' => $CurrencyID,
        'TaxModel'   => $TaxModel,
        'Customer'   => $Customer,
        'Quantity'   => 1,
        'IgnoreCanAddToBasket' => 1,
        'User'       => $User,
    });

    return {
            'LowestPrice' => $LowestPrice,
            'HighestPrice' => $HighestPrice
           };

}

#========================================================================================
# function     LISTPRICE
# state        public
#----------------------------------------------------------------------------------------
# syntax       #LISTPRICE(#Product, #Currency, #TaxModel, #ExraPrecision)
# example      #LOCAL("CurrencyID", #INPUT.Currency)
#                 #LOCAL("ListPrice", #LISTPRICE(#Object, #INPUT.Currency, #Shop.TaxModel, 1))
#                   #IF(#DEFINED(#ListPrice))
#                     #ListPrice[money]
#                   #ENDIF
#                 #ENDLOCAL
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Returns the list price of a product for a given currency and tax model.
#----------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #Currency | currency code, e.g. "EUR" | string
# input        #TaxModel | tax calculation model (0=net, 1=gross) | int
# input        #ExraPrecision | (optional, default=0) extra precision | int
#========================================================================================
sub LISTPRICE { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $CurrencyID, $TaxModel, $ExtraPrecision) = @$aParams;

    $ExtraPrecision = 0 unless defined $ExtraPrecision;
    return $Product->getListPrice($CurrencyID, $TaxModel, $ExtraPrecision);
}

#========================================================================================
# function     BUNDLEDLISTPRICESUM
# state        public
#----------------------------------------------------------------------------------------
# syntax       #BUNDLEDLISTPRICESUM(#Product, #Currency, #TaxModel)
# example      #LOCAL("SumPrice", #BUNDLEDLISTPRICESUM(#Object, #INPUT.Currency, #Shop.TaxModel))
#                 #SumPrice[money]
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Returns the sum of the list prices of all items of a bundle product.
#               See also: <function LISTPRICE>
#----------------------------------------------------------------------------------------
# input        #Product | product object (must be a bundle product) | object
# input        #Currency | currency code, e.g. "EUR" | string
# input        #TaxModel | tax calculation model (0=net, 1=gross) | int
# input        #User | (optional) user object | object
#========================================================================================
sub BUNDLEDLISTPRICESUM { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $CurrencyID, $TaxModel, $User) = @$aParams;

    my $Shop = $Product->getSite();
    my $Customer = undef;
    if (defined $User and not $User->get('IsAnonymous')) {
        $Customer = GetCustomerForUser ($User, $Shop);
        $TaxModel = $Customer->get('TaxModel') if defined $Customer;
    }

    my $Price = 0;
    my $Formatter = $Product->getSite()->formatter($CurrencyID);

    foreach my $BundleProductMap ( @{$Product->get('BundleProductMaps')}) {
        my $BundledProduct = $BundleProductMap->get('BundledProduct');
        my $BundledProductPrice = $self->LISTPRICE($Processor, [ $BundledProduct, $CurrencyID, $TaxModel ] );
        if ( defined $BundledProductPrice ) {
            my $ItemPrice = $BundledProductPrice * $BundleProductMap->get('BundledQuantity') / $BundledProduct->getWithParent('PriceQuantity');
            $Price += $Formatter->round($ItemPrice);
        }
    }

    return $Formatter->round($Price);
}

#========================================================================================
# function     PRICESAVINGS
# state        public
#----------------------------------------------------------------------------------------
# syntax       #PRICESAVINGS(#Product, #Price1, #Price2)
# example      #LOCAL("PriceSavings", #PRICESAVINGS(#Object, #DisplayPrice, #ManufacturerPrice))
#                 #IF(#DEFINED(#PriceSavings))<div>#PriceSavings</div>#ENDIF
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Returns the percentage or absolute value of the pricesavings.
#----------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #Price1 | lower price e.g. DisplayPrice | string
# input        #Price2 | comparison price e.g. ManufacturerPrice | int
#========================================================================================
sub PRICESAVINGS { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $Price, $ComparisonPrice) = @$aParams;

    my $Shop = $Product->getSite();
    my $ShowPriceSaving = $Shop->get('ShowPriceSavings');

    my $PriceSavings;

    if($ShowPriceSaving == 1) {
        $PriceSavings = $Processor->formatTLE('PriceSaving', 'money', $ComparisonPrice - $Price);
    } elsif ($ShowPriceSaving == 2) {
        $PriceSavings = $Processor->formatTLE('PriceSaving', 'float:0', 100 - (($Price * 100) / $ComparisonPrice)).' %'; ## no critic qw(ProhibitMagicNumbers)
    }

    return $PriceSavings;
}

#=========================================================================================================
# function     MANUFACTURERPRICE
# state        public
#---------------------------------------------------------------------------------------------------------
# syntax       #MANUFACTURERPRICE(#Product, #CurrencyID, #TaxModel, #User, #ExtraPrecision)
# example      #LOCAL("CurrencyID", #INPUT.Currency)
#                 #LOCAL("ListPrice", #MANUFACTURERPRICE(#Product, #CurrencyID, #Shop.TaxModel, #User, 1))
#                   #IF(#DEFINED(#ListPrice))
#                     #Price[money]
#                   #ENDIF
#                 #ENDLOCAL
#               #ENDLOCAL
#---------------------------------------------------------------------------------------------------------
# description  Returns the manufacturer's suggested retail price of a product for a given currency and
#               tax model. If user is defined the tax model of the user take effect.
#---------------------------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #CurrencyID | currency ID, e.g. "EUR" | string
# input        #TaxModel | tax calculation model (0=net, 1=gross) | int
# input        #User | (optional) user object | object
# input        #ExtraPrecision | (optional, default=0) extra precision | int
#=========================================================================================================
sub MANUFACTURERPRICE { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $CurrencyID, $TaxModel, $User, $ExtraPrecision) = @$aParams;

    TestObject('Product', $Product);
    TestString('CurrencyID', $CurrencyID, 3, 3);
    TestInteger('TaxModel', $TaxModel, 0, 1);

    my $Shop = $Product->getSite();
    my $Customer = undef;

    if (defined $User and not $User->get('IsAnonymous')) {
        $Customer = GetCustomerForUser ($User, $Shop);
        $TaxModel = $Customer->get('TaxModel') if defined $Customer;
    }

    # return undef price if the price is only shown for registered user & user is not logged in or is merchant
    return undef unless ((not $Shop->get('IsPriceOnlyForRegisteredUser')) or
                        (defined $User and not $User->get('IsAnonymous')) or
                        (defined $User and $User->get('IsMerchant')));

    $ExtraPrecision = 0 unless defined $ExtraPrecision;

    return $Product->getManufacturerPrice($CurrencyID, $TaxModel, $ExtraPrecision);
}

#=========================================================================================================
# function     ManufacturerPriceForCheapestVariation
# state        public
#---------------------------------------------------------------------------------------------------------
# syntax       #ManufacturerPriceForCheapestVariation(#Product, #CurrencyID, #TaxModel, #User, #ExtraPrecision)
# example      #LOCAL("CurrencyID", #INPUT.Currency)
#                 #LOCAL("ListPrice", #ManufacturerPriceForCheapestVariation(#Product, #CurrencyID, #Shop.TaxModel, #User, 1))
#                   #IF(#DEFINED(#ListPrice))
#                     #Price[money]
#                   #ENDIF
#                 #ENDLOCAL
#               #ENDLOCAL
#---------------------------------------------------------------------------------------------------------
# description  Returns the manufacturer price of the variation with the lowest price of a variation product 
#               for a given currency and tax model. If user is defined the tax model of the user take effect.
#---------------------------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #CurrencyID | currency ID, e.g. "EUR" | string
# input        #TaxModel | tax calculation model (0=net, 1=gross) | int
# input        #User | (optional) user object | object
# input        #ExtraPrecision | (optional, default=0) extra precision | int
#==========================================================================================================
sub ManufacturerPriceForCheapestVariation {
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $CurrencyID, $TaxModel, $User, $ExtraPrecision) = @$aParams;

    TestObject('Product', $Product);
    TestString('CurrencyID', $CurrencyID, 3, 3);
    TestInteger('TaxModel', $TaxModel, 0, 1);
    TestObject('User', $User, 'DE_EPAGES::Permission::API::Object::User') if defined $User;
    TestInteger('ExtraPrecision', $ExtraPrecision) if defined $ExtraPrecision;

    my $Shop = $Product->getSite();
    my $Customer = undef;

    if (defined $User and not $User->get('IsAnonymous')) {
        $Customer = GetCustomerForUser ($User, $Shop);
        $TaxModel = $Customer->get('TaxModel') if defined $Customer;
    }

    my $SuperProduct = $Product->isMaster ? $Product : $Product->superProduct;
    if (defined $SuperProduct) {

        my $hPriceParams = {
            CurrencyID  => $CurrencyID,
            TaxModel    => $TaxModel,
        };
        my $SubProduct = $SuperProduct->subProductWithLowestPrice($hPriceParams);
        if (defined $SubProduct) {
            return $SubProduct->getManufacturerPrice($CurrencyID, $TaxModel, $ExtraPrecision);
        }
        else {
            return undef;
        }
    }
    else {
        return undef;
    }
}

#===============================================================================================================
# function     ECOPARTICIPATION
# state        public
#---------------------------------------------------------------------------------------------------------------
# syntax       #ECOPARTICIPATION(#Product, #CurrencyID, #TaxModel, #User, #ExtraPrecision)
# example      #LOCAL("CurrencyID", #INPUT.Currency)
#                 #LOCAL("EcoParticipation", #ECOPARTICIPATION(#Product, #CurrencyID, #Shop.TaxModel, #User, 1))
#                   #IF(#DEFINED(#EcoParticipation))
#                     #EcoParticipation[money]
#                   #ENDIF
#                 #ENDLOCAL
#               #ENDLOCAL
#---------------------------------------------------------------------------------------------------------------
# description  Returns the list price of a product for a given currency and tax model. If user is defined the
#               tax model of the user take effect.
#---------------------------------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #CurrencyID | currency ID, e.g. "EUR" | string
# input        #TaxModel | tax calculation model (0=net, 1=gross) | int
# input        #User | (optional) user object | object
# input        #ExtraPrecision | (optional, default=0) extra precision | int
#===============================================================================================================
sub ECOPARTICIPATION { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $CurrencyID, $TaxModel, $User, $ExtraPrecision) = @$aParams;

    TestObject('Product', $Product, 'DE_EPAGES::Product::API::Object::Product');
    TestString('CurrencyID', $CurrencyID, 3, 3);
    TestInteger('TaxModel', $TaxModel, 0, 1);

    my $Shop = $Product->getSite();

    if (defined $User and not $User->get('IsAnonymous')) {
        my $Customer = GetCustomerForUser ($User, $Shop);
        $TaxModel = $Customer->get('TaxModel') if defined $Customer;
    }

    # return undef price if the price is only shown for registered user & user is not logged in
    return undef unless ((not $Shop->get('IsPriceOnlyForRegisteredUser')) or (defined $User and not $User->get('IsAnonymous')));

    $ExtraPrecision = 0 unless defined $ExtraPrecision;
    return $Product->getEcoParticipation($CurrencyID, $TaxModel, $ExtraPrecision);
}

#=============================================================================================
# function     DEPOSIT
# state        public
#---------------------------------------------------------------------------------------------
# syntax       #DEPOSIT(#Product, #CurrencyID, #TaxModel, #User, #ExtraPrecision)
# example      #LOCAL("CurrencyID", #INPUT.Currency)
#                 #LOCAL("Deposit", #DEPOSIT(#Product, #CurrencyID, #Shop.TaxModel, #User, 1))
#                   #IF(#DEFINED(#Deposit))
#                     #Deposit[money]
#                   #ENDIF
#                 #ENDLOCAL
#               #ENDLOCAL
#---------------------------------------------------------------------------------------------
# description  Returns the deposit price of a product for a given currency and tax model. If
#               user is defined the tax model of the user take effect.
#---------------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #CurrencyID | currency ID, e.g. "EUR" | string
# input        #TaxModel | tax calculation model (0=net, 1=gross) | int
# input        #User | (optional) user object | object
# input        #ExtraPrecision | (optional, default=0) extra precision | int
#=============================================================================================
sub DEPOSIT { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $CurrencyID, $TaxModel, $User, $ExtraPrecision) = @$aParams;

    TestObject('Product', $Product, 'DE_EPAGES::Product::API::Object::Product');
    TestString('CurrencyID', $CurrencyID, 3, 3);
    TestInteger('TaxModel', $TaxModel, 0, 1);

    my $Shop = $Product->getSite();

    if (defined $User and not $User->get('IsAnonymous')) {
        my $Customer = GetCustomerForUser ($User, $Shop);
        $TaxModel = $Customer->get('TaxModel') if defined $Customer;
    }

    # return undef price if the price is only shown for registered user & user is not logged in
    return undef unless ((not $Shop->get('IsPriceOnlyForRegisteredUser')) or (defined $User and not $User->get('IsAnonymous')));

    $ExtraPrecision = 0 unless defined $ExtraPrecision;
    return $Product->getDeposit($CurrencyID, $TaxModel, $ExtraPrecision);
}

#==============================================================================================================
# function     DISPLAYPRICE
# state        public
#--------------------------------------------------------------------------------------------------------------
# syntax       #DISPLAYPRICE(#Product, #CurrencyID, #TaxModel, #User, #NoFromPrice)
# example      #LOCAL("CurrencyID", #INPUT.Currency)
#                 #LOCAL("ListPrice", #DISPLAYPRICE(#Product, #CurrencyID, #Shop.TaxModel, #User, #NoFromPrice))
#                   #IF(#DEFINED(#ListPrice))
#                     #ListPrice[money]
#                   #ENDIF
#                 #ENDLOCAL
#               #ENDLOCAL
#--------------------------------------------------------------------------------------------------------------
# description  Returns the display price of a product for a given currency, tax model and customer. If user is
#               defined the tax model of the user take effect.
#--------------------------------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #CurrencyID | currency ID, e.g. "EUR" | string
# input        #TaxModel | tax calculation model (0=net, 1=gross) | int
# input        #User | (optional) user object | object
# input        #NoFromPrice | do not use the from price calculation | boolean
#==============================================================================================================
sub DISPLAYPRICE { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $CurrencyID, $TaxModel, $User, $NoFromPrice) = @$aParams;

    TestObject('Product', $Product, 'DE_EPAGES::Product::API::Object::Product');
    TestString('CurrencyID', $CurrencyID, 3, 3);
    TestInteger('TaxModel', $TaxModel, 0, 1);

    my $Shop = $Product->getSite();
    my $Customer = undef;

    if (defined $User and not $User->get('IsAnonymous')) {
        $Customer = GetCustomerForUser ($User, $Shop);
        $TaxModel = $Customer->get('TaxModel') if defined $Customer;
    }

    # return undef price if the price is only shown for registered user & user is not logged in
    return undef unless ((not $Shop->get('IsPriceOnlyForRegisteredUser')) or (defined $User and not $User->get('IsAnonymous')));

    my $Quantity = $Product->getWithParent('MinOrder');

    my $ShowFromPrice = ! $NoFromPrice;

    if ( $Product->isMaster ){
        if ( $ShowFromPrice && defined $Shop->get('VariationPriceFromText')) {
            return $Product->fromPrice({
                'CurrencyID' => $CurrencyID,
                'TaxModel'   => $TaxModel,
                'Customer'   => $Customer,
                'Quantity'   => $Quantity,
                'User'       => $User,
            });
        } else {
            return $Product->displayPrice({
                'CurrencyID' => $CurrencyID,
                'TaxModel'   => $TaxModel,
                'Customer'   => $Customer,
                'Quantity'   => $Quantity,
                'User'       => $User,
            });
        }
    } else {
        # Case SubProduct or Bundle Product or Product or ...
        if ( $ShowFromPrice && defined $Shop->get('VariationPriceFromText') && $Product->isSubProduct ) {
            return $Product->superProduct->fromPrice({
                'CurrencyID' => $CurrencyID,
                'TaxModel'   => $TaxModel,
                'Customer'   => $Customer,
                'Quantity'   => $Quantity,
                'User'       => $User,
            });
        } else {
            # Only SubProduct can query fromPrice everthing else uses displayPrice
            return $Product->displayPrice({
                'CurrencyID' => $CurrencyID,
                'TaxModel'   => $TaxModel,
                'Customer'   => $Customer,
                'Quantity'   => $Quantity,
                'User'       => $User,
            });
        }
    }
}

#==============================================================================
# function     HASPRICE
# state        public
#------------------------------------------------------------------------------
# syntax       #HASPRICE(#Product, #CurrencyID, #User)
# example      #IF(#HASPRICE(#Product, #INPUT.Currency, #Session.User.Object))
#                   <a href="...">add to basket</a>
#               #ENDIF
# example      #IF(#HASPRICE(#Product, "EUR", #Session.User.Object))
#                   <a href="...">add to basket</a>
#               #ENDIF
#------------------------------------------------------------------------------
# description  Returns true if a price exists for the given product, currency
#               and user, i.e. either a customer specific price for 1 item or
#               the list price.
#------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #CurrencyID | currency ID, e.g. "EUR" | string
# input        #User | user object | object
#==============================================================================
sub HASPRICE { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $CurrencyID, $User) = @$aParams;

    TestObject('Product', $Product, 'DE_EPAGES::Product::API::Object::Product');
    TestString('CurrencyID', $CurrencyID, 3, 3);
    TestObject('User', $User, 'DE_EPAGES::Permission::API::Object::User') if defined $User;

    my $Shop = $Product->getSite();

    # return false if the price is only shown for registered user & user is not logged in
    return 0 unless ((not $Shop->get('IsPriceOnlyForRegisteredUser')) or (defined $User and not $User->get('IsAnonymous')));

    return $Product->hasPrice($CurrencyID, { 'User' => $User });
}

#========================================================================================
# function     AppointmentDayList
# state        public
#----------------------------------------------------------------------------------------
# syntax       #AppointmentDayList(#Product, #Date)
# example      #LOOP(#AppointmentDayList(#Object, #NOW))
#               #ENDLOOP
#----------------------------------------------------------------------------------------
# description  Returns the <function DE_EPAGES::Calendar::API::Object::Calendar::dayList>
#               for the given product <function DE_EPAGES::Product::API::Object::Product::businessHours>.
#----------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #Date | date | DateTime
#========================================================================================
sub AppointmentDayList { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $Date) = @$aParams;
    my $DayList = $Product->businessHours->dayList($Date);
    foreach my $Entry (@$DayList) {
        @{$Entry->{'Appointments'}} = map { $_->tleHash } @{$Entry->{'Appointments'}}
    }
    return $DayList;
}

#=============================================================================================================
# function     DISPLAYPRICETAGINFOGROSS
# state        public
#-------------------------------------------------------------------------------------------------------------
# syntax       #DISPLAYPRICETAGINFOGROSS(#Product, #TaxModel, #User)
#               #LOCAL("DisplayPriceTagInfoGross", #DISPLAYPRICETAGINFOGROSS(#Product, #Shop.TaxModel, #User))
#                   #IF(#DisplayPriceTagInfoGross) #Shop.PriceTagInfoMultiProductGross #ENDIF
#               #ENDLOCAL
#-------------------------------------------------------------------------------------------------------------
# description  Returns true if gross price tag into needs to be displayed. If user is defined the tax model
#               of the user take effect.
#-------------------------------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #TaxModel | tax calculation model (0=net, 1=gross) | int
# input        #User | (optional) user object | object
#=============================================================================================================
sub DISPLAYPRICETAGINFOGROSS { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $TaxModel, $User) = @$aParams;

    TestObject('Product', $Product);
    TestInteger('TaxModel', $TaxModel, 0, 1);
    TestObject('User', $User, 'DE_EPAGES::Permission::API::Object::User') if defined $User;

    my $Shop = $Product->getSite();

    if (defined $User and not $User->get('IsAnonymous')) {
        my $Customer = GetCustomerForUser ($User, $Shop);
        $TaxModel = $Customer->get('TaxModel') if defined $Customer;
    }

    # return false if the price is only shown for registered user & user is not logged in
    return 0 unless ((not $Shop->get('IsPriceOnlyForRegisteredUser')) or (defined $User and not $User->get('IsAnonymous')));

    return 1 if $TaxModel == TAX_MODEL_GROSS;
}

#=========================================================================================
# function     DISPLAYFREETAXINFO
# state        public
#-----------------------------------------------------------------------------------------
# syntax       #DISPLAYFREETAXINFO($Product, $User)
#               #LOCAL("DisplayFreeTaxInfo", #DISPLAYFREETAXINFO($Product, $User))
#                   #IF(#DisplayFreeTaxInfo) #Shop.PriceTagInfoSingleProductFreeTax #ENDIF
#               #ENDLOCAL
#-----------------------------------------------------------------------------------------
# description  Returns true if the 'free tax' text should be displayed.
#-----------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #User | (optional) user object | object
#=========================================================================================
sub DISPLAYFREETAXINFO {
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $User) = @$aParams;

    TestObject('Product', $Product);
    TestObject('User', $User, 'DE_EPAGES::Permission::API::Object::User') if defined $User;

    return undef unless $Product->instanceOf('Product');

    my $Shop = $Product->getSite();
    my $TaxMatrix = $Shop->get('TaxMatrix');
    my $TaxClass = defined $Product->get('TaxClass') ? $Product->get('TaxClass') : $Product->getWithParent('TaxClass');
    my $TaxArea = $TaxMatrix->defaultTaxArea();

    # if user is a customer and has a own tax area override this tax area
    if (defined $User and not $User->get('IsAnonymous')) {
        my $Customer = GetCustomerForUser($User, $Shop);
        if (defined $Customer and defined $Customer->get('TaxArea')) {
            $TaxArea = $Customer->get('TaxArea');
        }
    }

    # return false if the price is only shown for registered user & user is not logged in
    return 0 unless ((not $Shop->get('IsPriceOnlyForRegisteredUser')) or (defined $User and not $User->get('IsAnonymous')));

    return 1 if $TaxMatrix->taxRate($TaxArea, $TaxClass) == 0;
}

#========================================================================================
# function     APPOINTMENTMENT_PRICE
# state        public
#----------------------------------------------------------------------------------------
# syntax       #APPOINTMENTMENT_PRICE(#Product, #CurrencyID, #TaxModel, #User, #Quantity)
#----------------------------------------------------------------------------------------
# description  Returns true if price tag into needs to be displayed.
#----------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #TaxModel | tax calculation model (0=net, 1=gross) | int
# input        #User | (optional) user object | object
#========================================================================================
sub APPOINTMENTMENT_PRICE { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Appointment, $Product, $CurrencyID, $TaxModel, $User, $Quantity) = @$aParams;
    if(not defined $TaxModel and defined $User) {
        my $Customer = GetCustomerForUser ($User, $Product->getSite());
        $TaxModel = $Customer->get('TaxModel') if defined $Customer;
    }
    my $Price = eval {
        return $Product->appointmentPrice($Appointment, {
            'CurrencyID' => $CurrencyID,
            'TaxModel'   => $TaxModel,
            'User'       => $User,
            'Quantity'   => $Quantity
        });
    };
    if (ExistsError()) {
        my $Error = GetError();
        return undef if $Error->code eq 'ProductNoPrice';
        Error($Error);
    }
    return $Price;
}

sub APPOINTMENT_PAYMENTS { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $User, $CountryID, $CurrencyID) = @$aParams;
    my $Shop = $Product->getSite();

    # find customer groups and taxmodel
    my $CustomerGroup = $Shop->get('NonMemberCustomerGroup');
    my $TaxModel      = $Shop->get('TaxModel');
    if (defined $User) {
        my $Customer = GetCustomerForUser ($User, $Shop);
        if (defined $Customer) {
            $CustomerGroup = $Customer->get('CustomerGroup');
            $TaxModel = $Customer->get('TaxModel')
        }
    }
    # find countryid
    if (not defined $CountryID) {
        ($CountryID) = map { $_->{'CountryID'} } grep { $_->{'IsDefault'} } @{$Shop->get('ShopCountries')};
        $CountryID = $User->get('BillingAddress') if defined $User;
    }
    my $hVars = {
        'CustomerGroup'    => $CustomerGroup,
        'BillingCountryID' => $CountryID,
        'CurrencyID'       => $CurrencyID,
        'TaxModel'         => $TaxModel,
    };
    # grep for region
    return [
        map  { $_->tleHash }
        sort { $a->get('Position')<=>$b->get('Position') }
        grep { $_->canAddToBasket($hVars) }
        @{$Shop->get('PaymentMethodsWithoutShipping')}
    ];
}

sub EXIST_APPOINTMENT_PAYMENT { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $User, $CountryID, $CurrencyID) = @$aParams;
    my $Shop = $Product->getSite();

    # find customer groups and taxmodel
    my $CustomerGroup = $Shop->get('NonMemberCustomerGroup');
    my $TaxModel      = $Shop->get('TaxModel');
    if (defined $User) {
        my $Customer = GetCustomerForUser ($User, $Shop);
        if (defined $Customer) {
            $CustomerGroup = $Customer->get('CustomerGroup');
            $TaxModel = $Customer->get('TaxModel')
        }
    }
    # find countryid
    if (not defined $CountryID) {
        ($CountryID) = map { $_->{'CountryID'} } grep { $_->{'IsDefault'} } @{$Shop->get('ShopCountries')};
        $CountryID = $User->get('BillingAddress') if defined $User;
    }
    my $hVars = {
        'CustomerGroup'    => $CustomerGroup,
        'BillingCountryID' => $CountryID,
        'CurrencyID'       => $CurrencyID,
        'TaxModel'         => $TaxModel,
    };
    # grep for region
    if (scalar (map  { $_->tleHash } sort { $a->get('Position')<=>$b->get('Position') }
        grep { $_->canAddToBasket($hVars) } @{$Shop->get('PaymentMethodsWithoutShipping')}) > 0){

        return 1;
    }
    else{
        return 0;
    }
}

#========================================================================================
# function     PREPAYMENT
# state        public
#----------------------------------------------------------------------------------------
# syntax       #PREPAYMENT(#Product, #Currency, #TaxModel, #ExraPrecision)
# example      #LOCAL("CurrencyID", #INPUT.Currency)
#                 #LOCAL("Prepayment", #PREPAYMENT(#Object, #INPUT.Currency, #Shop.TaxModel, 1))
#                   #IF(#DEFINED(#Prepayment))
#                     #Prepayment[money]
#                   #ENDIF
#                 #ENDLOCAL
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Returns the prepayment of an appointment for a given currency and tax model.
#----------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #Currency | currency code, e.g. "EUR" | string
# input        #TaxModel | tax calculation model (0=net, 1=gross) | int
# input        #ExraPrecision | (optional, default=0) extra precision | int
#========================================================================================
sub PREPAYMENT { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $CurrencyID, $TaxModel, $ExtraPrecision) = @$aParams;

    $ExtraPrecision = 0 unless defined $ExtraPrecision;
    return $Product->getPrepayment($CurrencyID, $TaxModel, $ExtraPrecision);
}

#========================================================================================
# function     GetUserDefinedAttributes
# state        public
#----------------------------------------------------------------------------------------
# syntax       #LOCAL("UserDefinedAttributes",
#                   #FUNCTION("GetUserDefinedAttributes", #Product, #SkipOwnTab)
#               )
# example      #LOCAL("UserDefinedAttributes",
#                   #FUNCTION("GetUserDefinedAttributes", #Product, #SkipOwnTab)
#               )
#                 #LOOP(#UserDefinedAttributes)
#                 #ENDLOOP
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Returns the user defined attributes that are displayed on
#              the "Description" tab on the product page.
#----------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #SkipOwnTab | String | String
#----------------------------------------------------------------------------------------
# return       #Result | array of user defined attributes | array
#========================================================================================
sub GetVisibleUserDefinedAttributes {
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product, $SkipOwnTab) = @$aParams;

    my $aReturnAttributes;

    foreach my $hUserDefAttrValues (@{$Product->get('UserDefinedAttributeValues')}) {
        my $Value = $hUserDefAttrValues->{'Value'};
        next unless defined $Value;

        my $Attribute = $hUserDefAttrValues->{'Attribute'};
        my $Type = $Attribute->type;

        next unless $Attribute->get('IsVisible');
        next if $Type eq 'PreDefCustomerString';

        if ($SkipOwnTab eq 'true') {
            next if $Attribute->get('ShowInOwnTab');
        }
        $hUserDefAttrValues->{'Attribute'} = $Attribute->tleHash();
        if ($Type eq 'PreDefLocalizedString' || $Type eq 'PreDefString') {
            $hUserDefAttrValues->{'Value'} = $Value->tleHash();
        }
        # Array of values has to made into a tle hash to access in the template
        if ($Type eq 'PreDefMultiLocalizedString' || $Type eq 'PreDefMultiString' ) {
            my $aValueArray;
            foreach my $val(@{$hUserDefAttrValues->{'Value'}}) {
                push @$aValueArray, $val->tleHash();
            }
            $hUserDefAttrValues->{'Value'} = $aValueArray;
        }
        push @$aReturnAttributes, $hUserDefAttrValues;
    }
    return $aReturnAttributes;
}

#========================================================================================
# function     UnavailableVariations
# state        public
#----------------------------------------------------------------------------------------
# syntax       #UnavailableVariations(#Product, #SelectParams)
# example      #LOCAL("UnavailableVariationIDs", #UnavailableVariations(#Object, #SelectParams))
#                 #IF(#KEY(#UnavailableVariationIDs, #ID)) ... #ENDIF
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Finds all AttributeValueIDs which are unavailable in context to already
#               selected attribute values and possible unselected ones.
#----------------------------------------------------------------------------------------
# input        #Product | super or sub product object | object
# input        #SelectParams | array of hashes including already selected attribute
#               object id's and predefattribute id's | ref.array.hash
# return       #Result | hash (id => 1) of unavailable attribute value id's | ref.hash
#========================================================================================
sub UnavailableVariations {
    my $self = shift;

    return _UnavailableVariations(\&_CombinationAvailable, @_);
}

#========================================================================================
# function     InvisibleVariations
# state        public
#----------------------------------------------------------------------------------------
# syntax       #InvisibleVariations(#Product, #SelectParams)
# example      #LOCAL("InvisibleVariationIDs", #InvisibleVariations(#Object, #SelectParams))
#                 #IF(#KEY(#InvisibleVariationIDs, #ID)) ... #ENDIF
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Finds all AttributeValueIDs which are invisible in context to already
#               selected attribute values and possible unselected ones.
#----------------------------------------------------------------------------------------
# input        #Product | super or sub product object | object
# input        #SelectParams | array of hashes including already selected attribute
#               object id's and predefattribute id's | ref.array.hash
# return       #Result | hash (id => 1) of invisible attribute value id's | ref.hash
#========================================================================================
sub InvisibleVariations {
    my $self = shift;

    return _UnavailableVariations(\&_CombinationVisible, @_);
}

#========================================================================================
# function     _UnavailableVariations
# state        private
#----------------------------------------------------------------------------------------
# syntax       _UnavailableVariations($CriterionFunctionRef, $TLEProcessor,
#                                      $Product, $ahSelectParams)
# example      _UnavailableVariations(\&_CombinationAvailable, @_)
# example      _UnavailableVariations(\&_CombinationVisible, @_)
#----------------------------------------------------------------------------------------
# description  Finds all AttributeValueIDs which are unavailable in context to already
#               selected attribute values and possible unselected ones based on the passed
#               criterion function.
#----------------------------------------------------------------------------------------
# input        $UnavailabilityCriterion | The criterion (function returning 0 or 1)
#               determining whether a variation is available or not | sub
# input        $Processor | the used TLE Processor | TLEProcessor
# input        $Product | super or sub product object | object
# input        $ahSelectParams | array of hashes including already selected attribute
#               object id's and predefattribute id's | ref.array.hash
# return       hash (id => 1) of unavailable attribute value id's | ref.hash
#========================================================================================
sub _UnavailableVariations {
    my ($UnavailabilityCriterion, $Processor, $aParams) = @_;
    my ($Product, $ahSelectParams) = @$aParams;

    # get all variations
    my $SuperProduct = $Product->get('SuperProduct') // $Product;
    my $SuperProductID = $SuperProduct->id;
    my $ahVariationTypes = $SuperProduct->get('VariationTypeValues');

    #-- Calculate the total number of possible combinations across all variation attributes/values
    my $TotalCombinations = 1;

    # build a variation hash for quicker look-up
    # format: 'AttributeID' => ['AttributeValueID1', 'AttributeValueID2', ...]
    my %VariationTypes = ();
    foreach my $hVariation (@$ahVariationTypes) {
        #-- When only 1 attribute, it still incurs a performance penalty
        $TotalCombinations *= scalar @{$hVariation->{'Values'}} > 1 ? scalar @{$hVariation->{'Values'}} : 1.2;

        $VariationTypes{$hVariation->{'Attribute'}->id} = [ map { $_->id } @{$hVariation->{'Values'}} ];
    }

    #-- When there are more than 60000 possible combinations, the recursive nature of the call below means that this takes an extraordinary
    #-- amount of time to process. This sub can also be called twice, once for unavailablevariations, and once for invisiblevariations
    #-- 60000 is a magic number that we have chosen to strike a balance between reality and performance
    #-- It is (assumed to be) exceptionally rare to have this many possible combination of variations, and in a lot of cases, the variation setup can be tweaked
    #-- to reduce the number of variations used
    #-- UNITY-8910
    return {} if $TotalCombinations > 60000;

    my @VariationTypeAttributeIDs = sort keys %VariationTypes;

    # which attribute value id's are already selected?
    # format: 'AttributeID' => 'AttributeValueID'

    # case 1: nothing selected
    my %SelectedVariations = ();

    # case 2: at least one selected and one not selected
    if ($ahSelectParams) {
        %SelectedVariations = map { $_->{'AttributeID'} => $_->{'AttributeValueID'} } @$ahSelectParams;
    }

    # case 3: everything is selected
    if ($Product->get('SuperProduct')) {
         %SelectedVariations = map { $_->id => $Product->getWithParent($_->alias)->id } map { $_->{'Attribute'} } @$ahVariationTypes;
    }

    my @SelectedAttributeValueIDs = values %SelectedVariations;
    my @SelectedAttributeIDs = keys %SelectedVariations;

    # which attribute id's aren't selected?
    my @UnselectedAttributeIDs = grep { not exists $SelectedVariations{$_} } @VariationTypeAttributeIDs;

    # fetch the mapping (Sub)ProductID => PreDefAttributes
    my $aaResult = DAL_GetVariationAttributesBySuperProductID($SuperProductID, \@VariationTypeAttributeIDs);

    # build matching hash for quick look-up
    # format: 'PreDefAttributeID1,PreDefAttributeID2' => 'SubProductID'
    my %VariationPreDefAttributes;
    foreach my $aRow (@$aaResult) {
        my $SubProductID = shift @$aRow;
        my $PreDefAttributes = join ',', sort(@$aRow);
        $VariationPreDefAttributes{$PreDefAttributes} = $SubProductID;
    }

    my @UnavailableAttributeValueIDs = ();
    my $hCombinations = {};

    # algo: we build several decision trees with selected, unselected attribute value id's in order to the current attribute
    # make an attribute fixed and loop over all possible attribute value id's from all other attributes

    # step 1: check all selected attributes
    foreach my $SelectedAttributeID (@SelectedAttributeIDs) {
        my @OtherAttributeIDs = grep { $_ != $SelectedAttributeID } @SelectedAttributeIDs;
        my @RearrangedAttributeIDs = ($SelectedAttributeID, @OtherAttributeIDs, @UnselectedAttributeIDs);

        $hCombinations = {};
        _CalculateCombinations($UnavailabilityCriterion, [], \@RearrangedAttributeIDs, $hCombinations, \%VariationTypes, \%VariationPreDefAttributes);

        foreach my $AttributeValueIDToTest (@{$VariationTypes{$SelectedAttributeID}}) {
            my $hSelectedCombination = $hCombinations->{$AttributeValueIDToTest};
            my $hOtherSelectedCombination = { %$hSelectedCombination };

            if (scalar @OtherAttributeIDs) {
                foreach my $OtherAttributeID (@OtherAttributeIDs) {
                    my $SelectedAttributeValueIDToTest = $SelectedVariations{$OtherAttributeID};
                    $hOtherSelectedCombination = $hOtherSelectedCombination->{$SelectedAttributeValueIDToTest};

                    push @UnavailableAttributeValueIDs, $AttributeValueIDToTest if $hOtherSelectedCombination->{'possible'} == 0;
                }
            }
            else {
                push @UnavailableAttributeValueIDs, $AttributeValueIDToTest if $hSelectedCombination->{'possible'} == 0;
            }
        }
    }

    # step 2: check all unselected attributes
    foreach my $UnselectedAttributeID (@UnselectedAttributeIDs) {
        my @AttributeValueIDsToTest = @{$VariationTypes{$UnselectedAttributeID}};

        my @OtherUnselectedAttributeIDs = grep { $_ != $UnselectedAttributeID } @UnselectedAttributeIDs;
        my @RearrangedAttributeIDs = (@SelectedAttributeIDs, $UnselectedAttributeID, @OtherUnselectedAttributeIDs);

        $hCombinations = {};
        _CalculateCombinations($UnavailabilityCriterion, [], \@RearrangedAttributeIDs, $hCombinations, \%VariationTypes, \%VariationPreDefAttributes);

        foreach my $AttributeValueIDToTest (@AttributeValueIDsToTest) {

            my $hSelectedCombination = { %$hCombinations };

            # move pointer thru all selected attribute value id's to the current unselected attribute value id
            foreach my $AttributeID (@SelectedAttributeValueIDs, $AttributeValueIDToTest) {
                $hSelectedCombination = { %$hSelectedCombination }->{$AttributeID};
            }

            push @UnavailableAttributeValueIDs, $AttributeValueIDToTest if $hSelectedCombination->{'possible'} == 0;
        }
    }

    return { map { $_ => 1 } @UnavailableAttributeValueIDs };
}

sub _CalculateCombinations {
    my ($UnavailabilityCriterion, $aFixedAttributeValueIDs, $aLoopableAttributesIDs, $hCombinations, $hVariationTypes, $hVariationPreDefAttributes) = @_;

    if (scalar @$aLoopableAttributesIDs == 0) {
        # rec. end
        $hCombinations->{'possible'} = $UnavailabilityCriterion->($aFixedAttributeValueIDs, $hVariationPreDefAttributes);
    }
    else {
        # rec. step down
        my @LoopableAttributesIDs = @$aLoopableAttributesIDs;
        my $AttributeID = shift @LoopableAttributesIDs;
        my $aAttributeValueIDs = $hVariationTypes->{$AttributeID};

        foreach my $AttributeValueID (@$aAttributeValueIDs) {
            my $aNewFixedAttributeValueIDs = [@$aFixedAttributeValueIDs, $AttributeValueID];
            $hCombinations->{$AttributeValueID} = {};

            _CalculateCombinations($UnavailabilityCriterion, $aNewFixedAttributeValueIDs, \@LoopableAttributesIDs, $hCombinations->{$AttributeValueID}, $hVariationTypes, $hVariationPreDefAttributes);

            # rec. step up
            my $Possible = 0;
            while (my ($Key, $hValue) = each %$hCombinations) {
                $Possible = 1 if ($Key =~ /^\d+$/ and $hValue->{'possible'});
            }

            $hCombinations->{'possible'} = $Possible;
        }
    }
}

sub _CombinationAvailable {
    my ($aAttributeValueIDs, $hVariationPreDefAttributes) = @_;

    my $TestPreDefAttributes = join ',', sort(@$aAttributeValueIDs);

    if (defined $hVariationPreDefAttributes->{$TestPreDefAttributes}) {
        # test if subproduct can be added to the basket
        my $SubProduct = LoadObject($hVariationPreDefAttributes->{$TestPreDefAttributes});

        return $SubProduct->canAddToBasket;
    }

    return 0;
}

#========================================================================================
# function     _CombinationVisible
# state        private
#----------------------------------------------------------------------------------------
# syntax       _CombinationVisible($aFixedAttributeValueIDs, $hVariationPreDefAttributes)
#----------------------------------------------------------------------------------------
# description  Checks whether a subproduct with the given variation attributes is
#               visible.
#----------------------------------------------------------------------------------------
# input        $aAttributeValueIDs | Variation attributes | ref.array.integer
# input        $hVariationPreDefAttributes | Hash mapping a variation attribute combination
#               to a SubProductID | ref.hash
# return       1 if the combination is visible, 0 otherwise | integer
#========================================================================================
sub _CombinationVisible {
    my ($aAttributeValueIDs, $hVariationPreDefAttributes) = @_;

    my $TestPreDefAttributes = join ',', sort(@$aAttributeValueIDs);

    if (defined $hVariationPreDefAttributes->{$TestPreDefAttributes}) {
        my $SubProduct = LoadObject($hVariationPreDefAttributes->{$TestPreDefAttributes});

        return $SubProduct->get('IsVisible');
    }

    return 0;
}

#========================================================================================
# function     APPOINTMENTMENT_PREPAYMENT
# state        public
#----------------------------------------------------------------------------------------
# syntax       #APPOINTMENTMENT_PREPAYMENT(#Product, #CurrencyID, #TaxModel, #User, #Quantity)
#----------------------------------------------------------------------------------------
# description  Returns the value for prepayment.
#----------------------------------------------------------------------------------------
# input        #Product | product object | object
# input        #CurrencyID | currency code, e.g. "EUR" | string
# input        #TaxModel | tax calculation model (0=net, 1=gross) | int
# input        #User | (optional) user object | object
# input        #Quantity | quantity | integer
#========================================================================================
sub APPOINTMENTMENT_PREPAYMENT { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Appointment, $Product, $CurrencyID, $TaxModel, $User, $Quantity) = @$aParams;
    if(not defined $TaxModel and defined $User) {
        my $Customer = GetCustomerForUser ($User, $Product->getSite());
        $TaxModel = $Customer->get('TaxModel') if defined $Customer;
    }
    my $Price = eval {
        return $Product->appointmentPrepayment($Appointment, {
            'CurrencyID' => $CurrencyID,
            'TaxModel'   => $TaxModel,
            'User'       => $User,
            'Quantity'   => $Quantity
        });
    };
    if (ExistsError()) {
        my $Error = GetError();
        return undef if $Error->code eq 'ProductNoPrice';
        Error($Error);
    }
    return $Price;
}

#========================================================================================
# function     CANDISPLAYFROMPRICE
# state        public
#----------------------------------------------------------------------------------------
# syntax       #CANDISPLAYFROMPRICE(#Product)
#----------------------------------------------------------------------------------------
# description  Returns true if "from prices" can be displayed.
#               - VariationPriceFromText must be set
#               - product must be the default sub product or
#                 product must be the master product and a available, cheaper sub product
#                 with an own price must exists
#----------------------------------------------------------------------------------------
# input        #Product | product object | object
#========================================================================================
sub CANDISPLAYFROMPRICE { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product) = @$aParams;

    return 0 unless defined $Product->getSite()->get('VariationPriceFromText');
    # test if we have a default sub product
    $Product = $Product->superProduct if $Product->get('IsDefault');
    # test if we have a master and found an sub product with own prices
    # note: existSubProductPrices returns false if $Product is not a master
    return $Product->existSubProductPrices ? 1 : 0;
}

#========================================================================================
# function     ProductURL
# state        public
#----------------------------------------------------------------------------------------
# syntax       #ProductURL(#Product)
#----------------------------------------------------------------------------------------
# description  Returns the URL to the product detail page.
#               If this function is called from a search results, additional paramters
#               are passed to the detail page in order to select the appropriate image
#               and variation values in case of a variation product.
#----------------------------------------------------------------------------------------
# input        #Product | product object | object
#========================================================================================
sub ProductURL { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Product) = @$aParams;

    my @Parameters;
    my $INPUT = $Processor->getInputContext;
    if( defined $INPUT->{'SearchParams'}) {
        push @Parameters, 'ViewAction=ViewProduct';
        push @Parameters, 'SearchParams='.UrlEncodeValue($INPUT->{'SearchParams'});
    }
    else {
        # from a category always link to default sub product instead of master
        $Product = $Product->defaultObject;
    }

    my $URL = '?ObjectPath='.$Processor->formatTLE('Path', 'url', $Product->get('Path'));
    return join('&', $URL, @Parameters);
}

1;
