#========================================================================================
# package      DE_EPAGES::Shop::API::TLE::WidgetHandler
# state        public
#----------------------------------------------------------------------------------------
# description  Implements system-wide TLE functions for dashboard widgets
#               <ul>
#                   <li>GroupWidgets</li>
#               </ul>
#========================================================================================
package DE_EPAGES::Shop::API::TLE::WidgetHandler;

use strict;
use DE_EPAGES::Core::API::Log qw( GetLog );
use DE_EPAGES::Core::API::PerlTools qw ( ElementOf );
use DE_EPAGES::Object::API::Factory qw( LoadObject );

#========================================================================================
# function     RegisterHandlerProc
# state        private
#----------------------------------------------------------------------------------------
# syntax       $RegisterHandlerProc( $hParams );
#----------------------------------------------------------------------------------------
# description  this function registers this attribute handle on the processor via hook
#               function on hook TLEProcessorRegistration
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments keys :
#               <ul>
#                   <li>Processor - tle processor - object
#               </ul> | ref.hash
#========================================================================================
sub RegisterHandlerProc {
    my ($Params) = @_;
    GetLog->debug( 'WidgetHandler.RegisterHandlerProc' );

    __PACKAGE__->new()->register( $Params->{'Processor'} );
    return;
}

#========================================================================================
# function     new
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler = new DE_EPAGES::Object::API::TLE::WidgetHandler;
# example      my $PrefixHandler  = new DE_EPAGES::Object::API::TLE::WidgetHandler;
#----------------------------------------------------------------------------------------
# description  create a new handler
#----------------------------------------------------------------------------------------
# return       $Handler | blessed object | object
#========================================================================================
sub new { bless {}, shift; }

#========================================================================================
# function     register
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler->register( $Processor );
#----------------------------------------------------------------------------------------
# description  this function register this handle on the processor
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
#========================================================================================
sub register {
    my $self = shift;
    my ($Processor) = @_;
    GetLog->debug( 'WidgetHandler.register' );

    $Processor->registerHandler('FunctionHandler', $self, 'GroupWidgets');
    return;
}

#========================================================================================
# function     GroupWidgets
# state        public
#----------------------------------------------------------------------------------------
# syntax       #GroupWidgets(#Shop, #WidgetGroup)
# example      #LOOP(#GroupWidgets(#Shop.Object, #Group.Object))
#                   #Alias<br />
#               #ENDLOOP
#----------------------------------------------------------------------------------------
# description  Returns the widget types assigned to a group depending on the shops features.
#----------------------------------------------------------------------------------------
# input        #Shop | shop object | object
# input        #WidgetGroup | widget group object | object
# return       #WigetsTypes | list of all widget types assigned to the group the shop
#                   has the feature for | ref.array.hash
#========================================================================================
sub GroupWidgets {
    my $self = shift;
    my ($Processor, $aParameter) = @_;
    my ($ShopID, $GroupID) = @$aParameter;
    return 'NoShop' unless defined $ShopID;
    return 'NoGroup' unless defined $GroupID;
    my $Shop = LoadObject($ShopID);
    my $Group = LoadObject($GroupID);

    my @ProhibitedWidgetTypesForCMS = qw (LastRegistrations MediaGalleryWidget);
    my $eCMSFeatureAvailable = $Shop->featureMaxValue('eCMS');

    return [
        map { $_->tleHash }
        grep { !($eCMSFeatureAvailable  && ElementOf($_->alias, @ProhibitedWidgetTypesForCMS)) }
        grep { not defined $_->get('Feature') or $Shop->featureMaxValue($_->get('Feature')) }
        @{ $Group->get('WidgetTypes') }
     ];
}

1;
