#========================================================================================
# package      DE_EPAGES::ShopMigration::API::ProductsCommon
# state        public
#----------------------------------------------------------------------------------------
# description common functions for exporting/importing products
#========================================================================================
package DE_EPAGES::ShopMigration::API::ProductsCommon;
use base qw(Exporter);

use strict;
use 5.010; # Needed because we use state vars

our @EXPORT_OK = qw(
    GetProductsIterator
    GetResetProductsIterator
    GetProducts
    SetProducts
    GetProductData
    SetProductData
    ExportProductFiles
    ImportProductFiles
);

use File::Spec::Functions;
use File::Basename qw(basename dirname);
use HTML::FormatText;
use HTML::TreeBuilder;
use DE_EPAGES::Core::API::File qw(AbsolutePath CopyFile ExistsDirectory ExistsFile
    RunOnDirectory RelativePath GetFileNames MoveFile fsplit);
use DE_EPAGES::Core::API::Error qw(Error ExistsError GetError);
use DE_EPAGES::Core::API::Image qw(TestImageFormat);
use DE_EPAGES::Core::API::Log qw(LogDebug);
use DE_EPAGES::Core::API::Object::DateTimeFormatter;
use DE_EPAGES::Core::API::PerlTools qw(ElementOf trim GetEpochAsMilli unique);
use DE_EPAGES::Core::API::Url qw(StringIsURL);
use DE_EPAGES::Core::API::Warning qw(Warning);
use DE_EPAGES::Core::API::TestParameter qw(TestString);
use DE_EPAGES::Database::API::Config qw(DatabaseConnectionData);
use DE_EPAGES::Database::API::Connection qw(RunOnStore GetCurrentDBHandle);
use DE_EPAGES::Object::API::Domain qw(GetUnitySubDomain);
use DE_EPAGES::Object::API::Factory qw(
    ExistsObjectByPath LoadObjectByPath Transaction LoadClassByAlias ExistsObject LoadObject LoadRootObject
);
use DE_EPAGES::Object::API::Language qw(GetPKeyLanguageByCode GetCodeByLanguageID);
use DE_EPAGES::Product::API::Constants qw(UNITY_IMAGE_ATTRIBUTES);
use DE_EPAGES::Product::API::Product qw(UsedVariationValues ScaleImageSizes);             ##no critic (RequireCorrectImports)
use DE_EPAGES::Product::DAL::Product qw(DAL_GetAllProductIDsAliasOrderedByShopID);        ##no critic (RequireCorrectImports)
use DE_EPAGES::Product::API::Shop qw (GetProductImageSizes);
use DE_EPAGES::Product::API::Constants qw(PRODUCT_IMAGE_SIZE_NAMES PRODUCT_IMAGES_SIZES); ##no critic (RequireCorrectImports)
use DE_EPAGES::ShopMigration::API::Cartridge;
use DE_EPAGES::ShopMigration::API::Constants qw(
    PRODUCTS_ATTRIBUTES_FOR_PREP_EXPORT
    PRODUCTS_IMAGE_ATTRIBUTES
    PRODUCTS_OBJECTATTRIBUTES_FOR_PREP_EXPORT
    PRODUCTS_LOCALIZEDATTRIBUTES_FOR_PREP_EXPORT
    PRODUCTS_ATTRIBUTES_FOR_BASE_EXPORT
    PRODUCTS_OBJECTATTRIBUTES_FOR_BASE_EXPORT
    PRODUCTS_LOCALIZEDATTRIBUTES_FOR_BASE_EXPORT
    PRODUCTS_SCRIBEATTRIBUTES_FOR_BASE_EXPORT
    PRODUCTS_FILE_ATTRIBUTES
    PRODUCTTYPE_SUPPORTED_ATTRIBUTE_TYPES
    PRODUCTS_ATTRIBUTES_DATETIME
    $EXPORT_FILES_AND_FILEATTRIBUTES
    PRODUCTS_SCALE_IMAGE_ATTRIBUTES
    MAX_IMAGE_FILENAME_LENGTH
);
use DE_EPAGES::ShopMigration::API::HTMLSanitizer qw(SanitizeHTMLForScribe);
use DE_EPAGES::ShopMigration::API::JSON qw( ParseDateTime FormatDateTime );
use DE_EPAGES::ShopMigration::API::ShopCommonData qw(GetActiveShopLanguages);
use DE_EPAGES::ShopMigration::API::ShortURL qw(InsertRedirect GetAllUILocalesForShop);
use DE_EPAGES::Object::API::Table::Locale qw(GetLanguageByLocaleID);
our $DEBUG = 0;
our $SKIP_UI = 0;

# - we need to cache all renamed images for a master and its subs as we only have a per-product view
# during _createScaledMediumImages() meaning we see each subproduct as-is and each of them can have
# a different set of selected images in the slideshow - we don't want to create shortened images
# for every subproduct ten times over
# - the hash is cleared on every new master
my %RenamedFiles = ();

#ERRORDEFINITION
sub ErrorDefinition {
    return {
        'REFERENCED_OBJECT_NOT_FOUND' => 'Cannot assign value to attribute #Name as #Path was not found below root or shop',
        'UI_STORE_NOT_DEFINED'        => 'InstanceUnityUI is not defined in Database.d/#Store#SPACE[0].conf.',
    };
}
#ERRORDEFINITION END

##no critic qw(ProhibitCascadingIfElse ProhibitDeepNests)

#========================================================================================
# function     GetProductsIterator
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $Iterator = GetProductsIterator($Shop);
#----------------------------------------------------------------------------------------
# description  gets an iterator that traverses all products
#               excluding appointments, bundles, customizable and download products
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
#----------------------------------------------------------------------------------------
# return       $Iterator | iterator | iterator
#========================================================================================
sub GetProductsIterator {
    my ($Shop) = @_;

    my $Iterator = DE_EPAGES::Core::API::Iterator->wrap(
        Source => _getExportableProductIDsAliasOrderedByShopID($Shop->id),
        Handle => sub {
            return $_[1]->[0];
        }
    );
    return $Iterator;
}

#========================================================================================
# function     GetResetProductsIterator
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $Iterator = GetResetProductsIterator($Shop);
#----------------------------------------------------------------------------------------
# description  gets an iterator that traverses all products
#               that are appointments, bundles, customizable or download products
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
#----------------------------------------------------------------------------------------
# return       $Iterator | iterator | iterator
#========================================================================================
sub GetResetProductsIterator {
    my ($Shop) = @_;

    my $Iterator = DE_EPAGES::Core::API::Iterator->wrap(
        Source => _getResetProductIDsAliasOrderedByShopID($Shop->id),
        Handle => sub {
            return $_[1]->[0];
        }
    );
    return $Iterator;
}

#========================================================================================
# function     _getExportableProductIDsAliasOrderedByShopID
#----------------------------------------------------------------------------------------
# syntax       $Iterator = _getExportableProductIDsAliasOrderedByShopID($ShopID);
#----------------------------------------------------------------------------------------
# description  Returns an iterator over the IDs of all shop products ordered by their alias
#               with the restriction that super products have to be in front of their subs
#               and excludes non-migratable items, that is:
#               appointments, bundles, customizable and download products
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id of the products | integer
# return       $Iterator | object iterator | DE_EPAGES::Core::API::Iterator
#========================================================================================
sub _getExportableProductIDsAliasOrderedByShopID {
    my ($ShopID) = @_;
    my $dbi = GetCurrentDBHandle();

    my $Iterator = $dbi->iterator(<<END_SQL , [ $ShopID, $ShopID ]);
SELECT productid FROM
(
    (SELECT sub.productid, subo.alias, po.alias AS orderBase, 20 AS orderPos
     FROM product p
     INNER JOIN object po ON po.objectid = p.productid
     INNER JOIN product sub ON sub.superproductid = p.productid
     INNER JOIN object subo ON subo.objectid = sub.productid
     WHERE p.superproductid IS NULL
      AND p.shopid = ?
      AND p.isappointment = 0
      AND p.iscustomizable = 0
      AND p.isdownloadproduct = 0
    )
    UNION
    (SELECT p.productid, po.alias, po.alias AS orderBase, 10 AS orderPos
     FROM product p
     INNER JOIN object po ON po.objectid = p.productid
     WHERE p.superproductid IS NULL
      AND p.shopid = ?
      AND p.isappointment = 0
      AND NOT EXISTS ( SELECT b.bundleproductid from bundleproductmap b where b.bundleproductid=p.productid)
      AND p.iscustomizable = 0
      AND p.isdownloadproduct = 0
    )
) p
ORDER BY orderBase, orderPos, alias
END_SQL

    return $Iterator;
}

#========================================================================================
# function     _getResetProductIDsAliasOrderedByShopID
#----------------------------------------------------------------------------------------
# syntax       $Iterator = _getResetProductIDsAliasOrderedByShopID($ShopID);
#----------------------------------------------------------------------------------------
# description  Returns an iterator over the IDs of products that need to be removed
#               in the base shop on final migration, that is:
#               appointments, bundles, customizable and download products
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id of the products | integer
# return       $Iterator | object iterator | DE_EPAGES::Core::API::Iterator
#========================================================================================
sub _getResetProductIDsAliasOrderedByShopID {
    my ($ShopID) = @_;
    my $dbi = GetCurrentDBHandle();

    my $Iterator = $dbi->iterator(<<END_SQL , [ $ShopID, $ShopID ]);
SELECT productid FROM
(
    (SELECT sub.productid, subo.alias, po.alias AS orderBase, 20 AS orderPos
     FROM product p
     INNER JOIN object po ON po.objectid = p.productid
     INNER JOIN product sub ON sub.superproductid = p.productid
     INNER JOIN object subo ON subo.objectid = sub.productid
     WHERE p.superproductid IS NULL
      AND p.shopid = ?
      AND ( p.isappointment = 1 OR
            p.iscustomizable = 1 OR
            p.isdownloadproduct = 1 )
        )
    UNION
    (SELECT p.productid, po.alias, po.alias AS orderBase, 10 AS orderPos
     FROM product p
     INNER JOIN object po ON po.objectid = p.productid
     WHERE p.superproductid IS NULL
      AND p.shopid = ?
      AND ( p.isappointment = 1 OR
            p.iscustomizable = 1 OR
            p.isdownloadproduct = 1 OR
            EXISTS ( SELECT b.bundleproductid from bundleproductmap b where b.bundleproductid=p.productid)
        )
    )
) p
ORDER BY orderBase, orderPos, alias
END_SQL

    return $Iterator;
}

#========================================================================================
# function     GetProducts
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $ahProducts = GetProducts($Shop);
#----------------------------------------------------------------------------------------
# description  gets all relevant data of products of a shop
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
#----------------------------------------------------------------------------------------
# return       $ahProducts | see GetProductData() | reference to array of hashes
#========================================================================================
sub GetProducts {
    my ($Shop) = @_;

    my @Items;
    my $Iterator = GetProductsIterator();
    while (my $ObjectID = <$Iterator>) {
        if (ExistsObject($ObjectID)) {
            my $Product = LoadObject($ObjectID);
            my $hInfo = GetProductData($Shop, $Product);
            push @Items, $hInfo if (keys %$hInfo);
        }
    }
    return \@Items;
}

#========================================================================================
# function     GetProductData
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $hProdcut = GetProductData($Shop, $Product);
#----------------------------------------------------------------------------------------
# description  Constructs a hash with the exportable fields for a single Product
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
# input        $Product | product | object
#----------------------------------------------------------------------------------------
# return       $hProduct | hash with following fields (scalars are converted to strings!)
#               <ul>
#                   <li>Attributes
#                       <ul>
#                           <li>Alias                             - string</li>
#                           <li>Position                          - integer</li>
#                           <li>Class                             - string</li>
#                           <li>Parent                            - string</li>
#                           <li>IsVisible                         - boolean</li>
#                           <li>TaxClass                          - string</li>
#                           <li>WeightUnit                        - string</li>
#                           <li>Weight                            - integer</li>
#                           <li>RefUnit                           - string</li>
#                           <li>RefAmount                         - integer</li>
#                           <li>RefContentAmount                  - integer</li>
#                           <li>Manufacturer                      - string</li>
#                           <li>ManufacturerSKU                   - string</li>
#                           <li>DeliveryPeriod                    - string</li>
#                           <li>Length                            - integer</li>
#                           <li>Height                            - integer</li>
#                           <li>Width                             - integer</li>
#                           <li>ImageLarge                        - string</li>
#                           <li>SuperProduct                      - string</li>
#                           <li>HasSubOwnPrices                   - boolean (always 1)</li>
#                           <li>UPCEAN                            - string</li>
#                           <li>ImagesSlideShowString             - string</li>
#                           <ol>ListPrices                        - ref array
#                               <ul>
#                                   <li> CurrencyID               - string</li>
#                                   <li> Deposit                  - float</li>
#                                   <li> EcoParticipation         - float</li>
#                                   <li> ListPrice                - float</li>
#                                   <li> ManufacturerPrice        - float</li>
#                               </ul>
#                           </ol>
#                           <li>IsDefault                         - boolean</li>
#                           <li>TaxClassDigital                   - string</li>
#                           <li>EnergyLabelsString                - string</li>
#                           <li>EnergyLabelSourceFile             - string</li>
#                           <li>EnergyLabelHighestEfficiencyClass - string</li>
#                           <li>SelectedVariations                - array of aliases</li>
#                       </ul>
#                   </li>
#                   <li>LocalizedAttributes
#                       <ul>
#                           <li>LanguageCode2            - en|de|..</li>
#                               <ul>
#                                   <li>Name                  - string</li>
#                                   <li>Keywords              - string</li>
#                                   <li>MainCharacteristics   - string</li>
#                                   <li>_ProductDataSheet     - string</li>
#                                   <li>YoutubeVideoId        - string (Prep Export only)</li>
#                                   <li>ProductVideoProvider  - string (Prep Export only)</li>
#                               </ul>
#                           </li>
#                       </ul>
#                   </li>
#               </ul>| reference to hash
#========================================================================================
sub GetProductData {
    my ($Shop, $Product) = @_;

    my $UsePrepAttributes = $Shop->get('IsMigrationPrepShop');
    my ($Attributes, $ObjectAttributes, $LocalizedAttributes, $ScribeAttributes) =
        _selectAttributesArrayRefsForProductsData($UsePrepAttributes);

    my $hValues = { Attributes => $Product->getWithParent($Attributes), LocalizedAttributes => {}, };
    # only get the ListPrices for the active taxmodel/currency instead of all
    $hValues->{Attributes}{ListPrices} = _getPrices($Shop, $Product);
    $hValues->{Attributes}{HasSubOwnPrices} = '1';
    $hValues->{Attributes}{RelativeFilePathPart} = _GetRelativeFilePathPart($Product, $Shop);
    # sometimes fields get treated as Integer instead of String thus mixing up the json file
    foreach my $AttributeName (keys %{$hValues->{Attributes}}) {
        if (defined $hValues->{Attributes}{$AttributeName} and not ref $hValues->{Attributes}{$AttributeName}) {
            $hValues->{Attributes}{$AttributeName} .= ''
        }
    }

    # set default value for OrderUnits
    my $OrderUnitPath = 'Units/piece';

    my $ProductOrderUnit = $Product->get('OrderUnit');
    if (defined $ProductOrderUnit) {
        $OrderUnitPath = $ProductOrderUnit->get('Path');
    } elsif (defined $Product->get('SuperProduct')) {
        # OrderUnits: If product has SuperProduct and no own OrderUnit, use OrderUnit from SuperProduct else
        # we use the default value for OrderUnits
        my $SuperProduct = $Product->get('SuperProduct');
        my $SuperProductOrderUnit = $SuperProduct->get('OrderUnit');
        if (defined $SuperProductOrderUnit) {
            $OrderUnitPath = $SuperProductOrderUnit->get('Path');
        }
    }

    $hValues->{Attributes}{OrderUnit} = LoadObjectByPath($OrderUnitPath);
    $hValues->{Attributes}{MinOrder} = $Product->get('MinOrder');
    $hValues->{Attributes}{PriceQuantity} = $Product->get('PriceQuantity') // '1';
    $hValues->{Attributes}{IntervalOrder} = $Product->get('IntervalOrder');
    # remove attributes with invalid image types
    foreach my $AttributeName (@{PRODUCTS_IMAGE_ATTRIBUTES()}) {
        if (defined $hValues->{Attributes}{$AttributeName}) {
            if ($AttributeName eq 'ImagesSlideShowString') {
                my @ValidImages = ();
                my @ImagesInString = trim(split /;/, $hValues->{Attributes}{$AttributeName});
                foreach my $FileName (@ImagesInString) {
                    if (StringIsURL($FileName)) {
                        push @ValidImages, $FileName;
                    } else {
                        eval {TestImageFormat($AttributeName, $FileName)};
                        if (ExistsError()) {
                            LogDebug(sprintf('Removing unsupported image %s on %s (%s)',
                                $FileName, $Product->pathString, $AttributeName), 1);
                        } else {
                            $FileName =~ s/_/-/g unless $Shop->get('IsMigrationPrepShop');
                            push @ValidImages, $FileName;
                        }
                    }
                }
                $hValues->{Attributes}{$AttributeName} = join(';', @ValidImages);
            } else {
                eval {TestImageFormat($AttributeName, $hValues->{Attributes}{$AttributeName})};
                if (ExistsError()) {
                    LogDebug(sprintf('Removing unsupported image %s on %s (%s)',
                        $hValues->{Attributes}{$AttributeName}, $Product->pathString, $AttributeName), 1);
                    $hValues->{Attributes}{$AttributeName} = undef;
                } else {
                    $hValues->{Attributes}{$AttributeName} =~ s/_/-/g unless $Shop->get('IsMigrationPrepShop');
                }
            }
        }
    };

    # UNITY-7460 export SelectedVariations as array of aliases for subproducts
    if ( exists($hValues->{Attributes}{SelectedVariations}) ) {
        my $SuperProduct = $Product->superProduct;
        my $aValues = [];
        if (defined $SuperProduct) {
            my @VariationTypes = sort map {$_->{'Attribute'}->alias} @{$SuperProduct->get('VariationTypeValues')};
            $aValues = \@VariationTypes;
        }
        $hValues->{Attributes}{SelectedVariations} = $aValues;
    }

    my $ProductClass = $Shop->get('ProductClass');
    $hValues->{Attributes}{Class} = $hValues->{Attributes}{Class} == $ProductClass ? undef : $Product->class->alias;
    foreach my $AttributeName (@$ObjectAttributes) {
        my $Value = $hValues->{Attributes}{$AttributeName};
        next if (not defined $Value or not ref $Value);
        if ($AttributeName eq 'TaxClass') {
            $hValues->{Attributes}{$AttributeName} = $Value->alias;
        } elsif ($AttributeName eq 'TaxClassDigital') {
            # UNITY-6872 : Non-download product with TaxClassDigital set
            $hValues->{Attributes}{$AttributeName} = undef;
        } else {
            $hValues->{Attributes}{$AttributeName} = $Shop->relativePath($Value);
        }
    }

    my $ActiveLanguages = GetActiveShopLanguages($Shop);
    foreach my $hLanguage (@$ActiveLanguages) {
        my $AttrInfo = $Product->getWithParent($LocalizedAttributes, $hLanguage->{'LanguageID'});
        # merge Text + Description on Base2Now export and delete "Text" as it is used as MetaDescription
        if (!$UsePrepAttributes) {
            my @Fields = qw(Description Text);
            my $NowDescription = sprintf('%s%s', @$AttrInfo{ @Fields });
            @{$AttrInfo}{@Fields} = ($NowDescription, undef);
            foreach my $FieldName (@$ScribeAttributes) {
                $AttrInfo->{$FieldName} = SanitizeHTMLForScribe($AttrInfo->{$FieldName});
            }
        }
        $hValues->{LocalizedAttributes}{$hLanguage->{'Code2'}} = $AttrInfo;
    }
    my $ProductTypeAttrInfo = _getProductUserDefinedAttributeValues($Shop, $Product);
    foreach my $Type (keys %$ProductTypeAttrInfo) {
        if ($Type eq 'LocalizedAttributes') {
            foreach my $Lang (keys %{$ProductTypeAttrInfo->{$Type}}) {
                foreach my $Entry (keys %{$ProductTypeAttrInfo->{$Type}{$Lang}}) {
                    $hValues->{$Type}{$Lang}{$Entry} = $ProductTypeAttrInfo->{$Type}{$Lang}{$Entry};
                }
            }
        } else {
            foreach my $Entry (keys %{$ProductTypeAttrInfo->{$Type}}) {
                $hValues->{$Type}{$Entry} = $ProductTypeAttrInfo->{$Type}{$Entry};
            }
        }
    }

    # UNITY-8409: Format DateTimes to default epages string "2009-08-28T00:00:00.00" in JSON
    my @DateTimeAttributes = @{PRODUCTS_ATTRIBUTES_DATETIME()};
    foreach my $AttributeName (@DateTimeAttributes) {
        if (defined $hValues->{Attributes}{$AttributeName}) {
            $hValues->{Attributes}{$AttributeName} = FormatDateTime($hValues->{Attributes}{$AttributeName});
        }
    }

    _convertScalarValuesToStringInHash($hValues);
    return $hValues;
}

sub _GetRelativeFilePathPart {
    my ($Object, $Site) = @_;

    my $FilePathPart = $Object->get('FilePathPart');
    $FilePathPart = RelativePath($Site->get('FilePath'), $Object->get('FilePath'))
        unless defined $FilePathPart;

    return $FilePathPart;
}

#========================================================================================
# function     _getPrices
# state        private
#----------------------------------------------------------------------------------------
# syntax       _getPrices($Shop, $Product)
#----------------------------------------------------------------------------------------
# description  constructs a array wit relevant prices (ListPrices, ManufacturerPrices
#               and so forth) for the given product as a hash (currently only for
#               the shop currency, but this can later be extended if needed
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
# input        $Product | product | object
#----------------------------------------------------------------------------------------
# return       $ahListPrices | array of hash with following fields | ref array hash
#               <ol>
#                   <ul>
#                       <li>CurrencyID           - string</li>
#                       <li>Deposit              - money (converted to String)</li>
#                       <li>EcoParticipation     - money (converted to String)</li>
#                       <li>ManufacturerPrice    - money (converted to String)</li>
#                       <li>ListPrice            - money (converted to String)</li>
#                    </ul>
#               </ol>
#========================================================================================
sub _getPrices {
    my ($Shop, $Product) = @_;

    my $CurrencyID = $Shop->get('CurrencyID');
    my $ahListPrices = [];
    my $hEntry = {
        CurrencyID        => $CurrencyID,
        Deposit           => $Product->getDeposit($CurrencyID),
        EcoParticipation  => $Product->getEcoParticipation($CurrencyID),
        ManufacturerPrice => $Product->getManufacturerPrice($CurrencyID,undef,1),
        ListPrice         => $Product->getListPrice($CurrencyID,undef,1),
    };
    my $SuperProduct = $Product->superProduct;
    if (defined $SuperProduct) {
        $hEntry->{Deposit} //= $SuperProduct->getDeposit($CurrencyID);
        $hEntry->{EcoParticipation} //= $SuperProduct->getEcoParticipation($CurrencyID);
        $hEntry->{ManufacturerPrice} //= $SuperProduct->getManufacturerPrice($CurrencyID,undef,1);
        $hEntry->{ListPrice} //= $SuperProduct->getListPrice($CurrencyID,undef,1);
    }
    foreach my $Field (keys %{$hEntry}) {
        $hEntry->{$Field} .= '' if (defined $hEntry->{$Field} and not ref $hEntry->{$Field});
    }
    push @$ahListPrices, $hEntry;
    return $ahListPrices;
}

#========================================================================================
# function     _convertScalarValuesToStringInHash
# state        private
#----------------------------------------------------------------------------------------
# syntax       _convertScalarValuesToStringInHash($hValues);
#----------------------------------------------------------------------------------------
# description  converts scalar values in the both hash lookups (Attributes and
#               LocalizedAttributes) to string values to have a consistent JSON quoting
#----------------------------------------------------------------------------------------
# input        $hValues         | Hash | ref hash
#========================================================================================
sub _convertScalarValuesToStringInHash {
    my ($hValues) = @_;

    # sometimes fields get treated as Integer instead of String thus mixing up the json file
    foreach my $Lookup (qw(Attributes LocalizedAttributes)) {
        foreach my $AttributeName (keys %{$hValues->{$Lookup}}) {
            if (defined $hValues->{$Lookup}{$AttributeName} and not ref $hValues->{$Lookup}{$AttributeName}) {
                $hValues->{$Lookup}{$AttributeName} .= ''
            }
        }
    }
}

#========================================================================================
# function     _getUserDefinedAttributeData
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $hAttributeData = _getUserDefinedAttributeData($Shop);
#----------------------------------------------------------------------------------------
# description  returns a hash with Type, IsHMTL and Alias of user-defined
#               attributes for all product types in the shop, the result is kept as
#               state variable so subsequent calls don't rebuild the hash
#----------------------------------------------------------------------------------------
# input        $Shop              | baseshop or prepshop | object
# return       $hAttributesData   | data for all user-defined product attributes | ref hash
#========================================================================================
sub _getUserDefinedAttributeData {
    my ($Shop) = @_;

    my @SupportedDefaultTypes = @{PRODUCTTYPE_SUPPORTED_ATTRIBUTE_TYPES()};
    my $hLookup = {};
    my $ProductTypes = $Shop->child('ProductTypes')->children();
    foreach my $ProductType ($Shop->get('ProductClass'), @$ProductTypes) {
        next unless $ProductType->class->alias eq 'ProductType';
        my @AttributeDetails = ();
        foreach my $Attribute (sort {$a->alias cmp $b->alias} @{$ProductType->get('UserDefinedAttributes')}) {
            my $Type = $Attribute->get('Type');
            my $Alias = $Attribute->alias;
            if (ElementOf($Type, @SupportedDefaultTypes)) {
                push @AttributeDetails, { Alias => $Alias, Type => $Type, IsHTML => $Attribute->get('IsHTML') };
            } else {
                LogDebug(sprintf('ProductType %s has unsupported type %s on ProductTypeAttribute %s',
                    $ProductType->alias, $Type, $Alias), 1);
            }
        }
        $hLookup->{$ProductType->id} = \@AttributeDetails;
    }
    my $hAttributesData = { $Shop->alias => $hLookup };
    return $hAttributesData;
}

sub _getCachedUserDefinedAttributeData {
    my ($Shop, $Reset) = @_;

    # We'll keep this as a state variable so subsequent calls only rebuild the hash if the $Shop->alias is different
    state $hAttributesData = _getUserDefinedAttributeData($Shop);
    if ($Reset) {
        $hAttributesData = {};
    } elsif (not exists $hAttributesData->{$Shop->alias}) {
        $hAttributesData = _getUserDefinedAttributeData($Shop) unless (exists $hAttributesData->{$Shop->alias});
    }
    return $hAttributesData;
}

#========================================================================================
# function     _getProductUserDefinedAttributeValues
# state        private
#----------------------------------------------------------------------------------------
# syntax       _getProductUserDefinedAttributeValues($Shop, $Product);
#----------------------------------------------------------------------------------------
# description  gets the transformed data for user-defined producttype attributes
#               The base types are converted to the now types which
#               only support (Localized)String values (see UNITY-6516)
#----------------------------------------------------------------------------------------
# input        $Shop        | baseshop or prepshop | object
# input        $Product     | product to get user defined attribute values from | object
# return       $AttrInfo    | values of user-defined product attributes | ref hash
#========================================================================================
sub _getProductUserDefinedAttributeValues {
    my ($Shop, $Product) = @_;

    my $hAttributesData = _getCachedUserDefinedAttributeData($Shop);
    my $AttrInfo = {};
    my $ProductType = $Product->get('Class');
    my $CurrencyID = $Shop->get('CurrencyID');
    my $ActiveLanguages = GetActiveShopLanguages($Shop);
    my @AttributeInfo = @{$hAttributesData->{$Shop->alias}{$ProductType->id}};
    foreach my $hLanguage (@$ActiveLanguages) {
        my $LanguageID = $hLanguage->{'LanguageID'};
        my $ShopLanguage = shift @{LoadClassByAlias('ShopLanguage')->getObjects('ByShopIDLanguageID', [ $Shop->id, $LanguageID ])};
        my $LocaleID = $Shop->localeForLanguageByShopLocaleIDAndLanguage($ShopLanguage);
        my $Locale = $Shop->siteLocale($LocaleID);
        my @AttributeNames = map {$_->{Alias}} @AttributeInfo;
        my $Values = $Product->getWithParent([ @AttributeNames ], $LanguageID);
        foreach my $Attribute (@AttributeInfo) {
            my $Alias = $Attribute->{Alias};
            my $IsLocalized = 1;
            my $Type = $Attribute->{Type};
            my $IsHTML = ($Type eq 'String' || $Type eq 'LocalizedString') ? $Attribute->{IsHTML} : 0;
            my $Value = $Values->{$Alias};

            if ($Type eq 'PreDefString') {
                $Value = defined $Value ? $Value->alias : '';
                $IsLocalized = 0;
            }
            elsif ($Type eq 'PreDefLocalizedString') {
                $Value = defined $Value ? $Value->alias : '';
                $IsLocalized = 0;
            }
            elsif ($Type eq 'Integer') {$Value = defined $Value ? $Locale->formatInteger($Value) : '';}
            elsif ($Type eq 'Float') {$Value = defined $Value ? $Locale->formatFloat($Value) : '';}
            elsif ($Type eq 'Money') {$Value = defined $Value ? $Locale->formatMoney($Value, $CurrencyID) : '';}
            elsif ($Type eq 'DateTime') {$Value = defined $Value ? $Locale->formatDateTime($Value, 'datetime') : '';}
            elsif ($Type eq 'Date') {$Value = defined $Value ? $Locale->formatDateTime($Value, 'date') : '';}
            elsif ($Type eq 'Time') {$Value = defined $Value ? $Locale->formatDateTime($Value, 'time') : '';}

            if ($IsHTML) {
                my $HTMLTree = HTML::TreeBuilder->new();
                $HTMLTree->parse($Value);
                $HTMLTree->eof();
                my $Formatter = HTML::FormatText->new();
                $Value = $Formatter->format($HTMLTree);
            }
            if ($IsLocalized) {
                $AttrInfo->{LocalizedAttributes}{$hLanguage->{'Code2'}}{$Alias} = $Value . '';
            } else {
                $AttrInfo->{Attributes}{$Alias} = $Value . '';
            }
        }
    }
    return $AttrInfo;
}

#========================================================================================
# function     SetProducts
# state        public
#----------------------------------------------------------------------------------------
# syntax       SetProducts($Shop, $ahProducts);
#----------------------------------------------------------------------------------------
# description  sets attributes on shopProducts as specified in the $ahProducts
#----------------------------------------------------------------------------------------
# input        $Shop         | baseshop or prepshop | object
# input        $ahProducts  | see GetProductData() | reference to array of hashes
#========================================================================================
sub SetProducts {
    my ($Shop, $ahProducts) = @_;

    foreach my $hProduct (@$ahProducts) {
        SetProductData($Shop, $hProduct);
    }
}

#========================================================================================
# function     ExportProductFiles
# state        public
#----------------------------------------------------------------------------------------
# syntax       ExportProductFiles($Product, $ToDirectory);
#----------------------------------------------------------------------------------------
# description  exports files belonging to PRODUCTS_FILE_ATTRIBUTES to the export dir but:
#               If we export from a BaseShop we use only ImageAttributes, replace Underscores with minuses and
#               scale the images from scratch
#               If we export from a PrepShop we just copy the whole shebang
#----------------------------------------------------------------------------------------
# input        $Product      | product | object
# input        $ToDirectory  | export directory | path
# input        $hOptions     | further flags to control exportFileData
#                               the following flags can be used:
#                                <ul>
#                                  <li>CopyMediumImages - also copy MediumImage file</li>
#                                </ul>
#                             | ref hash
# return       $aFiles       | list of all copied/generated files | ref array string
#========================================================================================
sub ExportProductFiles {
    my ($Product, $ToDirectory, $hOptions) = @_;

    return unless $EXPORT_FILES_AND_FILEATTRIBUTES;
    my $WebRootShopDir = $Product->getSite()->get('PublicPath');
    my $Shop = $Product->getSite();
    RunOnDirectory($WebRootShopDir,
        sub {
            my $TargetDir = AbsolutePath($ToDirectory);
            if ($Shop->get('IsMigrationPrepShop')) {
                my $BasePath = _GetRelativeFilePathPart($Product, $Shop);
                my $TargetPath = catdir($TargetDir, _GetRelativeFilePathPart($Product, $Shop));
                _copyFiles($BasePath, $TargetPath, { Type => 'ExportAll', Alias => $Product->alias, DEBUG => $DEBUG });
            } else {
                foreach my $AttributeName (@{PRODUCTS_FILE_ATTRIBUTES()}) {
                    exportFileData($Product, $AttributeName, $TargetDir, $hOptions);
                }
            }
        }
    );
}

#========================================================================================
# function     ImportProductFiles
# state        public
#----------------------------------------------------------------------------------------
# syntax       ImportProductFiles($Product, $FromDirectory, $OldPath, $hOptions);
#----------------------------------------------------------------------------------------
# description  imports all product image related files from the ExportDirectory
#               and translates the PublicPath to the new PublicPath
#               Scaling is only needed on Base2Prep
#----------------------------------------------------------------------------------------
# input        $Product        | product | object
# input        $OldPath        | original publicpath of the product | string
# input        $FromDirectory  | export base directory | path
# input        $hOptions       | further flags to control the import
#                                 the following flags can be used:
#                                   <ul>
#                                     <li>ScaleImages - create scaled versions of images</li>
#                                   </ul>
#                               | ref hash
#========================================================================================
sub ImportProductFiles {
    my ($Product, $FromDirectory, $OldPath, $hOptions) = @_;

    state $WebRootShopDir = catdir(LoadRootObject->get('PublicPath'), $Product->getSite()->get('FilePath'));
    my $Shop = $Product->getSite();
    RunOnDirectory($WebRootShopDir,
        sub {
            my $SourceDir = AbsolutePath(catdir($FromDirectory, $OldPath));
            my $TargetDir = $Product->get('PublicPath');
            # UNITY-8408 Handle legacy underscore files in BaseShop before copying Prep images
            #            that potentially contain underscores due to UNITY-8407
            # UNITY-6881 move old underscore images for consistency
            if (!$Shop->get('IsMigrationPrepShop') && !defined $Product->get('SuperProduct')) {
                _handleUnderScoreFiles($Product, $TargetDir, {
                    BackupDir => catdir(dirname($FromDirectory), 'Backup'),
                    RelPath   => _GetRelativeFilePathPart($Product, $Shop),
                    Type      => 'HandleUnderScoreFiles', Alias => $Product->alias, DEBUG => $DEBUG
                });
            }
            # UNITY-6990 skip the copy for sub products as it is already done on the master
            if (!defined $Product->get('SuperProduct')) {
                _copyFiles($SourceDir, $TargetDir, { Type => 'Import', Alias => $Product->alias, DEBUG => $DEBUG });
            }
            if ($hOptions->{ScaleImages}) {
                _shortenFileNamesInFolder($Product, $TargetDir);
                _createScaledMediumImages($Product);
            } else {
                $Product->addImageInformations([UNITY_IMAGE_ATTRIBUTES], undef);
                $Product->extendImageInformationsSlideShowString(undef);
            }
        });
}

sub exportFileData {
    my ($Product, $AttributeName, $TargetDir, $hOptions) = @_;

    my @Ignore = qw(ImageMediumLarge ImageMediumSmall ImageSmall ImageThumbnail ImageHotDeal);
    push @Ignore, 'ImageMedium' unless $hOptions->{CopyMediumImages};
    my $Shop = $Product->getSite();
    $hOptions->{ReplaceUnderscores} = $Shop->get('IsMigrationPrepShop') ? 0 : 1;
    my $BasePath = _GetRelativeFilePathPart($Product, $Shop);
    @{$hOptions}{qw(Type Alias DEBUG)} = ($AttributeName, $Product->alias, $DEBUG);

    if ($AttributeName eq 'ImagesSlideShowString') {
        my $ImagesSlideShowString = $Product->get($AttributeName);
        return unless defined $ImagesSlideShowString;
        my @ImagesInString = trim(split /;/, $ImagesSlideShowString);
        foreach my $FileName (@ImagesInString) {
            next if (StringIsURL($FileName));
            my $ImagePath = catfile($BasePath, $FileName);
            my $TargetFile = catfile($TargetDir, $ImagePath);
            eval {TestImageFormat($AttributeName, $ImagePath)};
            if (ExistsError()) {
                LogDebug(sprintf('Skipping unsupported image %s on %s (%s)',
                    $FileName, $Product->pathString, $AttributeName), 1);
                next;
            }
            _copyFile($ImagePath, $TargetFile, $hOptions) unless ExistsError();
        }
    }
    elsif ($AttributeName eq '_ProductDataSheet') {
        my $ActiveLanguages = GetActiveShopLanguages($Shop);
        foreach my $hLanguage (@$ActiveLanguages) {
            my $FileName = $Product->getWithParent('_ProductDataSheet', $hLanguage->{'LanguageID'});
            next unless defined $FileName;
            next if (StringIsURL($FileName));
            my $ImagePath = catfile($BasePath, $FileName);
            my $TargetFile = catfile($TargetDir, $ImagePath);
            local $hOptions->{ReplaceUnderscores} = 0;
            _copyFile($ImagePath, $TargetFile, $hOptions);
        }
    } else {
        return if ElementOf($AttributeName, @Ignore);
        my $FileName = $Product->getWithParent($AttributeName);
        return unless defined $FileName;
        return if (StringIsURL($FileName));
        my $ImagePath = catfile($BasePath, $FileName);
        my $TargetFile = catfile($TargetDir, $ImagePath);
        eval {TestImageFormat($AttributeName, $ImagePath)};
        if (ExistsError()) {
            LogDebug(sprintf('Skipping unsupported image %s on %s (%s)',
                $FileName, $Product->pathString, $AttributeName), 1);
            next;
        }
        local $hOptions->{ReplaceUnderscores} = 0 if ($AttributeName eq 'EnergyLabelSourceFile');
        _copyFile($ImagePath, $TargetFile, $hOptions) unless ExistsError();
    }
}

sub _isFileNewer {
    my ($Source, $Target) = @_;

    my $SourceModified = -M $Source;
    my $TargetModified = -M $Target // $SourceModified + 1;

    return $SourceModified < $TargetModified ? 1 : 0;
}

sub _copyFile {
    my ($Source, $Target, $hOptions) = @_;

    if (ExistsFile($Source) && _isFileNewer($Source, $Target)) {
        # UNITY-6986 only rename the basename not the whole directory (DemoShop content)
        my $TargetBaseName = basename($Target);
        $TargetBaseName =~ s/_/-/g if $hOptions->{ReplaceUnderscores};
        $Target = catfile(dirname($Target), $TargetBaseName);
        printf "[%s/%s] Copy %s -> %s\n", $hOptions->{Alias}, $hOptions->{Type}, basename($Source), $Target if $hOptions->{DEBUG};
        CopyFile($Source, $Target, { create_directory => 1 });
    }
}

sub _copyFiles {
    my ($SourceDir, $TargetDir, $hOptions) = @_;

    if (ExistsDirectory($SourceDir)) {
        my $aFiles = GetFileNames($SourceDir, { relative => 1, recursive => 0 });
        foreach my $File (@$aFiles) {
            # UNITY-6986 only rename the basename not the whole directory (DemoShop content)
            my $TargetBaseName = $File;
            $TargetBaseName =~ s/_/-/g if $hOptions->{ReplaceUnderscores};
            my $TargetFile = catfile($TargetDir, $TargetBaseName);
            my $SourceFile = catfile($SourceDir, $File);
            if (_isFileNewer($SourceFile, $TargetFile)) {
                printf "[%s/%s] Copy %s -> %s\n", $hOptions->{Alias}, $hOptions->{Type}, basename($SourceFile), $TargetFile if $hOptions->{DEBUG};
                CopyFile($SourceFile, $TargetFile, { create_directory => 1 });
            }
        }
    }
}

sub _handleUnderScoreFiles {
    my ($Product, $Target, $hOptions) = @_;

    my $aExcept = _getExceptFileNames($Product);
    if (ExistsDirectory($Target)) {
        my $aFiles = GetFileNames($Target, { relative => 1, recursive => 0 });
        foreach my $FileName (@$aFiles) {
            # don't move files that are currently used and not stored in epages-storage
            next if (ElementOf(basename($FileName), @$aExcept));
            my (undef, $BaseName, $Extension) = fsplit($FileName);
            my $ImageType = $BaseName =~ s/(_(?:h|m|s|ms|ml|xs))$//x ? $1 : '';
            if ($BaseName =~ s/_/-/g) {
                my $TargetName = join('', ($BaseName, $ImageType, $Extension));
                my $FullSource = catfile($Target, $FileName);
                my $FullTarget = catfile($Target, $TargetName);
                my $BackupTarget = catfile($hOptions->{BackupDir}, $hOptions->{RelPath}, $FileName);
                if (ExistsFile($FullTarget)) {
                    LogDebug(sprintf("[%s/%s] Moving legacy file to Backup [%s] %s\n",
                        $hOptions->{Alias}, $hOptions->{Type}, $hOptions->{BackupDir}, basename($FileName)), 1);
                    MoveFile($FullSource, $BackupTarget, { create_directory => 1 });
                } else {
                    LogDebug(sprintf("[%s/%s] Renaming file %s -> %s\n",
                        $hOptions->{Alias}, $hOptions->{Type}, basename($FileName), basename($FullTarget)), 1);
                    MoveFile($FullSource, $FullTarget);
                }
            }
        }
    }
}

sub _createScaledMediumImages {
    my ($Product) = @_;

    my $Shop = $Product->getSite();
    my $BasePath = _GetRelativeFilePathPart($Product, $Shop);
    # Note: ImageLarge and ImagesSlideShowString can overlap especially on variation products
    foreach my $AttributeName (@{PRODUCTS_SCALE_IMAGE_ATTRIBUTES()}) {
        my $hOptions = {
            Type => $AttributeName, Alias => $Product->alias, Name => $AttributeName, DEBUG => $DEBUG, Shop => $Shop
        };
        if ($AttributeName eq 'ImagesSlideShowString') {
            my $ImagesSlideShowString = $Product->get($AttributeName);
            my @ImagesInString = trim(split /;/, $ImagesSlideShowString);
            my $SomeEntryChanged = 0;
            my @SlideShowImages = ();
            foreach my $FileName (@ImagesInString) {
                if (StringIsURL($FileName)) {
                    push @SlideShowImages, $FileName;
                    next;
                }
                # UNITY-6990 shorten image names and move files on prep import
                my $NewName = $FileName;
                if (exists $RenamedFiles{$FileName}) {
                    $SomeEntryChanged = 1;
                    $NewName = $RenamedFiles{$FileName};
                }
                push @SlideShowImages, $NewName;
                my $ImagePath = catfile($BasePath, $NewName);
                _createScaledMediumImage($ImagePath, $hOptions);
            }
            # UNITY-7055 always set ImageSlideShowString to update ImageInformations
            # Always add ImageLarge to the SlideShow for PrepShop if not already part of it
            my $ImageLarge = $Product->getWithParent('ImageLarge');
            $ImageLarge = $RenamedFiles{$ImageLarge} if (exists $RenamedFiles{$ImageLarge});
            unshift @SlideShowImages, $ImageLarge if ($ImageLarge && (!ElementOf($ImageLarge, @SlideShowImages)));
            $Product->set({ 'ImagesSlideShowString' => join(';', @SlideShowImages) });
        }
        elsif ($AttributeName eq 'ImageLarge') {
            my $FileName = $Product->getWithParent($AttributeName);
            # UNITY-7055 try to use the first SlideShowImage if ImageLarge is not defined
            if (not defined $FileName) {
                my $ImagesSlideShowString = $Product->get('ImagesSlideShowString');
                my @ImagesInString = trim(split /;/, $ImagesSlideShowString);
                $FileName = $ImagesInString[0] if (scalar @ImagesInString);
            }
            next unless defined $FileName;
            next if (StringIsURL($FileName));
            # UNITY-6990 shorten image names and move files on prep import
            my $NewName = $FileName;
            $NewName = $RenamedFiles{$FileName} if (exists $RenamedFiles{$FileName});
            my $ImagePath = catfile($BasePath, $NewName);
            my $Result = _createScaledMediumImage($ImagePath, $hOptions);
            $Product->set($Result);
        }
    }
}

sub _getExceptFileNames {
    my ($Product, $BasePath) = @_;

    my @Except = ();
    push @Except, $Product->getWithParent('EnergyLabelSourceFile');
    my $ActiveLanguages = GetActiveShopLanguages($Product->get('Site'));
    foreach my $hLanguage (@$ActiveLanguages) {
        push @Except, $Product->getWithParent('_ProductDataSheet', $hLanguage->{'LanguageID'});
    }
    return \@Except;
}

sub _shortenFileNamesInFolder {
    my ($Product, $BasePath) = @_;

    %RenamedFiles = () unless ($Product->get('SuperProduct'));
    state $AllImageTypes = [ q(), qw(_h _m _s _ms _ml _xs) ];
    my $aExcept = _getExceptFileNames($Product);
    return unless (ExistsDirectory($BasePath));
    foreach my $FileName (@{GetFileNames($BasePath, { 'recursive' => 0, 'relative' => 1 })}) {
        # keep short names
        next if (length($FileName) < MAX_IMAGE_FILENAME_LENGTH);
        # don't rename files not handled by epages-storage
        next if (ElementOf(basename($FileName), @$aExcept));
        # don't rename sub image files before renaming the base image file
        my (undef, $BaseName, $Extension) = fsplit($FileName);
        next if ($BaseName =~ /(_(?:h|m|s|ms|ml|xs))$/);
        if (not exists $RenamedFiles{$FileName}) {
            my $AnyFileExists = 0;
            my $NewBaseName;
            do {
                $NewBaseName = sprintf('img%d', GetEpochAsMilli());
                foreach my $TestImageType (@$AllImageTypes) {
                    my $TestFileName = catfile($BasePath, join('', ($NewBaseName, $TestImageType, $Extension)));
                    $AnyFileExists = 1 if (ExistsFile($TestFileName));
                }
            } while ($AnyFileExists);
            # Now that we have found a non-existing shorter filename move all variants
            foreach my $ImageType (@$AllImageTypes) {
                my $OldFileName = join('', ($BaseName, $ImageType, $Extension));
                my $OldFileNamePath = catfile($BasePath, $OldFileName);
                next unless (ExistsFile($OldFileNamePath));
                my $NewFileName = join('', ($NewBaseName, $ImageType, $Extension));
                my $NewFileNamePath = catfile($BasePath, $NewFileName);
                LogDebug(sprintf("Renaming too long image file on product '%s' from '%s' to '%s'\n",
                    $Product->alias, $OldFileName, $NewFileName), 1);
                MoveFile($OldFileNamePath, $NewFileNamePath);
                $RenamedFiles{$OldFileName} = $NewFileName;
            }
        }
    }
}

sub _createScaledMediumImage {
    my ($ImagePath, $hOptions) = @_;

    TestString('ImagePath', $ImagePath);
    my ($FileNamePath, $Ext) = $ImagePath =~ /(.*)[.](.*?)$/;
    my $FileNameMediumImage = $FileNamePath . '_m.' . $Ext;
    my $FileNameMask = basename($FileNamePath) . '%s.' . $Ext;
    my $hResult = {};
    # Don't scale images for subproducts multiple times (same public path)
    unless (ExistsFile($FileNameMediumImage)) {
        printf "[%s/%s] Scaling image %s\n", $hOptions->{Alias}, $hOptions->{Name}, $ImagePath if $hOptions->{DEBUG};
        $hResult = ScaleImageSizes($ImagePath, _createScaleParameters($hOptions));
    } else {
        $hResult = {
            ImageHotDeal     => sprintf($FileNameMask, '_h'),
            ImageLarge       => sprintf($FileNameMask, ''),
            ImageMedium      => sprintf($FileNameMask, '_m'),
            ImageMediumLarge => sprintf($FileNameMask, '_ml'),
            ImageMediumSmall => sprintf($FileNameMask, '_ms'),
            ImageSmall       => sprintf($FileNameMask, '_s'),
            ImageThumbnail   => sprintf($FileNameMask, '_xs'),
        };
    }
    return $hResult;
}

sub _createScaleParameters {
    my ($hOptions) = @_;

    my $hParams = {};
    my @Sizes = keys %{PRODUCT_IMAGE_SIZE_NAMES()};

    # Sizes from Constants
    state $hProductImageSizes = GetProductImageSizes($hOptions->{Shop});
    foreach my $Size (@Sizes) {
        next if ($Size eq 'l');
        $hParams->{'SCALE_NAME_' . $Size} = PRODUCT_IMAGE_SIZE_NAMES->{$Size};
        $hParams->{'SCALE_WIDTH_' . $Size} = $hProductImageSizes->{$Size}->[0];
        $hParams->{'SCALE_HEIGHT_' . $Size} = $hProductImageSizes->{$Size}->[1];
    }
    $hParams->{donotoverwrite} = 1;
    return $hParams;
}

#========================================================================================
# function     SetProductData
# state        public
#----------------------------------------------------------------------------------------
# syntax       SetProductData($Shop, $hProduct);
#----------------------------------------------------------------------------------------
# description  sets attributes on a product as specified in the hash $hProduct
#----------------------------------------------------------------------------------------
# input        $Shop         | baseshop or prepshop | object
# input        $hProduct     | see GetProductData() | reference to hash
# return       $Product      | updated or inserted product | object
#========================================================================================
## no critic qw( ProhibitExcessComplexity )
sub SetProductData {
    my ($Shop, $hProduct) = @_;

    # reset the cached product type attributes as they can change to LocalizedString
    _getCachedUserDefinedAttributeData($Shop, 1);
    my $UsePrepAttributes = $Shop->get('IsMigrationPrepShop') ? 0 : 1;
    my ($Attributes, $ObjectAttributes, $LocalizedAttributes) = _selectAttributesArrayRefsForProductsData($UsePrepAttributes);

    my $SuperProductPath = $hProduct->{Attributes}{SuperProduct};
    delete $hProduct->{Attributes}{SuperProduct};
    _checkPathExists($Shop, $SuperProductPath, 'SuperProduct');
    my $SuperProduct = defined $SuperProductPath ? LoadObjectByPath($SuperProductPath, $Shop->id) : undef;
    my $ProductFolder = defined $SuperProduct ? $SuperProduct->folder('SubProducts') : $Shop->folder('Products');

    my $ParentPath = $hProduct->{Attributes}{Parent};
    _checkPathExists($Shop, $ParentPath, 'Parent');
    delete $hProduct->{Attributes}{Parent};
    # read-only attribute only required for file/image mapping later
    delete $hProduct->{Attributes}{RelativeFilePathPart};

    my $ClassAlias = $hProduct->{Attributes}{Class};
    delete $hProduct->{Attributes}{Class};
    my $ProductTypeFolder = $Shop->get('ProductTypes');
    my $ProductClass = defined $ClassAlias ? $ProductTypeFolder->child($ClassAlias) : $Shop->get('ProductClass');

    my $Alias = $hProduct->{Attributes}{Alias};
    # UNITY-6955 if product exists with different class as variation product rename the old one and insert new
    if ($ProductFolder->existsChild($Alias)) {
        my $OldProduct = $ProductFolder->child($Alias);
        if ($OldProduct->get('IsMaster') and $OldProduct->get('Class') ne $ProductClass) {
            my $NewAlias = sprintf('%s_%s', $Alias, time());
            LogDebug(sprintf("Renaming old variation master product '%s' to '%s' due to change in ProductType!\n",
                $Alias, $NewAlias), 1);
            $OldProduct->set({ 'Alias' => $NewAlias, 'IsVisible' => 0 });
        }
    }
    my $Product = $ProductFolder->existsChild($Alias) ?
        $ProductFolder->child($Alias) :
        $ProductClass->insertObject({ 'Alias' => $Alias, 'Parent' => $ProductFolder, 'SuperProduct' => $SuperProduct });
    my $IsBaseShop = $Shop->get('IsMigrationPrepShop') ? 0 : 1;
    if ($IsBaseShop) {
        my $OldGUID = $hProduct->{Attributes}{GUID};
        my $NewGUID = $Product->get('GUID');
        eval {
            _UpdateGUIDinUI($Shop, $OldGUID, $NewGUID) unless $SKIP_UI;
        };
        if (ExistsError()) {
            my $Error = GetError();
            if ($Error->code() eq 'UI_STORE_NOT_DEFINED') {
                Warning($Error);
            } else {
                Error($Error);
            }
        }
    }
    delete $hProduct->{Attributes}{GUID};
    # UNITY-7460 update SelectedVariations
    my $aSelectedVariationsAliases = $hProduct->{Attributes}{SelectedVariations};
    delete $hProduct->{Attributes}{SelectedVariations};

    # UNITY-8409: Convert serialized DateTimes attributes back to DateTime
    my @DateTimeAttributes = @{PRODUCTS_ATTRIBUTES_DATETIME()};
    foreach my $AttributeName (@DateTimeAttributes) {
        if (defined $hProduct->{Attributes}{$AttributeName}) {
            $hProduct->{Attributes}{$AttributeName} = ParseDateTime($hProduct->{Attributes}{$AttributeName});
        }
    }

    foreach my $Type (keys %$hProduct) {
        if ($Type eq 'Attributes') {
            my $aUserDefinedAttributes = $ProductClass->get('UserDefinedAttributes');
            foreach my $Key (keys %{$hProduct->{Attributes}}) {
                # Replace PreDefLocalizedString attributes alias with their PreDefAttributeValue
                if (ElementOf($Key, map {$_->alias} @{$aUserDefinedAttributes})) {
                    foreach my $UserAttribute (@$aUserDefinedAttributes) {
                        next unless ($Key eq $UserAttribute->alias);
                        if ($UserAttribute->get('Type') eq 'PreDefLocalizedString') {
                            $hProduct->{Attributes}{$Key} =
                                $hProduct->{Attributes}{$Key} eq '' ?
                                    undef :
                                    $UserAttribute->preDefAttributeValue($hProduct->{Attributes}{$Key});
                        }
                    }
                } elsif (!ElementOf($Key, @$Attributes)) {
                    die "Invalid Element for Attributes: $Key";
                }
            }
            foreach my $AttributeName (@$ObjectAttributes) {
                my $PathString = $hProduct->{Attributes}{$AttributeName};
                next unless defined $PathString;
                if ($AttributeName eq 'TaxClass') {
                    my $TaxMatrix = $Shop->get('TaxMatrix');
                    my $TaxClassID = $TaxMatrix->findChild($PathString);
                    $hProduct->{Attributes}{$AttributeName} = defined $TaxClassID ? LoadObject($TaxClassID) : $TaxMatrix->get('DefaultTaxClass');
                } elsif ($AttributeName eq 'TaxClassDigital') {
                    my $TaxMatrix = $Shop->get('TaxMatrixDigital');
                    my $TaxClassID = $TaxMatrix->findChild($PathString);
                    $hProduct->{Attributes}{$AttributeName} = LoadObject($TaxClassID) // $TaxMatrix->get('DefaultTaxClassDigital');
                } elsif (ExistsObjectByPath($PathString, $Shop->id)) {
                    $hProduct->{Attributes}{$AttributeName} = $Shop->path($PathString);
                } elsif (ExistsObjectByPath($PathString)) {
                    $hProduct->{Attributes}{$AttributeName} = LoadObjectByPath($PathString);
                } else {
                    Error('REFERENCED_OBJECT_NOT_FOUND', { Name => $AttributeName, Path => $PathString });
                }
            }
        }
        elsif ($Type eq 'LocalizedAttributes') {
            my $aPreDefAttributes = _getPreDefLocalizedStringAttributes($ProductClass);
            foreach my $LanguageCode2 (keys %{$hProduct->{LocalizedAttributes}}) {
                my @Allowed = map {$_->alias} @{$ProductClass->get('UserDefinedAttributes')};
                foreach my $Key (keys %{$hProduct->{LocalizedAttributes}{$LanguageCode2}}) {
                    if (!ElementOf($Key, @$LocalizedAttributes) && !ElementOf($Key, @Allowed)) {
                        die "Invalid Element for Attributes: $Key";
                    }
                    if (ElementOf($Key, @$aPreDefAttributes)) {
                        my $Value = $hProduct->{LocalizedAttributes}{$LanguageCode2}{$Key};
                        _checkPathExists($Shop, $Value, 'PreDefLocalizedString');
                        $hProduct->{LocalizedAttributes}{$LanguageCode2}{$Key} = $Shop->path($Value);
                    }
                }
            }
        } else {
            die "Invalid Type: $Type";
        }
    }

    Transaction(sub {
        # UNITY-6955 also change Class on import if necessary
        my $OldProductClass = $Product->get('Class');
        $Product->set({ Class => $ProductClass }) if ($ProductClass ne $OldProductClass);
        foreach my $Type (keys %$hProduct) {
            if ($Type eq 'Attributes') {
                my $hPrices = $hProduct->{Attributes}{ListPrices};
                delete $hProduct->{Attributes}{ListPrices};
                # need to set TaxClass first..
                $Product->set($hProduct->{Attributes});
                # then prices..
                foreach my $hPrices (@$hPrices) {
                    _setPrices($Product, $hPrices)
                }
            }
            elsif ($Type eq 'LocalizedAttributes') {
                foreach my $LanguageCode2 (keys %{$hProduct->{LocalizedAttributes}}) {
                    foreach my $Key (keys %{$hProduct->{LocalizedAttributes}{$LanguageCode2}}) {
                        my $LanguageID = GetPKeyLanguageByCode($LanguageCode2);
                        # Silently ignore fields for non-existing languages
                        if ($Shop->existsLanguage($LanguageID)) {
                            $Product->set($hProduct->{LocalizedAttributes}{$LanguageCode2}, $LanguageID);
                        }
                    }
                }
            }
        }
        _handleSelectedVariations($aSelectedVariationsAliases, $Product);

        # add redirect entries for each URI to epages-ui on final migration
        if ($IsBaseShop) {
            my $aLocales = GetAllUILocalesForShop($Shop);
            foreach my $LocaleID (@$aLocales) {
                last if $SKIP_UI;
                my $LanguageID = GetLanguageByLocaleID($LocaleID);
                my $FromSlug = $Product->get('URI', $LanguageID);
                if (defined $FromSlug) {
                    my $ToSlug = '/' . GetCodeByLanguageID($LanguageID) . '/p/' . $Product->guid;
                    InsertRedirect($Shop, $FromSlug, $ToSlug);
                }
            }
        }
    });
    return $Product;
}
## use critic

#========================================================================================
# function     _handleSelectedVariations
# state        private
#----------------------------------------------------------------------------------------
# syntax       _handleSelectedVariations($aSelectedVariationsAliases, $Product);
#----------------------------------------------------------------------------------------
# description  update SelectedVariations of master product with entries from
#               $aSelectedVariationsAliases
#----------------------------------------------------------------------------------------
# input        $aSelectedVariationsAliases  | array with used variation attributes | ref array
# input        $Product  | product | object
#========================================================================================
sub _handleSelectedVariations {
    my ($aSelectedVariationsAliases, $Product) = @_;
    my $SuperProduct = $Product->get('SuperProduct');
    my $Class = $Product->class();

    my $VariationColumns = {};
    $VariationColumns->{$_} = {} foreach @$aSelectedVariationsAliases;
    if (defined $SuperProduct) {
        # Fetch existing attribute values either from SuperProduct if existing or add them from $Class
        my @VariationAttributes = map {$_->{'Attribute'}} @{$SuperProduct->get('VariationTypeValues')};
        delete $VariationColumns->{ $_->alias } foreach @VariationAttributes;
        foreach my $AttrAlias (keys %$VariationColumns) {
            push @VariationAttributes, $Class->attribute($AttrAlias) if ($Class->existsAttribute($AttrAlias));
        }
        # Extend SelectedVariations if a value on current Product is set and not already in SelectedVariations
        if (scalar @VariationAttributes) {
            my @VariationValueIDs;
            foreach my $Attribute (@VariationAttributes) {
                my $AttrAlias = $Attribute->alias;
                my $Value = $Product->getWithParent($AttrAlias);
                push @VariationValueIDs, $Value->id if defined $Value;
            }
            #update super products SelectedVariations
            my @ExistingSelectedVariations = map {$_->{'SelectedID'}} @{$SuperProduct->get('SelectedVariations')};
            my @UniqueValues = unique(@ExistingSelectedVariations, @VariationValueIDs);
            $SuperProduct->set({ 'SelectedVariations' => [ map {{ 'SelectedID' => $_ }} @UniqueValues ] });
        }
    }
}

#========================================================================================
# function     _getPreDefLocalizedStringAttributes
# state        private
#----------------------------------------------------------------------------------------
# syntax       _getPreDefLocalizedStringAttributes($ProductType);
#----------------------------------------------------------------------------------------
# description  returns the names of all user defined attributes for a product type
#----------------------------------------------------------------------------------------
# input        $ProductClass  | producttype | object
# return       $aAttributes   | all user-defined attribute aliases | ref array
#========================================================================================
sub _getPreDefLocalizedStringAttributes {
    my ($ProductClass) = @_;
    return [ map {$_->alias} grep {$_->get('Type') eq 'PreDefLocalizedString'} @{$ProductClass->get('UserDefinedAttributes')} ];
}

#========================================================================================
# function     _checkPathExists
# state        private
#----------------------------------------------------------------------------------------
# syntax       _checkPathExists($Shop, $Path, $Name);
#----------------------------------------------------------------------------------------
# description  returns the names of all user defined attributes for a product type
#----------------------------------------------------------------------------------------
# input        $Shop  | shop object to check path from | object
# input        $Path  | path to check | string
# input        $Name  | additonal info text for error | string
# return       $aAttributes   | all user-defined attribute aliases | ref array
#========================================================================================
sub _checkPathExists {
    my ($Shop, $Path, $Name) = @_;
    return unless defined $Path;
    unless (ExistsObjectByPath($Path, $Shop->id) || ExistsObjectByPath($Path)) {
        Error('REFERENCED_OBJECT_NOT_FOUND', { Name => $Name, Path => $Path });
    }
}

#========================================================================================
# function     _setPrices
# state        private
#----------------------------------------------------------------------------------------
# syntax       _setPrices($Product, $hPrices)
#----------------------------------------------------------------------------------------
# description  sets the prices on a given product
#----------------------------------------------------------------------------------------
# input        $Product  | product to set prices on | object
# input        $hPrices  | hash with all price infos | ref hash
#========================================================================================
sub _setPrices {
    my ($Product, $hPrices) = @_;

    my $CurrencyID = $hPrices->{'CurrencyID'};
    $Product->setDeposit($CurrencyID, undef, $hPrices->{Deposit}) if defined $hPrices->{Deposit};
    $Product->setEcoParticipation($CurrencyID, undef, $hPrices->{EcoParticipation}) if defined $hPrices->{EcoParticipation};
    $Product->setManufacturerPrice($CurrencyID, undef, $hPrices->{ManufacturerPrice}) if defined $hPrices->{ManufacturerPrice};
    $Product->setListPrice($CurrencyID, undef, $hPrices->{ListPrice}) if defined $hPrices->{ListPrice};
}

#========================================================================================
# function     _setPrices
# state        private
#----------------------------------------------------------------------------------------
# syntax       _selectAttributesArrayRefsForProductsData($UsePrepAttributes)
#----------------------------------------------------------------------------------------
# description  sets the prices on a given product
#----------------------------------------------------------------------------------------
# input        $UsePrepAttributes  | flag that decides which constants to use | boolean
#               1 -> use Constants named *PREP* (PrepExport and BaseImport)
#               0 -> use Constants named *BASE* (BaseExport and PrepImport)
#========================================================================================
# Helper function that copies the proper (constant) ArraysRefs depending to use on Get/Set
sub _selectAttributesArrayRefsForProductsData {
    my ($UsePrepAttributes) = @_;

    my ($Attributes, $ObjectAttributes, $LocalizedAttributes, $ScribeAttributes);
    if ($UsePrepAttributes) {
        $Attributes = [ @{PRODUCTS_ATTRIBUTES_FOR_PREP_EXPORT()} ];
        $ObjectAttributes = [ @{PRODUCTS_OBJECTATTRIBUTES_FOR_PREP_EXPORT()} ];
        $LocalizedAttributes = [ @{PRODUCTS_LOCALIZEDATTRIBUTES_FOR_PREP_EXPORT()} ];
        $ScribeAttributes = [];
    } else {
        $Attributes = [ @{PRODUCTS_ATTRIBUTES_FOR_BASE_EXPORT()} ];
        $ObjectAttributes = [ @{PRODUCTS_OBJECTATTRIBUTES_FOR_BASE_EXPORT()} ];
        $LocalizedAttributes = [ @{PRODUCTS_LOCALIZEDATTRIBUTES_FOR_BASE_EXPORT()} ];
        $ScribeAttributes = [ @{PRODUCTS_SCRIBEATTRIBUTES_FOR_BASE_EXPORT()} ];
    }
    return ($Attributes, $ObjectAttributes, $LocalizedAttributes, $ScribeAttributes);
}

#========================================================================================
# function     _UpdateGUIDinUI
# state        private
#----------------------------------------------------------------------------------------
# syntax       _UpdateGUIDinUI($OldGUID, $NewGUID)
#----------------------------------------------------------------------------------------
# description  updates the prdocutids in the json field 'content' in the contens
#               table in order display the correct products in the productslider
#----------------------------------------------------------------------------------------
# input        $Shop         | baseshop or prepshop | object
# input        $OldGUID      | (old) guid in the prepshop | string
# input        $NewGUID      | (new) guid in the base shop | string
#========================================================================================
sub _UpdateGUIDinUI {
    my ($Shop, $OldGUID, $NewGUID) = @_;
    my $Store = LoadRootObject()->get('StoreName');
    my $Slug = GetUnitySubDomain($Shop->alias);
    my $UIStore = DatabaseConnectionData($Store)->{InstanceUnityUI};
    if (!defined $UIStore || !$UIStore) {
        Error('UI_STORE_NOT_DEFINED', { Store => $Store })
    } else {
        RunOnStore(
            Store => $UIStore,
            Sub   => sub {
                my $dbi = GetCurrentDBHandle();
                my $sth = $dbi->executeStatement(<<SQL,
UPDATE contents
SET content = REPLACE(content, ?, ?)
WHERE id IN (
  SELECT c2.id
  FROM (SELECT id, pageid from contents) as c2
  INNER JOIN pages p ON c2.pageid=p.id
  WHERE p.shopid IN (
    SELECT id FROM shops WHERE shopGateID = (
      SELECT id FROM shopGates WHERE slug=?
    )
  )
)
AND content like ?
SQL
                    [ $OldGUID, $NewGUID, $Slug, sprintf('%%%s%%%', $OldGUID) ]);
                LogDebug(sprintf("Updating in %s.contents [%s]: GUID %s -> GUID %s: affected %s rows\n",
                    $UIStore, $Slug, $OldGUID, $NewGUID, $sth->rows), 1);
            }
        );
    }
}

1;
