use strict;
use Test::More;
use Time::HiRes qw ( time );
use JSON qw ( from_json to_json );
use Readonly;
use XML::SemanticDiff;
use DE_EPAGES::Core::API::Error qw ( HandleError );
use DE_EPAGES::Core::API::File  qw ( AbsolutePath DeleteFile GetFileContent WriteFile CorrectPathSlashes);
use DE_EPAGES::Core::API::Script  qw ( RunScript );
use DE_EPAGES::Database::API::Connection qw ( GetTestSite RunOnStore );
use DE_EPAGES::Object::API::Factory qw ( ExistsObjectByPath JoinObjectPath LoadClassByAlias LoadObjectByPath LoadRootObject ResetFactoryCaches );
use DE_EPAGES::Object::API::XML::ObjectExportDriver qw ( ExportObjects );
use DE_EPAGES::Object::API::XML::ObjectImportHandler qw ( ImportObjects );
use DE_EPAGES::WebInterface::API::MessageCenter qw ( SynchronizeCache );

Readonly::Scalar my $EXPORT_SHOP_TIME            => 10;  # 6.14.1: 8.1 ; 6.13.2: 11.5; 6.13.3: 13; 6.13.4: 11.
Readonly::Scalar my $CREATE_SHOP_TIME            => 7;   # 6.14.1: 5.5 ; 6.13.2: 10  ; 6.13.3: 12. 6.13.4:  9.
Readonly::Scalar my $EXPORT_CSS_STLYESHEETS_TIME => 3;   #               6.13.3: 4 ; 6.13.4:  3.
Readonly::Scalar my $REEXPORT_IMPORTED_SHOP_TIME => 6.5; # 6.14.1: 5.1 ; 6.13.3: 9 ; 6.13.4:  9.5.
Readonly::Scalar my $IMPORT_SHOP_TIME            => 46;  # 6.14.1: 45  ; 6.13.2: 36
Readonly::Scalar my $DELETE_SHOP_TIME            => 38;  # 6.14.1: 37  ; 6.13.2: 45

Readonly::Scalar my $DEFAULT_SHOP_XML => join ('/', $ENV{EPAGES_CARTRIDGES}, 'DE_EPAGES', 'Design', 'Database', 'XML', 'DefaultShop.xml');

## no critic qw(ProhibitInterpolationOfLiterals)

#========================================================================================
# function     _TimeOK
# state        private
#----------------------------------------------------------------------------------------
# syntax       _TimeOK( $MaxTime, $Comment, $Code );
#----------------------------------------------------------------------------------------
# description  Tests that the code block runs fast enough.
#----------------------------------------------------------------------------------------
# input        $MaxTime | maximum runtime in seconds | float
# input        $Comment | comment | string
# input        $Code | code to execute | code reference
#========================================================================================
sub _TimeOK {
    my ($MaxTime, $Comment, $Code) = @_;

    my $time0 = time();
    $Code->();
    my $Time = time() - $time0;
    SKIP: {
        skip( 'Forced skip of timing tests.', 1 ) if $ENV{'SKIP_TIMING'};
        cmp_ok( $Time, '<', $MaxTime, sprintf("$Comment: %0.1f s", $Time) );
    }
    return;
}

# Add missing RoofXXX attributes to Style tag without altering /Shops/DemoShop
#
# BugFix: EPG-29965 AttributeSource: EPG-27383
sub _DirtyHackAddStyleTagAttributes($){
    my ($Content) = @_;

    # Path: DE_EPAGES/Design/Database/XML/DefaultShop.xml
    my $Path = CorrectPathSlashes( $DEFAULT_SHOP_XML );
    my $DefaultShopXML = GetFileContent($Path);

    my $StyleTagAttributes;
    if ( $$DefaultShopXML =~ m/<Style([^>]+?)>/x ){
        # parser expect attributes to be in single line
        my @AttributeLine = split( /\n/, $1 );
        # filter out non Roof-Attributes and drop duplicate RoofIsVisible (already in $Content)
        my @RoofAttributes = grep { $_ =~ m/^Roof/ && $_ !~ m/^RoofIsVisible/ } @AttributeLine;
        $StyleTagAttributes = join ( ' ', @RoofAttributes );
    } else {
        die "Couldn't extract Style attributes from file: $Path";
    }

    # inject attributes in Style tag
    $Content =~ s/<Style /<Style $StyleTagAttributes /x;

    return $Content;
}

sub _TestXMLExportImport($$$$) {
    my ($ShopConfigService, $StoreName, $ShopAlias, $NewAlias) = @_;

    # delete the clone shop if it exists from a previous test
    if( $ShopConfigService->exists({ 'Alias' => $NewAlias })->result ) {
        note 'delete previously created clone shop';
        $ShopConfigService->deleteShopRef({ 'Alias' => $NewAlias });
    }

    # Export shop
    my $FileName = 'DemoShop.xml';
    _TimeOK( $EXPORT_SHOP_TIME, 'export shop', sub {
        ExportObjects( '/Shops', $FileName, { withdefaults => 1 }, [$ShopAlias], \&_TransForm );
    });

    # Transform old alias into new one.
    my $Content = ${ GetFileContent( $FileName, { utf8 => 1 } ) };
    $Content =~ s/\Q$ShopAlias\E/$NewAlias/g;

    if ( $Content !~ m/RoofBackgroundColor/ ) {
        Test::More::diag 'EPG-29965 - Roof attributes are missing in Style tag -> add Roof attributes from DefaultShop.xml';
        # EPG-29965 - avoid semantic diff after patching 6.15.x -> 6.16.0
        # Caveat: We no longer test Style attributes fully :-\
        $Content = _DirtyHackAddStyleTagAttributes($Content);
    }

    my $ImportFile = 'DemoShopXMLimport.xml';
    WriteFile( $ImportFile, \$Content, { utf8 => 1 } );
    ok( 1, 'transform shop alias' );

    my $NewPath = JoinObjectPath( ['Shops', $NewAlias] );
    if( ExistsObjectByPath( $NewPath ) ) {
        note 'delete shop object of clone shop';
        LoadObjectByPath( $NewPath )->delete;
    }

    # Before creating a new shop, reset the default locale to en_GB
    is( LoadRootObject()->get('LocaleID'), 'en_GB', 'default system locale is en_GB, see DE_EPAGES::Object::API::Cartridge::_activateLocales' );

    my $EmptyFile = 'Empty.xml';
    my $Tag = '<epages />';
    WriteFile( $EmptyFile, \$Tag, { utf8 => 1 } );

    # create an empty shop
    _TimeOK( $CREATE_SHOP_TIME, 'create shop', sub {
        $ShopConfigService->create( {
            Database                => $StoreName,
            ShopType                => 'Demo',
            ShopAlias               => $NewAlias,
            Alias                   => $NewAlias,
            ImportFiles             => [ AbsolutePath($EmptyFile) ],
        });
    });

    # update cache for store database
    ResetFactoryCaches();

    # Import shop data from XML file
    _TimeOK( $IMPORT_SHOP_TIME, "import $ImportFile", sub {
        ImportObjects( $ImportFile, '/Shops' );
    });

    # inform other application servers
    SynchronizeCache();

    # export CSS stylesheets
    my $Shop = LoadObjectByPath( $NewPath );
    _TimeOK( $EXPORT_CSS_STLYESHEETS_TIME, 'export CSS stylesheets', sub {
        $_->writeCSS foreach @{ $Shop->get('Styles') };
    });

    # Export shop data again
    my $ExportFile = 'DemoShopXMLexport.xml';
    _TimeOK( $REEXPORT_IMPORTED_SHOP_TIME, 're-export imported shop', sub {
        ExportObjects( '/Shops', $ExportFile, { withdefaults => 1 }, [$NewAlias], \&_TransForm );
    });

    my $Equal = 1;

    my $diff = XML::SemanticDiff->new( keeplinenums => 1, keepdata => 1 );

    foreach my $change ($diff->compare($ImportFile, $ExportFile)) {
        my $Message = "$change->{message} in context $change->{context}".
            " (between lines $change->{startline} and $change->{endline})\n";
        $Message .= "  got     : $change->{new_value}\n" if defined $change->{new_value};
        if (defined $change->{old_value}){
            $Message .= "  expected: $change->{old_value}\n";
        } else {
            $Message .= "  expected: Value is undef, the generated file $ImportFile".
                " don't contain attribute/element with matching name.".
                " Check source /Shops/DemoShop for missing attribute/element)\n";
        }
        if (defined $change->{new_value} and defined $change->{old_value}) {
            my $NewValue = to_json(from_json($change->{new_value}), {pretty => 1});
            $NewValue = join('', sort map { s/,//g; $_ } split(/\n/, $NewValue));
            my $OldValue = to_json(from_json($change->{old_value}), {pretty => 1});
            $OldValue = join('', sort map { s/,//g; $_ } split(/\n/, $OldValue));
            next if $NewValue eq $OldValue;
        }
        Test::More::diag $Message;
        $Equal = 0;
    }
    ok( $Equal, 'export file equals import file' );

    # cleanup
    SKIP: {
        skip 'don\'t cleanup, because export files are different', 1 unless $Equal;

        _TimeOK( $DELETE_SHOP_TIME, 'deleteShopRef', sub {
            $ShopConfigService->deleteShopRef({ 'Alias' => $NewAlias });
        });

        DeleteFile( $_ ) foreach( $EmptyFile, $FileName, $ImportFile, $ExportFile );
    }
}

RunScript( 'Sub' => sub {

    my $NumTests = 10; ## no critic qw(ProhibitMagicNumbers)
    plan tests => $NumTests;

    if ($ENV{'SKIP_NON_WORKING'}) {
        plan skip_all => 'Test does not work reliable on patched installations.'
    }

    my $ShopAlias = 'DemoShop';
    my $StoreName = GetTestSite();
    my $ExistsDemoShop;
    RunOnStore( Store => $StoreName, Sub => sub {
        $ExistsDemoShop = ExistsObjectByPath( JoinObjectPath([ 'Shops', $ShopAlias ]) );

        # set the attribute NewsletterCampaign.TestEMail from Shop.Address.EMail
        # which is done by DE_EPAGES::ShopNewsletter::Hooks::NewsletterCampaign::OnInsertNewsletterCampaign
        # on XML import
        if( $ExistsDemoShop && LoadClassByAlias('Shop')->existsAttribute('NewsletterCampaigns') ) {
            my $Shop = LoadRootObject()->child('Shops')->child($ShopAlias);
            my $Address = $Shop->get('Address');
            if( $Address ) {
                my $Email = $Address->get('EMail');
                if( $Email ) {
                    foreach my $Campaign ( @{$Shop->get('NewsletterCampaigns')} ) {
                        $Campaign->set({ 'TestEMail' => $Email }) unless defined $Campaign->get('TestEMail');
                    }
                }
            }
        }
    });

    SKIP: {
        skip "shop $ShopAlias does not exists in database $StoreName", $NumTests unless $ExistsDemoShop;

        my $User = '/Providers/Distributor/Users/admin';
        my $Password = 'admin';
        my $WebServiceProxy;
        RunOnStore( 'Store' => 'Site', 'Sub' => sub {
            $WebServiceProxy = LoadRootObject()->folder('Stores')->child('Site')->get('WebServiceProxy');
        });
        ok( 1, "get web service proxy ($WebServiceProxy)" );

        # create a web service client
        my $ShopConfigService = DE_EPAGES::WebService::API::Client
            ->uri('urn://epages.de/WebService/ShopConfigService/2009/09')
            ->proxy( $WebServiceProxy, timeout => 900 );
        $ShopConfigService->userinfo("$User:$Password");

        my $NewAlias = 'DemoShopXMLimport';
        RunOnStore( Store => $StoreName, Sub => sub {
            eval { _TestXMLExportImport($ShopConfigService, $StoreName, $ShopAlias, $NewAlias) };
            HandleError( Finally => sub {
                SynchronizeCache();
            });
        });
    }
});


sub _TransForm {
    my ($TagName, $hAttributes) = @_;
    delete $hAttributes->{'GUID'};
    delete $hAttributes->{'CreationDate'};
    return;
}
